////////////////////////////////////////////////////////////////////////////
//
// string.h - string class which takes care of leaks and provides fast
//            string inequality test.
//
////////////////////////////////////////////////////////////////////////////
#ifndef _SSTRING_H_
#define _SSTRING_H_

#include <string.h>
#include "misc.h"

class ostream;
class istream;

// access to global string table w/o using String class
extern const char* strlit(const char*);
extern const char* rstrlit(const char*);
extern const char* strnlit(const char*, int);

// String class - all strings created with this class are entered
// into a global string table.  All common strings will point to
// the same entry in the table.  Strings in this table cannot be
// modified or all hell will break loose.
class String {
 public:
  // routines to convert String to int and char* for convenience
  const char* ToStr() const { return ptr; }
  operator const char*() const { return ptr; }
  operator const int() const { return (const int)ptr; }

  // length of string
  int Length() const { return ptr ? strlen(ptr) : 0; }

  // is this an empty string?
  int Null() const { return !ptr; }

  // allow access string like a char array
  const char& operator[](int i) const { return ptr[i]; }
  //int operator[](char c) const { char* p = ptr ? index(ptr, c) : 0; return p ? p - ptr : -1; }

  // obvious string operators
  int operator < (const String& s) const { return strcmp(ptr, s.ptr); }
  int operator > (const String& s) const { return strcmp(ptr, s.ptr); }
  int operator == (const String& s) const { return (ptr == s.ptr); }
  int operator == (const char* s) const { return (ptr == strlit(s)); }
  int operator != (const String& s) const { return (ptr != s.ptr); }
  String& operator=(const String& s) { ptr=s.ptr; return *this; }
  String& operator=(const char* s) { ptr=strlit(s); return *this; }
  String operator+(const String& s) const;

  friend String operator+(const String&, const int);
  friend String operator+(const char*, const String&);
  friend String operator+(const String&, const char*);

  // substring operator
  String operator()(int, int) const;

  // new empty string
  String() : ptr(0) { ; }

  // new string from single character
  String(char c) : ptr(0) { static char tmp[2]; *tmp = c; ptr = strlit(tmp); }

  // allow string to be initialized like "s = 0;" to make life easier
  String(int) : ptr(0) { ; }

  // copy a string
  String(const String& s) : ptr(s.ptr) { ; }

  // create new string
  String(const char* s) : ptr(strlit(s)) { ; }

  // destroy a string
  ~String() { ; }

  // enter char* into global string table
  static String Init(const char*);

  // string I/O operators
  friend ostream& operator<<(ostream& os, const String& s);
  friend istream& operator>>(istream& is, String& s);
 private:
  const char* ptr;  // pointer to actual string
};

// hash a string
extern unsigned int HashString(const char*&);

#endif
