////////////////////////////////////////////////////////////////////////////
//
// Vector.h - templatized adjustable vector class
//
////////////////////////////////////////////////////////////////////////////
#ifndef _VECTOR_H_
#define _VECTOR_H_

#include "SimpleObject.h"
#include "tsearch.h"

// heuristic for vector growth function
#define NewSize(a) ((a) + ((a) < 50 ? 2 : (a)))

// simple template vector class
template <class T> class Vector : public SimpleObject {
 public:
  // is the vector empty
  Boolean IsEmpty() const { return MakeBoolean(!_active); }

  // return the size of the vector
  unsigned int Size() const { return _active; }
  unsigned int Length() const { return _active; }

  // modify the size of the vector
  void SetSize(unsigned int, unsigned int=NotFound);
  void Expand(unsigned int n)
    {
      if(_active + n > _total)
	SetSize(_active + n, NewSize(_active + n));
      else
	_active += n;
    }
  void Shrink(unsigned int n) { assert(_active >= n); _active -= n; }

  // last vector element
  T& Tail() const { return _data[_active - 1]; }

  // return nth element (guarded access)
  T& Elt(unsigned int i) const { return (i >= _active) ? dummy() : _data[i]; }

  // return ith element from end of vector
  T& RElt(unsigned int i) const { return Elt(_active - i - 1); }

  // access vector like an array
  T& operator[](unsigned int i) const { return _data[i]; }

  // find the index of an element
  unsigned int IndexOf(T d) const { return lsearch_index(d, _data, _active); }

  // find the index of an element in a sorted vector
  unsigned int SortedIndexOf(T d) const { return bsearch_index(d, _data, _active, True); }

  // find the index of an element if it were to be inserted in a sorted vector
  unsigned int FindSortedIndexOf(T d) const { return bsearch_index(d, _data, _active, False); }

  // find element
  T& Find(T d) const { return Elt(IndexOf(d)); }
  T& FindSorted(T d) const { return Elt(SortedIndexOf(d)); }

  // test membership
  Boolean IsMember(T d) const { return MakeBoolean(IndexOf(d) != NotFound); }
  Boolean IsMemberSorted(T d) const { return MakeBoolean(SortedIndexOf(d) != NotFound); }

  // push and pop vector like a stack
  T Pop() { T res = _data[0]; Remove((unsigned int)0); return res; }
  T& Push(T d) { return Insert(0, d, InsertBefore); }

  // append elements to vector
  T& Append(T d) { Expand(1); return (_data[_active-1] = d); }
  Boolean AppendNew(T d) { return !IsMember(d) ? Append(d), True : False; }

  // insert elements
  T& Insert(unsigned int, T, Where = InsertBefore);
  T& InsertSorted(T d) { return Insert(FindSortedIndexOf(d), d); }

  // remove elements
  void Remove(unsigned int);
  void Remove(T d) { Remove(IndexOf(d)); }
  void RemoveSorted(T d) { Remove(SortedIndexOf(d)); }

  // swap vector elements
  void Swap(unsigned int i, unsigned int j) { tswap(T, _data[i], _data[j]); }

  // remove all vector elements
  void Erase() { while(Size()) Pop(); }

  // copy another vector
  void CopyVector(const Vector<T>& v);

  // append a vector to this one
  void AppendVector(const Vector<T>& v);

  // comparison operators
  Boolean operator==(const Vector<T>&) const;
  Boolean operator!=(const Vector<T>&) const;

  // print this vector
  ostream& Print(ostream&,int=0) const;

  // create an empty vector
  Vector() : _data(0), _active(0), _total(0) { }

  // create a vector with 1 element
  Vector(const T& data) : _data(0), _active(0), _total(0) { Append(data); }

  // copy another vector
  Vector(const Vector<T>& v) : _data(0), _active(0), _total(0) { CopyVector(v); }

  // destroy a vector
  ~Vector() { delete[] _data; }
 private:
  // dummy value for non-existent elements
  static T& dummy() { static T dummy; return dummy; }
  T* _data;                // vector data
  unsigned int _active;    // number of "active" elements
  unsigned int _total;     // true size of _data
};

// vector that owns it's data (i.e. data is deleted when vector is)
template <class T> class VectorOwn : public Vector<T> {
 public:
  VectorOwn() : Vector<T>() { ; }
  VectorOwn(const T& d) : Vector<T>(d) { ; }
  VectorOwn(const Vector<T>& v) : Vector<T>(v) { ; }
  ~VectorOwn() { for(unsigned int i = 0; i < Size(); i++) delete (*this)[i]; }
};

#if 0
template <class T> class VectorItr {
 public:
  VectorItr(const Vector<T>&);
  ~VectorItr();
 private:
  Vector<T>* _v;
  int _index;
};
#endif

#ifdef __GNUC__
#include "Vector.c"
#endif

#endif
