#include <stream.h>
#include <math.h>

#ifndef __GNUC__
#include "hash.h"
#define inline
#else
#include "misc.h"
#endif

// possible improvements:
// keep tail ptr for each bucket list???
// cache last lookup? (or cache hash keys?)

#define resizeThreshold		0.5

// compare two hash keys
template <class K, class D>
inline Boolean HashTable<K, D>::KeyCmp(const K& a, const K& b) const
{
  return _Compare ? _Compare(a, b) : MakeBoolean(a == b);
}

// default hash function
template <class K, class D>
inline unsigned int HashTable<K, D>::KeyGen(const K& a) const
{
  unsigned int hash_key = 0;
  if(_Compute) {
    hash_key = _Compute(a);
  } else {
    char* p = (char*)&a;
    for(int i = 0; i < sizeof(a); i++)
      hash_key += *p++;
  }
  return hash_key % _numBuckets;
}

// lookup hash bucket given a key
template <class K, class D>
inline HashBucket<K, D>* HashTable<K, D>::LookupBucket(const K& key) const
{
  unsigned int index = KeyGen(key);
  for(HashBucket<K,D>* b = _buckets[index]; b && !KeyCmp(key, b->_key); b = b->_next);
  return b;
}

// hash Add/Set/Replace
template <class K, class D>
inline Boolean HashTable<K, D>::DoHashOp(HashOp op, const K& key, const D& data)
{
  unsigned int index = KeyGen(key);
  for(HashBucket<K,D>* b = _buckets[index]; b && !KeyCmp(key, b->_key); b = b->_next);

  switch(op)
    {
    case HashAdd:
      if(b) return False;
    case HashSet:
      if(!b) {
	b = NewBucket(key, data);
	if(b->_next = _buckets[index])
	  _numCollisions++;
	_buckets[index] = b;
	//_tails[index] = b;
	if(_resizeCount < ++_numEntries)
	  Resize(True);
      } else {
	b->SetValues(key, data);
      }
      return True;
    case HashReplace:
      if(b) b->SetValues(key, data);
      return False;
    default:
      return False;
    }
}

// create a new hash bucket
template <class K, class D>
inline HashBucket<K, D>* HashTable<K, D>::NewBucket(const K& key, const D& data)
{
  return new HashBucket<K,D>(key, data);
}

// remove an element
template <class K, class D>
inline Boolean HashTable<K, D>::Remove(const K& key)
{
  unsigned int index = KeyGen(key);
  HashBucket<K,D>* b = _buckets[index];
  HashBucket<K,D>* last = 0;

  while(b && !KeyCmp(key, b->_key)) {
    last = b;
    b = b->_next;
  }

  if(b) {
    if(!last)
      _buckets[index] = b->_next;
    else
      last->_next = b->_next;
    //if(_tails[index] == b) _tails[index] = last;
    b->_next = 0;
    delete b;
    _numEntries--;
    return True;
  }
  return False;
}


// change size of hash table
// grow == False -> keep same size but rehash all data!
template <class K, class D>
inline void HashTable<K, D>::Resize(Boolean grow)
{
  if ((grow && Primes[_sizeIndex+1]) || !grow)
    {
      // join all buckets into singly linked list so they can be reused.
      HashBucket<K,D>* b = 0;
      HashBucket<K,D>* head = 0;
      for(unsigned int i = 0; i < _numBuckets; i++)
        {
	  if(_buckets[i])
	    {
	      if(!head)
		head = _buckets[i];
	      if(b)
		b->_next = _buckets[i];
	      else
		b = _buckets[i];
	      while(b->_next) b = b->_next;
	    }
        }

      _numCollisions = 0;

      // not sure if zero check is necessary
      delete[] _buckets;
      //delete[] _tails;

      _numBuckets = Primes[grow ? ++_sizeIndex : _sizeIndex];

      _buckets = new HashBucket<K,D>*[_numBuckets];
      MemClear(_buckets, sizeof(*_buckets)*_numBuckets);

      //_tails = new HashBucket<K,D>*[_numBuckets];
      //MemClear(_tails, sizeof(*_buckets)*_numBuckets);

      _resizeCount = (unsigned int)ceil(resizeThreshold*_numBuckets);

      for(b = head; b; b = head)
        {
          head = b->_next;
          unsigned int index = KeyGen(b->_key);
          if(b->_next = _buckets[index])
            _numCollisions++;
          _buckets[index] = b;
        }
    }
}

// create a new hash table
template <class K, class D>
inline HashTable<K, D>::HashTable(HashCompareParam(cmpr),
				  HashComputeParam(cmpu),
				  unsigned int isize,
				  K dummy_key,
				  D dummy_data)
: _Compute(cmpu), _Compare(cmpr),
  _buckets(0), _sizeIndex(1),
  _numBuckets(0), _numEntries(0),
  _numCollisions(0), _resizeCount(0),
  _dummy_key(dummy_key), _dummy_data(dummy_data)
{
  isize = (unsigned int)(isize/resizeThreshold);

  // find closest prime to passed size.
  for (_sizeIndex = 1; Primes[_sizeIndex]; _sizeIndex++)
    if (Primes[_sizeIndex] >= isize)
      break;

  if (Primes[_sizeIndex])
    Resize(False);
  else
    assert(!"requested hash table size too big!");
}

// destroy a hash table
template <class K, class D>
inline HashTable<K, D>::~HashTable()
{
  if(_buckets)
    for(unsigned int i = 0; i < _numBuckets; i++)
      delete _buckets[i];
  delete[] _buckets;
}

// map a function to all hash elements
template <class K, class D>
inline void HashTable<K, D>::Map(void (*fn)(HashTable<K, D>*,
					    HashBucket<K, D>*, void*),
				 void* env)
{
  HashTableItr<K,D>* hitr = new HashTableItr<K,D>(this);
  hitr->Map(fn, env);
  delete hitr;
}

// print hash table
template <class K, class D>
inline ostream& HashTable<K, D>::Print(ostream& os, int) const
{
  os << "HashTable: " << _numBuckets << ", " << _numEntries;
  os << ", " << _numCollisions << endl;
  for (unsigned int i = 0; i < _numBuckets; i++)
    for(HashBucket<K,D>* b = _buckets[i]; b; b = b->_next)
      os << "Bucket " << i << ": " << b->_key << ", " << b->_data << endl;
  return os;
}

// get next bucket for hash table iterator
template <class K, class D>
inline HashBucket<K,D>* HashTable<K, D>::NextBucket(unsigned int& bucketindex,
						    const HashBucket<K,D>* b) const
{
  if (b && b->_next)
    return b->_next;
  else
    {
      if(b && !b->_next)
	bucketindex++;

      if (_numBuckets <= bucketindex)
	return 0;

      do
	{
	  b = _buckets[bucketindex];
	}
      while(!b && ++bucketindex < _numBuckets);
      return (HashBucket<K,D>*)b;
    }
}

/////////////////////////////////////////////////////////////////////////////
//
// Hashtable Iterator
//
/////////////////////////////////////////////////////////////////////////////

// map a function to all elements in an iterator
template <class K, class D>
inline void HashTableItr<K, D>::Map(void (*fn)(HashTable<K, D>*,
					       HashBucket<K, D>*, void*),
				    void* env)
{
  while(!IsEmpty()) {
    fn((HashTable<K,D>*)_hash, (HashBucket<K,D>*)_bucket, env);
    Next();
  }
}

// advance iterator
template <class K, class D>
inline void HashTableItr<K, D>::Next()
{
  if(_bucket && _bucket->Next()) {
    _bucket = _bucket->Next();
  } else {
    _bucket = 0;
    _index++;
    while(!_bucket && _index < _hash->_numBuckets) {
      _bucket = _hash->_buckets[_index];
      if(!_bucket) _index++;
    }
  }
  //cout << "Next() -> " << _bucket << ", " << _index << endl << flush;
}
