////////////////////////////////////////////////////////////////////////////
//
// hash.h - templatized hash table
//
////////////////////////////////////////////////////////////////////////////
#ifndef _HASH_H_
#define _HASH_H_

#include "Object.h"

// table of primes for hash table sizes
extern unsigned int Primes[];

// make a ptr hash table with Boolean flags for owning

// different hash operations
enum HashOp { HashAdd, HashReplace, HashSet };

// some defines to make typing easier
#if 0
typedef Boolean (*HashCompare)(const K&, const K&);
typedef unsigned int (*HashCompute)(const K&);
typedef void (*HashMap)(HashTable<K, D>*, HashBucket<K, D>*, void*);
#else
#define HashCompare Boolean (*)(const K&, const K&)
#define HashCompute unsigned int (*)(const K&)
#define HashMap void (*)(HashTable<K, D>*, HashBucket<K, D>*, void*)

#define HashCompareParam(x) Boolean (*x)(const K&, const K&)
#define HashComputeParam(x) unsigned int (*x)(const K&)
#define HashMapParam(x) void (*x)(HashTable<K, D>*, HashBucket<K, D>*, void*)
#endif

// hash bucket
template <class K, class D> class HashBucket {
 public:
  // next hash bucket in chain
  HashBucket*& Next() const { return ((HashBucket<K,D>*)this)->_next; }

  // hash bucket key
  const K& Key() const { return _key; }

  // hash bucket data
  D& Data() const { return ((HashBucket<K,D>*)this)->_data; }

  // assign hash bucket
  const HashBucket& operator=(const HashBucket& hb)
  { _key = hb._key; _data = hb._data; return *this; }

  // new empty hash bucket
  HashBucket() : _next(0) { ; }

  // new hash bucket from key and data
  HashBucket(const K& k, const D& d) : _key(k), _data(d), _next(0) { ; }

  // set hash bucket key and data
  virtual void SetValues(const K& k, const D& d) { _key = k; _data = d; }

  // delete hash bucket
  virtual ~HashBucket() { delete _next; }
 protected:
  K _key;             // key
  D _data;            // data
  HashBucket* _next;  // pointer to next hash bucket in chain
  friend class HashTable<K, D>;
};

// hash bucket that owns it's key
template <class K, class D> class HashBucketOwnKey : public HashBucket<K, D> {
 public:
  void SetValues(const K& k, const D& d) { delete _key; _key = k; _data = d; }
  HashBucketOwnKey(const K& k, const D& d) : HashBucket<K, D>(k, d) { ; }
  ~HashBucketOwnKey() { delete _key; }
};

// hash bucket that owns it's data
template <class K, class D> class HashBucketOwnData : public HashBucket<K, D> {
 public:
  void SetValues(const K& k, const D& d) { delete _data; _key = k; _data = d; }
  HashBucketOwnData(const K& k, const D& d) : HashBucket<K, D>(k, d) { ; }
  ~HashBucketOwnData() { delete _data; }
};

// hash bucket that owns it's key and data
template <class K, class D> class HashBucketOwnKeyAndData : public HashBucket<K, D> {
 public:
  void SetValues(const K& k, const D& d) { delete _key; delete _data; _key = k; _data = d; }
  HashBucketOwnDataAndKey(const K& k, const D& d) : HashBucket<K, D>(k, d) { ; }
  ~HashBucketOwnKeyAndData() { delete _key; delete _data; }
};

// hash table class
template <class K, class D> class HashTable : public SimpleObject {
 public:
  // add a new element to a hash table (fails if it already exists)
  Boolean Add(const K& k, const D& d) { return DoHashOp(HashAdd, k, d); }

  // replace an existing element (fails if it doesn't exist)
  Boolean Replace(const K& k, const D& d) { return DoHashOp(HashReplace, k, d); }

  // set a hash element
  Boolean Set(const K& k, const D& d) { return DoHashOp(HashSet, k, d); }

  // remove a hash element (fails if element doesn't exist)
  Boolean Remove(const K&);

  // lookup a hash element by key and return it's data
  D& Lookup(const K& k) const { HashBucket<K, D>* b = LookupBucket(k); return b ? b->_data : _dummy_data; }

  // hash table membership
  Boolean IsMember(const K& k) const { return MakeBoolean(LookupBucket(k)); }

  // find bucket list (i.e. all elements that have collided on a certain key)
  const HashBucket<K, D>* LookupBucketList(const K& k) const { return _buckets[KeyGen(k)]; }

  // rehash all elements
  void Rehash() { Resize(False); }

  // map a function over all hash table elements
  void Map(HashMap, void*);

  // hash table statistics
  unsigned int NumEntries() const { return _numEntries; }
  unsigned int NumBuckets() const { return _numBuckets; }
  unsigned int NumCollisions() const { return _numCollisions; }

  // print hashtable
  ostream& Print(ostream&, int=0) const;

  // create new hash table
  HashTable(HashCompare = 0, HashCompute = 0, unsigned int = 0,
	    K dummy_key = 0, D dummy_data = 0);

  // destroy a hash table
  virtual ~HashTable();
 protected:
  // create a new bucket
  virtual HashBucket<K, D>* NewBucket(const K& key, const D& d);

  // lookup a bucket by key
  HashBucket<K, D>* LookupBucket(const K&) const;

  // get next bucket after current one (for iterators)
  HashBucket<K, D>* NextBucket(unsigned int&, const HashBucket<K, D>*) const;

  // default key compare
  Boolean KeyCmp(const K& a, const K& b) const;

  // default hash function
  unsigned int KeyGen(const K& a) const;

  // perform add/set/replace
  Boolean DoHashOp(HashOp, const K&, const D&);

  // grow or rehash the table
  void Resize(Boolean);

 private:
  unsigned int (*_Compute)(const K&);      // hash function
  Boolean (*_Compare)(const K&, const K&); // element comparison function
  HashBucket<K, D>** _buckets;             // bucket list
  unsigned short _sizeIndex;               // pointer to table size in Primes[]
  unsigned int _numBuckets;                // number of buckets
  unsigned int _numEntries;                // number of entries
  unsigned int _numCollisions;             // number of collisions
  unsigned int _resizeCount;               // number of elements before resize
  K _dummy_key;                            // dummy key when key doesn't exist
  D _dummy_data;                           // dummy data when data doesn't exist

  friend class HashTableItr<K, D>;
};

// hash table that owns it's keys
template <class K, class D> class HashTableOwnKey : public HashTable<K, D> {
 public:
  HashTableOwnKey(HashCompareParam(cmpr) = 0, HashComputeParam(cmpu) = 0,
		  unsigned int size = 0, K dkey = 0, D ddata = 0)
    : HashTable<K, D>(cmpr, cmpu, size, dkey, ddata) { ; }
 protected:
  HashBucket<K, D>* NewBucket(const K& k, const D& d) { return new HashBucketOwnKey<K,D>(k, d); }
};

// hash table that owns it's data
template <class K, class D> class HashTableOwnData : public HashTable<K, D> {
 public:
  HashTableOwnData(HashCompareParam(cmpr) = 0, HashComputeParam(cmpu) = 0,
		   unsigned int size = 0, K dkey = 0, D ddata = 0)
    : HashTable<K, D>(cmpr, cmpu, size, dkey, ddata) { ; }
 protected:
  HashBucket<K, D>* NewBucket(const K& k, const D& d) { return new HashBucketOwnData<K,D>(k, d); }
};

// hash table that owns it's keys and data
template <class K, class D> class HashTableOwnKeyAndData : public HashTable<K, D> {
 public:
  HashTableOwnKeyAndData(HashCompareParam(cmpr) = 0, HashComputeParam(cmpu) = 0,
			 unsigned int size = 0, K dkey = 0, D ddata = 0)
    : HashTable<K, D>(cmpr, cmpu, size, dkey, ddata) { ; }
 protected:
  HashBucket<K, D>* NewBucket(const K& k, const D& d) { return new HashBucketOwnKeyAndData<K,D>(k, d); }
};

// hash table iterator (add stuff like dump)
template <class K, class D> class HashTableItr {
 public:
  // advance iterator
  void Next();

  // map function to all elements in iterator
  void Map(HashMap, void* env);

  // is the iterator finished?
  Boolean IsEmpty() const { return MakeBoolean(!_bucket && _index >= _hash->_numBuckets); }

  // return key of current element
  const K& Key() const { return (!IsEmpty() && _bucket) ? _bucket->Key() : _hash->_dummy_key; }

  // return data of current element
  D& Data() const { return (!IsEmpty() && _bucket) ? _bucket->Data() : _hash->_dummy_data; }

  // create an iterator from a hash table
  HashTableItr(const HashTable<K,D>* h) : _hash(h), _bucket(0), _index(-1) { Next(); }

  // destroy iterator
  virtual ~HashTableItr() { ; }
 private:
  const HashTable<K, D>* _hash;     // hash table this iterator is working on
  const HashBucket<K, D>* _bucket;  // current bucket
  int _index;                       // current bucket list index
};

#ifdef __GNUC__
#include "hash.c"
#endif

#endif
