////////////////////////////////////////////////////////////////////////////
//
// FreeList.h - fast specialized memory allocator
//
////////////////////////////////////////////////////////////////////////////
#ifndef _FREELIST_H_
#define _FREELIST_H_

#ifdef USE_FREELIST

#include <new.h>
#include <sys/types.h>
#include "misc.h"

// internal macros
#define DataAddr(fle) ((void*)((char*)(fle) + sizeof(FreeListElem*)))
#define HeaderAddr(fle) ((FreeListElem*)((char*)(fle) - sizeof(FreeListElem*)))

// element in the free list
struct FreeListElem {
  FreeListElem* _next;
  ~FreeListElem() { delete _next; }
};

// free list does _not_ overload array allocators!
class FreeList {
 public:
  // remove an element from the free list
  void* Pop()
    {
      FreeListElem* tmp = _cache;
      _cache = tmp->_next;
      _size--;
      return DataAddr(tmp);
    }
  // return an element to the free list
  void Push(void* tmp)
    {
      FreeListElem* old = HeaderAddr(tmp);
      old->_next = _cache;
      _cache = old;
      _size++;
    }

  // allocate an element from the free list
  void* Alloc(size_t, void*);

  // free an element
  void Free(void* addr) { if(addr) { Push(addr); _num_frees++; } }

  // create a new free list
  FreeList(char*, int obj_size, Boolean = False, int = 0, int = 0, int = 0, void* = 0);

  // destroy free list
  ~FreeList();
 private:
  void Expand(int, void*);  // grow free list
  char* _class_name;        // class name this free list is for
  Boolean _verbose;         // print new/delete statistics
  int _obj_size;            // size of objects in free list
  FreeListElem* _cache;     // free list
  void* _place;             // default placement operator for new
  int _size;                // size of free list
  int _total;               // total # bytes in free list
  int _limit;               // maxmimum free list size
  int _expand;              // free list is expandable
  int _num_allocs;          // number of allocs
  int _num_frees;           // number of frees
  int _num_chunks;          // number of free list chunks
  FreeListElem* _chunks;    // list of chunks
};

// declare use of free list in class definition
#define FreeListDefaultDecls(class_name)\
 private: \
  static FreeList& freelist() \
    { static FreeList fl(#class_name, sizeof(class_name)); return fl; } \
 public: \
  void* operator new(size_t s, void* p=0) { return freelist().Alloc(s, p); } \
  void operator delete(void* p) { freelist().Free(p); } \
  friend class FreeList

// like FreeListDefaultDecls but have more control over options
#define FreeListDecls(class_name, verb, init, limit, exp, place)\
 private: \
  static FreeList& freelist() \
    { \
      static FreeList fl(#class_name, sizeof(class_name), verb, init, limit, exp, place); \
      return fl; \
    } \
 public: \
  void* operator new(size_t s, void* p=0) { return freelist().Alloc(s, p); } \
  void operator delete(void* p) { freelist().Free(p); } \
  friend class FreeList

#else
// macros still need to be defined if we aren't using the freelist
#define FreeListDefaultDecls(class_name)
#define FreeListDecls(class_name, verb, init, limit, exp, place)
#endif

#endif
