#include <stream.h>
#include <stdlib.h>
#define _TRACE_DEF_
#include "new.h"

// simple system to track news and deletes
#ifdef TRACE_NEW

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

extern int InitMemTracker(char* logfile, int doublefree);
extern void PrintMemoryStats(int dummy, int sort);

struct MemRecord;

static int DoubleFreeTrack = 1;
static FILE* LogFile = 0;
static MemRecord* Records = 0;
static int NumRecords = 0;
static int did_init = InitMemTracker(0, DoubleFreeTrack);

typedef int (*comparator)(const void*, const void*);

struct MemRecord {
  void* address;
  char* file;
  unsigned int size;
  unsigned short line;
  unsigned short freed;
};

int InitMemTracker(char* logfile, int doublefree)
{
  DoubleFreeTrack = doublefree;
  if(logfile)
    LogFile = fopen(logfile, "w");
  on_exit(PrintMemoryStats, BySize);
  return 1;
}

static void PrintMemRecord(FILE* f, MemRecord* m, int af)
{
  if(f) {
    if(af)
      fprintf(f, "%x - %d bytes allocated at line %d in file %s.\n",
	      (int)m->address, m->size, m->line, m->file);
    else
      fprintf(f, "%x - %d bytes freed at line %d in file %s.\n",
	      (int)m->address, m->size, m->line, m->file);
  }
}

/* make realloc much bigger here!!!!! */
static void AddMemRecord(void* add, unsigned int size, char* file, unsigned int line)
{
  int m = 0;
  int res = -1;

  if(Records)
    Records = (MemRecord*)realloc(Records, (NumRecords + 1) * sizeof(MemRecord));
  else
    Records = (MemRecord*)malloc(sizeof(MemRecord));

  if(NumRecords) {
    unsigned int u = NumRecords - 1;
    unsigned int l = 0;
    m = (l+u)>>1;
    res = add - Records[m].address;
    while(res && l < u) {
      if(res > 0)
        l = m + 1;
      else if(res < 0)
        u = m ? m - 1 : l++, 0;
      else
        break;
      m = (l + u) >> 1;
      res = add - Records[m].address;
    }

    if(res > 0) m++;

    if(res && NumRecords - m > 0)
      {
        /*printf("copy %d,%d to %d\n", m, (NumRecords - m), m+1);*/
        MemCopy(&Records[m], &Records[m+1], (NumRecords - m) * sizeof(MemRecord));
      }
  }

  if(res)
    NumRecords++;

  Records[m].address = add;
  Records[m].size = size;
  Records[m].file = file;
  Records[m].line = line;
  Records[m].freed = 0;

  PrintMemRecord(LogFile, &Records[m], 1);
}

static int CompareMemRecordsByAddress(char* a, MemRecord* b)
{
  return (a - (char*)b->address);
}

static MemRecord* FindMemRecord(void* add)
{
  return (MemRecord*)bsearch(add, Records, NumRecords,
			     sizeof(MemRecord),
			     (comparator)CompareMemRecordsByAddress);
}

static void RemoveMemRecord(void* add)
{
  MemRecord* m = FindMemRecord(add);

  if(m) {
    if(DoubleFreeTrack) {
      PrintMemRecord(LogFile, m, 0);
      if(m->freed) {
        if(m->freed == 1)
          printf("%dnd free of ", m->freed+1);
        else if(m->freed == 2)
          printf("%drd free of ", m->freed+1);
        else
          printf("%dth free of ", m->freed+1);
        PrintMemRecord(stderr, m, 0);
      }
      m->freed++;
    } else {
      int i = m - Records;
      if(NumRecords - i - 1 > 0)
        MemCopy(&Records[i+1], &Records[i], (NumRecords - i - 1) * sizeof(MemRecord));
      NumRecords--;
      if(!NumRecords) {
        free(Records);
        Records = 0;
      }
    }
  } else {
    printf("Memory at %x not allocated by new!\n", (int)add);
  }
}

static int sortkind = BySize;

static int CompareMemRecords(MemRecord* a, MemRecord* b)
{
  return (sortkind == BySize) ? (a->size - b->size) : strcmp(a->file, b->file);
}

void PrintMemoryStats(int dummy, int sort)
{
  int i, total = 0;
  int num = 0;

  if(LogFile)
    fclose(LogFile);

  if(Records && sort != NoSort) {
    sortkind = sort;
    qsort(Records, NumRecords, sizeof(MemRecord), (comparator)CompareMemRecords);
  }

  for(i = 0; i < NumRecords; i++) {
    MemRecord* m = &Records[i];
    if(!m->freed)
      num++;
  }

  if(num) {
    printf("\n%d Memory leaks:\n", num);
    for(i = 0; i < NumRecords; i++) {
      MemRecord* m = &Records[i];
      if(!m->freed) {
        PrintMemRecord(stderr, m, 1);
        total += m->size;
      }
    }
    printf("\nTotal bytes leaked = %d.\n", total);
  } else {
    printf("\nNo leaks detected.\n");
    if(Records) free(Records);
  }
}

void* operator new(size_t size, const char* file, int line)
{
  void* add = (void*)malloc(size);
  assert(add);
  AddMemRecord(add, size, (char*)file, line);
  return add;
}

void* operator new[](size_t size, const char* file, int line)
{
  void* add = (void*)malloc(size);
  assert(add);
  AddMemRecord(add, size, (char*)file, line);
  return add;
}

void* renew(void* add, unsigned int size, char* file, unsigned int line)
{
  RemoveMemRecord(add);
  add = (void*)realloc(add, size);
  assert(add);
  AddMemRecord(add, size, file, line);
  return add;
}

void operator delete(void* add)
{
  if(add) {
    RemoveMemRecord(add);
    free(add);
  }
}

#elif defined(OVERLOAD_NEW)

void* operator new(size_t size, void* p)
{
  //cout << "calling new(" << size << ")" << endl;
  return malloc(size);
}

void* operator new[](size_t size, void* p)
{
  //cout << "calling new[](" << size << ")" << endl;
  return malloc(size);
}

void* operator new(size_t size)
{
  //cout << "calling new(" << size << ")" << endl;
  return malloc(size);
}

void* operator new[](size_t size)
{
  //cout << "calling new[](" << size << ")" << endl;
  return malloc(size);
}

void operator delete(void* p)
{
  //cout << "calling delete" << endl;
  free(p);
}

#endif
