#include <assert.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <sys/time.h>
#include "timer.h"
#include "misc.h"

/* max size of timer description string */
#define MAX_DESC_SIZE    80

struct timer {
  struct timeval timer;     /* time this timer was last started */
  float total_time;         /* total amount of time this timer has run */
  float max_time;           /* largest time slice this timer has run */
  float min_time;           /* smallest time slice this timer has run */
  int num;                  /* number of times this timer has run */
  char desc[MAX_DESC_SIZE]; /* description of this timer */
  int is_on;                /* flag, is the timer on? */
};

/* storage for timer objects */
static struct timer timers[MAX_TIMERS];
static int num_timers = 0;

/* print info about timer i */
void timer_print(tid_t i)
{
  if(i > 0 && i <= num_timers) {
    struct timer* t = &timers[i-1];
    if(t->num)
      printf("Timer(%c) %s: Total(%d) = %.3f Ave = %.3f Min/Max = %.3f/%.3f",
	     t->is_on ? 'R' : 'S', t->desc, t->num, t->total_time/1000,
	     t->total_time/1000/t->num, t->min_time/1000, t->max_time/1000);
    else
      printf("Timer(%c) %s: Total(%d) = %.3f Ave = %.3f Min/Max = %.3f/%.3f",
	     t->is_on ? 'R' : 'S', t->desc, t->num, t->total_time/1000, 0.0,
	     t->min_time/1000, t->max_time/1000);
  }
}


/* print all timers */
void timer_print_all(void)
{
  int i;
  for(i = 0; i <= num_timers; i++) {
    timer_print(i);
    printf("\n");
  }
}

/* get the current timer delta from a running timer.
   (e.g. the current time - start time of the given timer) */
float timer_delta(tid_t i)
{
  if(i > 0 && i <= num_timers) {
    struct timer* t = &timers[i-1];
    struct timeval et;

    if(t->is_on) {
      gettimeofday(&et, 0);
      return (et.tv_sec - t->timer.tv_sec) * 1e6 + (et.tv_usec - t->timer.tv_usec);
    }
  }
  return 0.0;
}

/* return the total time recorded by the given timer */
float timer_total(tid_t i)
{
  if(i > 0 && i <= num_timers) {
    struct timer* t = &timers[i-1];
    return t->total_time;
  }
  return 0.0;
}

/* return the max time interval recorded by the given timer */
float timer_max(tid_t i)
{
  if(i > 0 && i <= num_timers) {
    struct timer* t = &timers[i-1];
    return t->max_time;
  }
  return 0.0;
}

/* return the min time interval recorded by the given timer */
float timer_min(tid_t i)
{
  if(i > 0 && i <= num_timers) {
    struct timer* t = &timers[i-1];
    return t->min_time;
  }
  return 0.0;
}

/* return the number of intervals this timer has measured */
int timer_num(tid_t i)
{
  if(i > 0 && i <= num_timers) {
    struct timer* t = &timers[i-1];
    return t->num;
  }
  return 0;
}

/* find out if the given timer is on */
int timer_is_on(tid_t i)
{
  if(i > 0 && i <= num_timers) {
    struct timer* t = &timers[i-1];
    return t->is_on;
  }
  return 0;
}

/* used to print timer reports at program exit */
static void report_time()
{
  int i;
  printf("\n");
  for(i = 1; i <= num_timers; i++) {
    if(timer_is_on(i))
      printf("Timer %s left running!\n", timers[i-1].desc);
    timer_print(i);
    printf("\n");
  }
  printf("\n");
}

/* start the given timer */
void timer_start(tid_t ti)
{
  if(ti > 0 && ti <= num_timers) {
    struct timer* t = &timers[ti-1];
    if(!t->is_on) {
      gettimeofday(&t->timer, 0);
      t->is_on = 1;
    } else {
      printf("Warning: starting timer %s multiple times\n", t->desc);
    }
  }
}

/* stop the given timer */
void timer_stop(tid_t ti)
{
  if(ti > 0 && ti <= num_timers) {
    struct timer* t = &timers[ti-1];
    struct timeval et;
    if(t->is_on) {
      float total;
      gettimeofday(&et, 0);
      total = (et.tv_sec - t->timer.tv_sec) * 1e6 + (et.tv_usec - t->timer.tv_usec);
      t->total_time += total;
      if(total > t->max_time)
	t->max_time = total;
      if(total < t->min_time)
	t->min_time = total;
      t->num++;
      t->is_on = 0;
    } else {
      printf("Warning: stopping timer %s multiple times\n", t->desc);
    }
  }
}

/* create a new timer with the given description */
tid_t timer_create(const char* desc)
{
  static int init = 0;
  tid_t ct = ++num_timers;

#if 0
  if(!init) {
    on_exit(report_time, 0);
    init = 1;
  }
#endif

  if(num_timers < MAX_TIMERS) {
    timers[ct-1].total_time = 0.0;
    timers[ct-1].num = 0;
    assert(strlen(desc) < MAX_DESC_SIZE);
    strcpy(timers[ct-1].desc, desc);
    timers[ct-1].is_on = 0;
    timers[ct-1].max_time = 0;
    timers[ct-1].min_time = 1e60;
    return ct;
  } else {
    --num_timers;
    return 0;
  }
}
