////////////////////////////////////////////////////////////////////////////
//
// dxmodules.c - interface file for all dx extensions
//
////////////////////////////////////////////////////////////////////////////
#include "dxfield.h"
#include "planet.h"
#include "ocean.h"
#include "terrain.h"
#include "dxcolor.h"
#include "cloud.h"
#include "array.h"

// must set defaults in case input objects are nil... stupid..

extern "C" {

Error m_Planet(DXObject *in, DXObject *out)
{
  Real rad = 100.0, delta = 0.0, cap = 0.0, min, max;
  int iter = 500, seed = 0;
  DXField f;

#ifdef DEBUG
  cout << "m_Planet:" << endl;
#endif

  if(!dxscanf(in, "%rF%f%d%f%d%f", &f, &rad, &iter, &delta, &seed, &cap))
    return ERROR;

  Real rdelta = delta * (rad * 0.05) / iter;
  Planet p(f, rad, iter, rdelta, 0.0, seed, cap, False, False);

#ifdef DEBUG
  f.Print(cout, 1); cout << endl << flush;
#endif

  f.MinMaxHeights(min, max);
  out[0] = f;
  out[1] = ToDX(min);
  out[2] = ToDX(max);

  return OK;
}

// should add option for "planar" oceans
// ocean should be a separate field which can be combined with the original
// field, should also have opacities set according to depth.
// should also recolor original field so the colors below sea level are
// some "dark" color
Error m_Ocean(DXObject *in, DXObject *out)
{
  Real level = 0.5, shape = 0.08, t = 10.0, V = 10, ang = 0.0;
  Real K0 = 80.0, Kx = 80.0, Kz = 80.0, g = 9.81;
  Real min, max, omin, omax;
  Real sea_level;
  char* kind = 0;
  DXField shore;
  int land = 0;
  int i;

#ifdef DEBUG
  cout << "m_Ocean:" << endl;
#endif

  if(!dxscanf(in, "%rF%s%f%f%f%f%f%f%f%f%f", &shore, &kind,
	      &level, &shape, &t, &V, &ang, &K0, &Kx, &Kz, &g))
    return ERROR;

  DXField ocean(shore, True, True);
  shore.MinMaxHeights(min, max);
  sea_level = ((max - min) * level) + min;
  ocean.SetHeight(sea_level);

#ifdef DEBUG
  cout << "Shore: " << min << "/" << max << "/" << shore.AverageHeight() << endl;
#endif

  if(!strcmp(kind, "flat")) {
    // do nothing
  } else if(!strcmp(kind, "noise")) {
    int expo = (int)ceil(log(max(shore.Width(), shore.Height()))/log(2));
    int size = (int)pow(2, expo);
    MyField tmp(size, size);

    tmp.SpectralSynthesis2d(1.0 + shape);
    tmp *= 0.8;

    for(int y = 0; y < ocean.Height(); y++)
      for(int x = 0; x < ocean.Width(); x++)
	ocean.Height(x, y) = tmp.Height(x, y) + sea_level;

  } else if(!strcmp(kind, "simple")) {
    Real scale = (shape / 5.0) * (max - min);
    Real scalef = 2.5 * scale;
    for(int y = 0; y < ocean.Height(); y++) {
      Real phi = (M_PI * y) / 20.0;
      for(int x = 0; x < ocean.Width(); x++) {
	Real rx = M_PI * (rrand(2.0) - 1.0) / 4.0;
	Real ry = M_PI * (rrand(2.0) - 1.0) / 4.0;
	Real theta = (M_PI * x) / 15.0;
	Real& h = ocean.Height(x, y);
	Real d = (scalef * (h - shore.Height(x, y))/sea_level) + scale;
	h += d * (fsin(theta+phi+rx) * fcos(phi+ry));
      }
    }
  } else if(!strcmp(kind, "realistic")) {
    Ocean o(shore, V, shape, level, K0, Kx, Kz, g);
#ifdef DEBUG
    cout << o << endl;
#endif
    MyField* tmp = o.Compute(t, deg2rad(ang), level, False);
    ocean.CopyData(*tmp);
    delete tmp;
  }

  // set opacticies (should use some sort of power fn for opacities)
  FloatArray opac(ocean.Size());
  ocean.InitRGB();
  ocean.MinMaxHeights(omin, omax);
  for(i = 0; i < shore.Size(); i++) {
    Real oh = ocean.Height(i);
    Real depth = oh - shore.Height(i);
    if(depth >= 0.0) {
      if(omin == omax)
	ocean.color(i) = RGB(0.0, 0.0, 0.95);
      else
	ocean.color(i) = RGB(0.0, 0.0, 0.95 - (0.10 * (oh-omin)/(omax-omin)));
      opac[i] = 0.6 + depth / (sea_level - min);
      if(opac[i] > 1.0) opac[i] = 1.0;
      if(opac[i] < 0.0) opac[i] = 0.0;
      land++;
    } else {
      ocean.color(i) = RGB(0.0, 0.0, 0.0);
      ocean.Height(i) = sea_level;
      opac[i] = 0.0;
    }
  }
#ifdef DEBUG
  cout << "ocean/land = " << land << "/" << shore.Size() << endl;
#endif
  Transpose(opac.Data(), ocean.Width(), ocean.Height(), 1);
  ocean.SetComponent("opacities", opac, True);

  ocean.MinMaxHeights(min, max);
  out[0] = ocean;
  out[1] = ToDX(min);
  out[2] = ToDX(max);

  return OK;
}

Error m_Water(DXObject *in, DXObject *out)
{
  Real sea = 0.5, stop = 0.01, evap = 1.0, rain = 0.05, thresh = 0.05;
  Real min, max;
  int freq = -1, imax = 0;
  DXField f;

#ifdef DEBUG
  cout << "m_Water:" << endl;
#endif

  if(!dxscanf(in, "%F%f%f%f%f%d%f%d", &f, &sea, &stop,
	      &evap, &rain, &freq, &thresh, &imax))
    return ERROR;

  FloatArray opac(f.Size());
  MyField* water = f.Water(sea, stop, evap, rain, freq, thresh, opac.Data(), imax);
  f.CopyData(*water);
  Transpose(opac.Data(), f.Width(), f.Height(), 1);
  f.SetComponent("opacities", opac);
  delete water;

  f.MinMaxHeights(min, max);
  out[0] = f;
  out[1] = ToDX(min);
  out[2] = ToDX(max);

  return OK;
}

// not sure if this guy should take ambient light as parameter
Error m_Cloud(DXObject *in, DXObject *out)
{
  Real shine = 0.1, Ka = 0.8, Kd = 0.8, Ks = 0.2;
  Real c0 = 1.1, fx0 = 0.015, fy0 = 0.040, t0 = 5.5;
  Real Kt = 0.8, D = 0.6, T1 = 0.4, T2 = 0.2;
  RGBColor rgb = { 0.2, 0.2, 0.2 };
  Point3d P, Q, R, T;
  Point fp, tp, loc;
  DXField cloudf;
  DXLight tmpl;
  Camera c;
  int n = 7;

#ifdef DEBUG
  cout << "m_Cloud:" << endl;
#endif

  if(!dxscanf(in, "%rF%P%P%P%P%rC%rL%f%f%f%f%f%f%f%f%d%f%f%f%f",
	      &cloudf, &P, &Q, &R, &T, &c, &tmpl, &shine, &Ka, &Kd,
	      &Ks, &Kt, &D, &T1, &T2, &n, &c0, &fx0, &fy0, &t0))
    return ERROR;

  if(!DXQueryDistantLight(tmpl, &loc, &rgb)) {
    DXSetError(ERROR_BAD_PARAMETER, "light should be a distant light");
    return ERROR;
  }

  Light light(RGB(rgb.r, rgb.g, rgb.b), Vector3d(loc.x, loc.y, loc.z));

  DXGetView(c, &fp, &tp, 0);
  Point3d from(fp.x, fp.y, fp.z), to(tp.x, tp.y, tp.z);

  // get distance to image plane from camera! this is a hack
  tp.x -= fp.x; tp.y -= fp.y; tp.z -= fp.z;
  Real focal = sqrt(tp.x*tp.x + tp.y*tp.y + tp.z*tp.z);
  //Real focal = 1000.0;
  //

#ifdef DEBUG
  cout << "Light: "; light.Print(cout); cout << endl << flush;
  cout << "Camera: from: " << from << endl << flush;
  cout << "Camera: to: " << to << endl << flush;
  cout << "Camera: focal: " << focal << endl << flush;
#endif

  // not sure if we Ks & Kd are being factored too many
  // times into the lighting equation (e.g. once by World and once by Sky...)
  // not sure if DX Image module is also redoing the same stuff!
  // Options module controls rendering options
#if 0
  World w(light, from, focal, Ka, Kd, Ks);
#else
  World w(light, from, focal, 1.0, 1.0, 1.0);
#endif

  Sky sky(T1, T2, D, Ka, Kt, Ks, n, c0, fx0, fy0, t0);
  FloatArray opac(cloudf.Size());
  Boolean planar = True;
  Cloud* cloud = 0;

  if(!in[1]) {
    Plane pl(cloudf(0), cloudf(cloudf.Width()-1), cloudf(0,cloudf.Height()-1));
    cloud = new Cloud2d(pl, shine);
  } else {
    // for now control points are parameters to module
#if 0
    //
    // determine control points
    //
    MyField tmp(cloudf, True, True);

    Point3d P, Pmin, Pmax, Q, Qmin, Qmax, R, Rmin, Rmax;
    tmp.MinMaxDim(0, Pmin, Pmax);
    tmp.MinMaxDim(1, Qmin, Qmax);
    tmp.MinMaxDim(2, Rmin, Rmax);

    Point3d T((Pmax.x + Pmin.x)/2.0,
	      (Qmax.y + Qmin.y)/2.0,
	      (Rmax.z + Rmin.z)/2.0);
    tmp -= T;

    while(Pmax.y != Pmin.y && Pmax.z != Pmin.z) {

      // rotate so Pmax.y==Pmin.y && Pmax.z==Pmin.z

      tmp.MinMaxDim(0, Pmin, Pmax);
    }
#endif
    Ellipsoid el(P, Q, R, T);
    cloud = new Cloud3d(el, shine);
  }

#ifdef DEBUG
  cout << *cloud << endl << flush;
#endif

  float* opacd = opac.Data();
  sky.Texturize(w, *cloud, cloudf, opacd);
  Transpose(opacd, cloudf.Width(), cloudf.Height(), 1);
  cloudf.SetComponent("opacities", opac);
  delete cloud;

  out[0] = cloudf;

  return OK;
}

Error m_TwoDfbm(DXObject *in, DXObject *out)
{
  Real sigma = 1.0, H = 0.8, min, max;
  Boolean flag = True;
  int level = -1;
  DXField f;

#ifdef DEBUG
  cout << "m_TwoDfbm:" << endl;
#endif

  if(!dxscanf(in, "%rF%d%f%f%d", &f, &level, &sigma, &H, &flag))
    return ERROR;

  int l2 = (int)round(log(f.Width())/log(2));
  if(level > l2 || level < 0)
    level = l2;

  if((1 << l2) != f.Width() || (1 << l2) != f.Height()) {
    DXSetError(ERROR_BAD_PARAMETER, "Field size must be power of 2");
    return ERROR;
  }

  f.TwoDfbm(level, sigma, H, flag);
  f.ScaleHeights((f.Height() + f.Width())/20.0);

  f.MinMaxHeights(min, max);
  out[0] = f;
  out[1] = ToDX(min);
  out[2] = ToDX(max);

  return OK;
}


Error m_SpectralSynthesis(DXObject *in, DXObject *out)
{
  Real H = 0.8, min, max;
  DXField f;

#ifdef DEBUG
  cout << "m_SpectalSynthesis:" << endl;
#endif

  if(!dxscanf(in, "%rF%f", &f, &H))
    return ERROR;

  int l2 = (int)round(log(f.Width())/log(2));
  if((1 << l2) != f.Width() || (1 << l2) != f.Height()) {
    DXSetError(ERROR_BAD_PARAMETER, "Field size must be power of 2");
    return ERROR;
  }

  f.SpectralSynthesis2d(H);

  f.MinMaxHeights(min, max);
  out[0] = f;
  out[1] = ToDX(min);
  out[2] = ToDX(max);

  return OK;
}


Error m_Crater(DXObject *in, DXObject *out)
{
  Real dp = 1.5, db = sqrt(0.5), slope = 0.9, min, max;
  int iter = 50000;
  DXField f;

#ifdef DEBUG
  cout << "m_Crater:" << endl;
#endif

  if(!dxscanf(in, "%rF%d%f%f%f", &f, &iter, &dp, &db, &slope))
    return ERROR;

  f.Craterize(iter, dp, db, slope);
  f.ScaleHeights((f.Height() + f.Width())/30.0);

  f.MinMaxHeights(min, max);
  out[0] = f;
  out[1] = ToDX(min);
  out[2] = ToDX(max);

  return OK;
}

Error m_Raise(DXObject *in, DXObject *out)
{
  char* kind = "sphere";
  int iter = 500;
  Real min, max;
  DXField f;

#ifdef DEBUG
  cout << "m_Raise:" << endl;
#endif

  if(!dxscanf(in, "%rF%s%d", &f, &kind, &iter))
    return ERROR;

  if(strcmp(kind, "sphere") && strcmp(kind, "pyramid") && strcmp(kind, "cone")) {
    DXSetError(ERROR_BAD_PARAMETER, "unknown raise shape %s", kind);
    return ERROR;
  }

  f.Raise(iter, kind);

  f.MinMaxHeights(min, max);
  out[0] = f;
  out[1] = ToDX(min);
  out[2] = ToDX(max);

  return OK;
}


Error m_ThermalErode(DXObject *in, DXObject *out)
{
  Real talus = 45.0, perc = 0.05, wind = 0.0, dir = 0.0, min, max;
  int iter = 100;
  DXField f;

#ifdef DEBUG
  cout << "m_ThermalErode:" << endl;
#endif

  if(!dxscanf(in, "%rF%f%f%d%f%f", &f, &talus, &perc, &iter, &wind, &dir))
    return ERROR;

  f.T_Erode(talus, perc, iter, wind, deg2rad(dir));

  f.MinMaxHeights(min, max);
  out[0] = f;
  out[1] = ToDX(min);
  out[2] = ToDX(max);

  return OK;
}


Error m_HydraulicErode(DXObject *in, DXObject *out)
{
  Real kc = 3.0, kd = 0.1, ks = 0.3, rain = 0.001, freq = 0.93, sea = 0.0;
  int iter = 500;
  Real min, max;
  DXField f;

#ifdef DEBUG
  cout << "m_HydraulicErode:" << endl;
#endif

  if(!dxscanf(in, "%rF%f%f%f%f%f%d%f", &f, &kc, &kd, &ks, &rain, &freq, &iter, &sea))
    return ERROR;

  f.H_Erode(kc, kd, ks, rain, iter, (int)((1.0 - freq) * iter), sea);

  f.MinMaxHeights(min, max);
  out[0] = f;
  out[1] = ToDX(min);
  out[2] = ToDX(max);

  return OK;
}


Error m_Erode(DXObject *in, DXObject *out)
{
  Real talus = 45.0, perc = 0.05, wind = 0.0, dir = 0.0;
  Real kc = 3.0, kd = 0.1, ks = 0.3, rain = 0.001, freq = 0.93, sea = 0.0;
  int iter = 500;
  Real min, max;
  DXField f;

  if(!dxscanf(in, "%rF%d%f%f%f%f%f%f%f%f%f%f", &f, &iter, &talus, &perc,
	      &wind, &dir, &kc, &kd, &ks, &rain, &freq, &sea))
    return ERROR;

  int rfreq = (int)((1.0 - freq) * iter);
  f.Erode(iter, talus, perc, wind, deg2rad(dir), kc, kd, ks, rain, rfreq, sea);

  f.MinMaxHeights(min, max);
  out[0] = f;
  out[1] = ToDX(min);
  out[2] = ToDX(max);

  return OK;
}



Error m_TerrainColormap(DXObject *in, DXObject *out)
{
  Real sea = 0.5, beach = 0.05, plains = 0.30, hills = 0.10, mnt = 0.05;
  Real min = 0.0, max = 1.0, dark = 0.0;
  Field cmap;

#ifdef DEBUG
  cout << "m_TerrainColormap:" << endl;
#endif

  if(!dxscanf(in, "%rO%f%f%f%f%f%f%f%f", &cmap, &min, &max, &sea,
	      &beach, &plains, &hills, &mnt, &dark))
    return ERROR;

  TerrainPalette tp(sea, beach, plains, hills, mnt);
  if(dark != 0.0) tp.Darken(dark);
  ColorMap tcmap(cmap, tp, min, max);
  out[0] = tcmap;

  return OK;
}


Error m_TerrainColorize(DXObject *in, DXObject *out)
{
  Real sea = 0.5, beach = 0.05, plains = 0.30, hills = 0.10, mnt = 0.05;
  Real min = 0.0, max = 1.0, dark = 0.0;
  DXField f;

#ifdef DEBUG
  cout << "m_TerrainColorize:" << endl;
#endif

  if(!dxscanf(in, "%rF%f%f%f%f%f%f", &f, &sea, &beach, &plains, &hills, &mnt, &dark))
    return ERROR;

  TerrainPalette tp(sea, beach, plains, hills, mnt);
  if(dark != 0.0) tp.Darken(dark);
  f.Colorize(tp, min, max);

  out[0] = f;

  return OK;
}


Error m_LoadPNM(DXObject *in, DXObject *out)
{
  char* name = 0;

#ifdef DEBUG
  cout << "m_LoadPNM:" << endl;
#endif

  if(!dxscanf(in, "%rs", &name))
    return ERROR;

  MyField mf(name, 0);
  DXField dxf(mf);
  out[0] = dxf;

  return OK;
}


Error m_SavePNM(DXObject *in, DXObject *out)
{
  char* file = "test.ppm";
  DXField f;

#ifdef DEBUG
  cout << "m_SavePNM:" << endl;
#endif

  if(!dxscanf(in, "%rF%rs", &f, &file))
    return ERROR;

  if(!f.IsColored()) {
#ifdef DEBUG
    cout << "Field not colored!  doing auto color" << endl << flush;
#endif
    f.IntensityMap();
  }

  f.Save(file);
  return OK;
}

};

// overload new and delete to use DX memory allocators

#if 0
void* operator new(size_t size, void* p)
{
  return DXAllocateZero(size);
}

void* operator new[](size_t size, void* p)
{
  return DXAllocateZero(size);
}

void* operator new(size_t size)
{
  return DXAllocateZero(size);
}

void* operator new[](size_t size)
{
  return DXAllocateZero(size);
}

void operator delete(void* p)
{
  DXFree(p);
}
#endif
