#include "cloud.h"
#include "hmap.h"

// apply cloud texture to the given height map.
// World w contains the light source and viewing vector
// Cloud c contains the cloud to be rendered
// opac should point to an array that holds the opacities for each point
// store values in color field!
// field coords must be normalized to be in [0, 1]!
void Sky::Texturize(const World& w, const Cloud& c, MyField& f, Real* opac) const
{
  PercCounter pc("texturize", f.Size()*2);
  Real* tv = new Real[f.Size()];
  Real TR, mint = 0.0, maxt = 0.0;
  RGB I;

  f.InitRGB();

  // first compute the texture value at each point
  for(int i = 0; i < f.Size(); i++, pc++) {
    Real v = TextureVal(c, f(i));
    if(!i || mint > v) mint = v;
    if(!i || maxt < v) maxt = v;
    tv[i] = v;
  }

  // "color" f by computing cloud intensity and opacity at all points
  // texture values must be normalized for this part
  for(i = 0; i < f.Size(); i++, pc++) {
    Real ntv = (tv[i] - mint) / (maxt - mint);
    Render(w, c, f(i), ntv, I, TR);
    f.color(i) = I;
    opac[i] = 1.0 - TR;
  }

  delete[] tv;
}

// compute the texture value for c at the given point
Real Sky::TextureVal(const Cloud& c, const Point3d& p) const
{
  // 3d clouds use ellipsoidal texture and 2d clouds use planar texture
  if(c.IsA(Cloud3d::Id()))
    return _et(p.x, p.y, p.z);
  else
    return _pt(p.x, p.y, p.z);
}

// compute cloud intensity given Id, Is and It
RGB Sky::Intensity(const Cloud& c, const RGB& Id, const RGB& Is, Real It) const
{
  Real I1r = (1 - _s) * Id.r() + _s * Is.r();
  Real I2r = (1 - _t) * I1r + _t * It * c.color().r();
  Real Ir = (1 - _a) * I2r + _a;

  Real I1g = (1 - _s) * Id.g() + _s * Is.g();
  Real I2g = (1 - _t) * I1g + _t * It * c.color().g();
  Real Ig = (1 - _a) * I2g + _a;

  Real I1b = (1 - _s) * Id.b() + _s * Is.b();
  Real I2b = (1 - _t) * I1b + _t * It * c.color().b();
  Real Ib = (1 - _a) * I2b + _a;

  return RGB(Ir, Ig, Ib);
}


// compute translucence at given point
// making D, T1 & T2 really small will negate effects of translucence
Real Sky::Translucence(const World& w, const Cloud& c, Real It, const Point3d& p) const
{
  Real TR;
  if(c.IsA(Cloud3d::Id())) {
    Real g = ((Cloud3d&)c).g(p.x, p.y, w.F());
    // g is not in range when on wrong side of ellipse
    //assert(g >= 0.0 && g <= 1.0);
    TR = 1.0 - (It - _T1 - (_T2 - _T1) * (1.0 - g)) / _D;
    TR = fixrrng(TR, 0.0, 1.0);
  } else {
    Real sum = _T1 + _D;
    if(It >= sum)
      TR = 0.0;
    else if(sum > It && It >= _T1)
      TR = 1.0 - (It - _T1) / _D;
    else
      TR = 1.0;
  }
//  assert(TR >= 0.0 && TR <= 1.0);
  return TR;
}

// compute translucence and intensity at given position
void Sky::Render(const World& w, const Cloud& cloud, const Point3d& pos,
		 Real It, RGB& I, Real& TR) const
{
  Vector3d N = cloud.Normal(pos);
  RGB Id = w.Id(N, cloud.color());
  RGB Is = w.Is(N, cloud.color(), cloud.Shininess());
  I = Intensity(cloud, Id, Is, It);
  TR = Translucence(w, cloud, It, pos);
}


// create a new sky
// these defaults are totally bogus
Sky::Sky(Real T1, Real T2, Real D, Real a, Real t, Real s,
	 int n, Real C0, Real fx0, Real fy0, Real t0)
: _T1(T1), _T2(T2), _D(D), _a(a), _t(t), _s(s),
  _pt(n, C0, fx0, fy0, t0), _et(n, C0, fx0, fy0, t0)
{
}

// destroy a sky
Sky::~Sky()
{
}

// print a cloud
ostream& Cloud::Print(ostream& os, int) const
{
  return Object::Print(os);
}

ostream& Cloud3d::Print(ostream& os, int) const
{
  Cloud::Print(os);
  os << "(" << _e << ")";
  return os;
}

// compute 3d cloud normal
Vector3d Cloud3d::Normal(const Point3d& p) const
{
  return _e.Normal(p);
}

// compute 2d cloud normal
Vector3d Cloud2d::Normal(const Point3d&) const
{
  return _p.Normal();
}

ostream& Cloud2d::Print(ostream& os, int) const
{
  Cloud::Print(os);
  os << "(" << _p << ")";
  return os;
}
