////////////////////////////////////////////////////////////////////////////
//
// cloud.h - create & render planar and ellipsoidal clouds
//           (see Gardner paper for more info)
//
////////////////////////////////////////////////////////////////////////////
#ifndef _CLOUD_H_
#define _CLOUD_H_

#include "light.h"
#include "world.h"
#include "texture.h"
#include "quadric.h"
#include "plane.h"

class Cloud;
class MyField;

// Sky class - controls rendering of clouds
// and holds parameters used for rendering
class Sky {
 public:
  // given a world a cloud a field and an array of reals
  // apply the cloud texture to the given field and compute
  // the translucence values for every location and store
  // them in the given array
  void Texturize(const World&, const Cloud&, MyField&, Real*) const;

  // compute the texture value for the given cloud at the given point
  Real TextureVal(const Cloud& c, const Point3d& p) const;

  // compute intensity given the diffuse, specular and texture intensities
  RGB Intensity(const Cloud&, const RGB& Id, const RGB& Is, Real It) const;

  // compute the translucence at a given point
  // for the given cloud and texture value
  Real Translucence(const World&, const Cloud& c, Real It, const Point3d&) const;

  // compute intensity and translucence values
  // for the given cloud at a given point
  void Render(const World&, const Cloud&, const Point3d&, Real, RGB&, Real&) const;

  // Create a new "sky"
  // (Note: these defaults are pretty bogus)
  Sky(Real T1 = 0.8, Real T2 = 0.6, Real D = 0.8,
      Real a = 0.5, Real t = 0.8, Real s = 0.3,
      int n = 7, Real C0 = 0.5, Real fx0 = 0.5,
      Real fy0 = 0.5, Real t0 = 1.0);

  // destroy a sky
  ~Sky();
 private:
  // _T1, _T2, _D should have "high" values
  Real _T1; // translucence threshold
  Real _T2; // translucence threshold
  Real _D;  // range of texture vals which give translucence
  Real _a;  // fraction of ambient reflection
  Real _t;  // fraction of texture shading
  Real _s;  // fraction of specular reflection
  PlanarTexture _pt;       // texture used for planar clouds
  EllipsoidalTexture _et;  // texture used for ellipsoidal clouds
};

// abstract cloud class
class Cloud : public Object {
  DeclareStandardMethods(Cloud, Object);
 public:
  // return the shininess of this cloud
  Real Shininess() const { return _e; }

  // return the normal vector of this cloud at the given point
  virtual Vector3d Normal(const Point3d&) const = 0;

  // get the color of this cloud
  RGB& color() const { return ((Cloud*)this)->_c; }

  // destroy a cloud
  ~Cloud() { }
 protected:
  // create a white cloud
  Cloud(Real e = 0.1) : _c(RGB(1.0, 1.0, 1.0)), _e(e) { }

  // create cloud of an arbitrary color
  Cloud(const RGB& c, Real e = 0.1) : _c(c), _e(e) { }
 private:
  RGB _c;   // cloud color
  Real _e;  // cloud shininess
};

// 2d (planar) cloud
class Cloud2d : public Cloud {
  DeclareStandardMethods(Cloud2d, Cloud);
 public:
  Vector3d Normal(const Point3d&) const;
  Real TextureVal(const Texture&, const Point3d&) const;
  Cloud2d(const Plane& p, Real e = 0.1) : Cloud(e), _p(p) { }
  ~Cloud2d() { }
 private:
  Plane _p;  // plane that defines this cloud
};

// 3d (ellipsoidal) cloud
class Cloud3d : public Cloud {
  DeclareStandardMethods(Cloud3d, Cloud);
 public:
  Vector3d Normal(const Point3d&) const;
  Real TextureVal(const Texture&, const Point3d&) const;
  Real g(Real x, Real y, Real f) const
    { return _e.LimbXY(x, y, f)/_e.MaxLimbXY(x, y, f); }
  Cloud3d(const Ellipsoid& ellip, Real e = 0.1) : Cloud(e), _e(ellip) { }
  ~Cloud3d() { }
 private:
  Ellipsoid _e;  // ellipsoid that defines this cloud
};

#endif
