// terrain erosion - see Musgrave paper for more info
#include "trig.h"
#include "hmap.h"

class Tdiff {
 public:
  Tdiff() : adiff(0.0), naptr(0) { }
  Tdiff(int ok, Real alt, Real* cp, Real* np)
    {
      naptr = 0;
      if(ok && (adiff = alt - *cp - T) > 0.0) {
	naptr = np;
        total += adiff;
        maxad = max(maxad, adiff);
      }
    }
  Real adiff;
  Real* naptr;
  static Real total;
  static Real maxad;
  static Real T;
};

Real Tdiff::total = 0.0;
Real Tdiff::maxad = 0.0;
Real Tdiff::T = 0.0;

// thermal erosion.  the talus angle is the angle needed between adjacent
// elements for erosion to take place (e.g. slope of sides).  perc is the
// amount of material that may be eroded from any vertex.  iter is the
// number of iterations to perform
void MyField::T_Erode(Real talus, Real perc, int iter, Real wind, Real ang)
{
  Real dx = fabs((*this)(1).x - (*this)(0).x);
  Real winds[8] = { 0.0, 0.0, 0.0,  0.0, 0.0,  0.0, 0.0, 0.0 };
  Real T = atan(deg2rad(talus)) * dx;   // convert talus angle to simple #
  Real* curr_a = Heights();             // get heights
  Real* new_a = Heights();              // tmp heights array
  Real bsum = 0.0, asum = 0.0;
  Tdiff diffs[8];
  int w = Width();
  int h = Height();
  int wadj = w - 2;
  int i;

#ifdef DEBUG
  cout << "Talus angle: " << T << endl;
#endif

  Real xwind = wind * fcos(ang);
  Real ywind = wind * fsin(ang);

  winds[0] += ywind*fcos(deg2rad(45.0));
  winds[1] += ywind;
  winds[2] += ywind*fcos(deg2rad(45.0));
  winds[5] -= ywind*fcos(deg2rad(45.0));
  winds[6] -= ywind;
  winds[7] -= ywind*fcos(deg2rad(45.0));

  winds[2] -= xwind*fcos(deg2rad(45.0));
  winds[4] -= xwind;
  winds[7] -= xwind*fcos(deg2rad(45.0));
  winds[0] += xwind*fcos(deg2rad(45.0));
  winds[3] += xwind;
  winds[5] += xwind*fcos(deg2rad(45.0));

#ifdef DEBUG
  cout << "winds: ";
#endif
  for(i = 0; i < 8; i++) {
    if(winds[i] < 0.0)
      winds[i] = 1.0 / (1.0 - winds[i]);
    else
      winds[i] += 1.0;
#ifdef DEBUG
    cout << winds[i] << " ";
#endif
  }
#ifdef DEBUG
  cout << endl;
#endif

  for(i = 0; i < Size(); i++)
    bsum += curr_a[i];

  Tdiff::T = T;
  PercCounter pc("terode", iter);
  for(i = 0; i < iter; i++, pc++) {
    Real* altp = curr_a;
    Real* naltp = new_a;

    for(int y = 0; y < h; y++) {
      Boolean up = MakeBoolean(y > 0);
      Boolean down = MakeBoolean(y < h - 1);

      for(int x = 0; x < w; x++, altp++, naltp++) {
	Boolean left = MakeBoolean(x > 0);
	Boolean right = MakeBoolean(x < w - 1);
	Real* captr = altp - w - 1;
	Real* naptr = naltp - w - 1;
	Tdiff* diffp = diffs;
	Real alt = *altp;

        Tdiff::total = 0.0;
        Tdiff::maxad = 0.0;

	// for each vertex, find out if any of the neighboring vertices
	// exceed the talus angle.  if so drop a percentage of "rubble"
	// from this vertex onto the adjacent vertex.  The amount of
	// rubble is determined by the difference in height the talus
	// angle and the "perc" factor

	*diffp++ = Tdiff(up&&left, alt, captr, naptr); captr++; naptr++;
	*diffp++ = Tdiff(up, alt, captr, naptr); captr++; naptr++;
	*diffp++ = Tdiff(up&&right, alt, captr, naptr); captr+=wadj; naptr+=wadj;

	*diffp++ = Tdiff(left, alt, captr, naptr); captr+=2; naptr+=2;
	*diffp++ = Tdiff(right, alt, captr, naptr); captr+=wadj; naptr+=wadj;

	*diffp++ = Tdiff(down&&left, alt, captr, naptr); captr++; naptr++;
	*diffp++ = Tdiff(down, alt, captr, naptr); captr++; naptr++;
	*diffp = Tdiff(down&&right, alt, captr, naptr);

	//Real dump = perc * Tdiff::maxad;

	// add option so rubble is not prop to T but the height diff
	if(Tdiff::total > 0.0) {
	  Real* wp = winds;
	  for(; diffp >= diffs; diffp--, wp++) {
	    if(diffp->naptr) {
	      Real rperc = min(0.4, *wp * perc);
	      Real rubble = rperc * diffp->adiff;
	      *diffp->naptr += rubble;
	      *naltp -= rubble;
	    }
	  }
	}
      }
    }
    MemCopy(new_a, curr_a, w*h*sizeof(*new_a));
  }
  SetHeights(curr_a);

  for(i = 0; i < Size(); i++)
    asum += curr_a[i];

#ifdef DEBUG
  cout << "altitudes, before/after: " << bsum << "/" << asum << endl;
#endif

  delete[] curr_a;
  delete[] new_a;
}

// structure that holds the altitude, water amount
// and sediment amount for each vertex
class AWS {
 public:
#if 1
  void Rain(double rain) { water += alt * rain; }
#else
  void Rain(double rain) { water += (sed + alt) * rain; }
#endif
  AWS() : alt(0.0), water(0.0), sed(0.0) { }
  double alt, water, sed;
};

class DW {
 public:
  DW(AWS* cp=0, AWS* np=0, double ad=0.0) : cptr(cp), nptr(np), adiff(ad) { }
  AWS* cptr;
  AWS* nptr;
  double adiff;
};

class DWP {
 public:
  DW* operator[](int i) { return dwp + i; }
  void Compute(int ok, double curr_a, double curr_w, int adj)
    {
      if(ok) {
	double diff = (curr_a + curr_w) - (cptr->alt + cptr->water);
	if(diff > 0.0) {
	  dwp[n++] = DW(cptr, nptr, diff);
	  adsum += diff;
	  maxad = max(maxad, diff);
	  // may be too expensive to compute for all cases!
	  // compute only when needed?
	}
      }
      cptr+=adj; nptr+=adj;
    }
  DWP(AWS* cp, AWS* np)
    : cptr(cp), nptr(np), maxad(0.0), adsum(0.0), n(0) { }
  DW dwp[8];   // dw info array
  AWS* cptr;   // pointer to current adjacent vertex
  AWS* nptr;   // pointer to new adjacent vertex
  double maxad;  // maximum altitude difference among all adjacent vertices
  double adsum;  // total difference in altitudes
  int n;       // number of lower adjacent vertices
};


// perform hydraulic ersion on a field
// kc is the sediment capacity for each water particle
//  (i.e. how much sediment it can carry from one vertex to the next)
// kd is the deposition constant (i.e. how much sediment can be dropped
//  at each vertex in 1 iteration)
// ks is the soil softness which controls how much sediment can be
//  washed away by the water.
// iter is the number of iterations to perform
// freq is the frequency of rainfalls (-1 = no rain ever)
// rain is the amount of rain that falls per rainfall
// the algorithm always starts off with 1 rainfall
// sum delta-alts and distribute according to height diff
// what about water which hits a "valley"?  should make lakes and stuff
// scale ks (soil softness according to height?)

// if kc is too big then weird things happen
// (i.e. get heights taller than original landscape)
// probably when heights are all equal (i.e. on a flat plane)?
// and there is alot of sediment at a single vertex that has no
// where to go?

void MyField::H_Erode(Real kc, Real kd, Real ks, Real rain, int iter, int freq,
		      Real sea_level)
{
  AWS* curr_aws_array = new AWS[Size()];
  Real bsum = 0.0, asum = 0.0, ssum = 0.0;
  PercCounter pc("werode", iter);
  Real omin, omax, sea;
  int w = Width();
  int h = Height();
  int wadj = w - 2;
  int i, x, y;

  MinMaxHeights(omin, omax);

  sea = sea_level * (omax - omin);

  // initialize water, sediment and data arrays
  // make sure all heights are >= 0
  for(x = 0; x < w*h; x++) {
    curr_aws_array[x].alt = Height(x) - omin;
    curr_aws_array[x].Rain(rain);
#ifdef DEBUG
    bsum += curr_aws_array[x].alt;
#endif
  }

  for(i = 0; i < iter; i++, pc++) {
    AWS* curr_aws = curr_aws_array;

    for(y = 0; y < h; y++) {
      Boolean up = MakeBoolean(y > 0);
      Boolean down = MakeBoolean(y < h - 1);

      for(x = 0; x < w; x++, curr_aws++) {
	DWP dwp(curr_aws - w - 1, curr_aws - w - 1);
	Boolean left = MakeBoolean(x > 0);
	Boolean right = MakeBoolean(x < w - 1);
	double curr_a = curr_aws->alt;
	double curr_w = curr_aws->water;
	double curr_s = curr_aws->sed;   // amount of sediment at this vertex

	// top row
	dwp.Compute(left&&up, curr_a, curr_w, 1);
	dwp.Compute(up, curr_a, curr_w, 1);
	dwp.Compute(right&&up, curr_a, curr_w, wadj);

	// middle row
	dwp.Compute(left, curr_a, curr_w, 2);
	dwp.Compute(right, curr_a, curr_w, wadj);

	// bottom row
	dwp.Compute(left&&down, curr_a, curr_w, 1);
	dwp.Compute(down, curr_a, curr_w, 1);
	dwp.Compute(right&&down, curr_a, curr_w, 0);

	if(!dwp.n) {
	  // all adjacent vertices are higher, dump some sediment here
#if 1
	  double minad = FLT_MAX;
	  DW* cdw = dwp[0];	          // pointer to dw info
	  for(int k = dwp.n; k > 0; k--, cdw++)
	    if(minad > cdw->adiff)
	      minad = cdw->adiff;
	  double sdump = min(kd * curr_s, minad);
#else
	  double sdump = kd * curr_s;
#endif
	  curr_aws->alt += sdump;
	  curr_aws->sed -= sdump;
	  //assert(curr_aws->alt <= (omax - omin));
	  //assert(curr_aws->sed >= 0.0);
	} else {
	  // amount of water that can be dumped from this vertex
	  double wdump = min(curr_w, dwp.maxad);
	  double dperc = wdump / dwp.adsum;
	  double cs = kc * wdump;  	  // sediment capacity of wdump
	  DW* cdw = dwp[0];	          // pointer to dw info

	  // dump all water from this vertex
	  curr_aws->water -= wdump;
	  //assert(curr_aws->water >= 0.0);

	  if(curr_s >= cs) {
	    // if there is more sediment than the water can hold
	    // we deposit the difference at the current vertex
	    // and dump the maximum amount to the adjacent verticies
	    for(int k = dwp.n; k > 0; k--, cdw++) {
	      // dump fraction of water and sediment to adjacent vertex
	      double dw = dperc * cdw->adiff;
	      cdw->nptr->water += dw;
	      cdw->nptr->sed += kc * dw;
	    }
	    curr_aws->alt += kd * (curr_s - cs);
	    curr_aws->sed = (1.0 - kd) * (curr_s - cs);
	    //assert(curr_aws->sed >= 0.0);
	    //assert(curr_aws->alt <= (omax - omin));
	  } else {
	    double erode = min(curr_a, ks * (cs - curr_s));
	    double sperc = (curr_s + erode) / dwp.adsum;

	    // the water can carry more sediment than is available
	    // so we erode a little from the current vertex and
	    // add it to the amount transfered to the adjacent vertex
	    for(int k = dwp.n; k > 0; k--, cdw++) {
	      // dump fraction of water and sediment to adjacent vertex
	      double dw = dperc * cdw->adiff;
	      double sdump = sperc * cdw->adiff;
	      cdw->nptr->water += dw;
	      cdw->nptr->sed += sdump;
	    }
	    curr_aws->alt -= erode;
	    curr_aws->sed = 0.0;
	    //assert(curr_aws->alt >= 0.0);
	    //assert(curr_aws->alt <= (omax - omin));
	  }
	}
      }
    }
    /* add more rain every freq iterations */
    // should rain be proportional to alt + sed (i.e. for really big sed)
    if(freq > 0 && (i + 1) % freq == 0)
      for(x = 0; x < w*h; x++)
	curr_aws_array[x].Rain(rain);

    if(sea_level > 0.0) {
      for(x = 0; x < w*h; x++) {
	if(curr_aws_array[x].alt <= sea) {
	  curr_aws_array[x].water = 0.0;
	  curr_aws_array[x].sed = 0.0;
	}
      }
    }
  }

  // readjust heights
  for(i = 0; i < Size(); i++) {
#ifdef DEBUG
    asum += curr_aws_array[i].alt;
    ssum += curr_aws_array[i].sed;
#endif
#if 1
    // problem if there is too much sediment contained at this vertex!
    Height(i) = curr_aws_array[i].alt + curr_aws_array[i].sed + omin;
#else
    Height(i) = curr_aws_array[i].alt + omin;
#endif
  }

#ifdef DEBUG
  cout << "altitudes, before/after: " << bsum << "/";
  cout << "(" << asum << "/" << asum + ssum << ")" << endl;
#endif

  delete[] curr_aws_array;
}



#define tstep(ok, adj)\
if(ok && (adiff = alt - captr->alt - T) > 0.0) {\
  *awsp = captr;\
  *diffp = adiff;\
  total += adiff;\
} else {\
  *awsp = 0;\
}\
awsp++; diffp++; captr+=adj


void MyField::Erode(int iter, Real talus, Real perc, Real wind, Real dir,
		    Real kc, Real kd, Real ks, Real rain, int freq,
		    Real sea_level)
{
  AWS* curr_aws_array = new AWS[Size()];
  Real bsum = 0.0, asum = 0.0, ssum = 0.0;
  Real dx = fabs((*this)(1).x - (*this)(0).x);
  Real winds[8] = { 0.0, 0.0, 0.0,  0.0, 0.0,  0.0, 0.0, 0.0 };
  Real T = atan(deg2rad(talus)) * dx;   // convert talus angle to simple #
  Real omin, omax, sea;
  int w = Width();
  int h = Height();
  int wadj = w - 2;
  int i, x, y;

  MinMaxHeights(omin, omax);

  sea = sea_level * (omax - omin);

  // initialize water, sediment and data arrays
  // make sure all heights are >= 0
  for(x = 0; x < w*h; x++) {
    curr_aws_array[x].alt = Height(x) - omin;
    curr_aws_array[x].Rain(rain);
#ifdef DEBUG
    bsum += curr_aws_array[x].alt;
#endif
  }

  Real xwind = wind * fcos(dir);
  Real ywind = wind * fsin(dir);

  winds[0] += ywind*fcos(deg2rad(45.0));
  winds[1] += ywind;
  winds[2] += ywind*fcos(deg2rad(45.0));
  winds[5] -= ywind*fcos(deg2rad(45.0));
  winds[6] -= ywind;
  winds[7] -= ywind*fcos(deg2rad(45.0));

  winds[2] -= xwind*fcos(deg2rad(45.0));
  winds[4] -= xwind;
  winds[7] -= xwind*fcos(deg2rad(45.0));
  winds[0] += xwind*fcos(deg2rad(45.0));
  winds[3] += xwind;
  winds[5] += xwind*fcos(deg2rad(45.0));

#ifdef DEBUG
  cout << "winds: ";
#endif
  for(i = 0; i < 8; i++) {
    if(winds[i] < 0.0)
      winds[i] = 1.0 / (1.0 - winds[i]);
    else
      winds[i] += 1.0;
#ifdef DEBUG
    cout << winds[i] << " ";
#endif
  }
#ifdef DEBUG
  cout << endl;
#endif

  PercCounter pc("erode", iter);
  for(i = 0; i < iter; i++, pc++) {
    AWS* curr_aws = curr_aws_array;

    // do one thermal erosion step
    for(y = 0; y < h; y++) {
      Boolean up = MakeBoolean(y > 0);
      Boolean down = MakeBoolean(y < h - 1);

      for(x = 0; x < w; x++, curr_aws++) {
	Boolean left = MakeBoolean(x > 0);
	Boolean right = MakeBoolean(x < w - 1);
	AWS* captr = curr_aws - w - 1;
	Real alt = curr_aws->alt;
	Real diffs[8];
	AWS* aptrs[8];
	Real* diffp = diffs;
	AWS** awsp = aptrs;
	Real total = 0.0;
	Real adiff;

	tstep(left&up, 1);
	tstep(up, 1);
	tstep(right&up, wadj);
	tstep(left, 2);
	tstep(right, wadj);
	tstep(left&down, 1);
	tstep(down, 1);
	tstep(right&down, 1);

	if(total > 0.0) {
	  Real* wp = winds;
	  diffp--; awsp--;
	  for(; diffp >= diffs; diffp--, awsp--, wp++) {
	    if(*awsp) {
	      Real rperc = min(0.4, *wp * perc);
	      Real rubble = rperc * *diffp;
	      (*awsp)->alt += rubble;
	      curr_aws->alt -= rubble;
	    }
	  }
	}
      }
    }

    // do one hydraulic erosion step
    curr_aws = curr_aws_array;
    for(y = 0; y < h; y++) {
      Boolean up = MakeBoolean(y > 0);
      Boolean down = MakeBoolean(y < h - 1);

      for(x = 0; x < w; x++, curr_aws++) {
	DWP dwp(curr_aws - w - 1, curr_aws - w - 1);
	Boolean left = MakeBoolean(x > 0);
	Boolean right = MakeBoolean(x < w - 1);
	double curr_a = curr_aws->alt;
	double curr_w = curr_aws->water;
	double curr_s = curr_aws->sed;   // amount of sediment at this vertex

	// top row
	dwp.Compute(left&&up, curr_a, curr_w, 1);
	dwp.Compute(up, curr_a, curr_w, 1);
	dwp.Compute(right&&up, curr_a, curr_w, wadj);

	// middle row
	dwp.Compute(left, curr_a, curr_w, 2);
	dwp.Compute(right, curr_a, curr_w, wadj);

	// bottom row
	dwp.Compute(left&&down, curr_a, curr_w, 1);
	dwp.Compute(down, curr_a, curr_w, 1);
	dwp.Compute(right&&down, curr_a, curr_w, 0);

	if(!dwp.n) {
	  // all adjacent vertices are higher, dump some sediment here
#if 1
	  double minad = FLT_MAX;
	  DW* cdw = dwp[0];	          // pointer to dw info
	  for(int k = dwp.n; k > 0; k--, cdw++)
	    if(minad > cdw->adiff)
	      minad = cdw->adiff;
	  double sdump = min(kd * curr_s, minad);
#else
	  double sdump = kd * curr_s;
#endif
	  curr_aws->alt += sdump;
	  curr_aws->sed -= sdump;
	  //assert(curr_aws->alt <= (omax - omin));
	  //assert(curr_aws->sed >= 0.0);
	} else {
	  // amount of water that can be dumped from this vertex
	  double wdump = min(curr_w, dwp.maxad);
	  double dperc = wdump / dwp.adsum;
	  double cs = kc * wdump;  	  // sediment capacity of wdump
	  DW* cdw = dwp[0];	          // pointer to dw info

	  // dump all water from this vertex
	  curr_aws->water -= wdump;
	  //assert(curr_aws->water >= 0.0);

	  if(curr_s >= cs) {
	    // if there is more sediment than the water can hold
	    // we deposit the difference at the current vertex
	    // and dump the maximum amount to the adjacent verticies
	    for(int k = dwp.n; k > 0; k--, cdw++) {
	      // dump fraction of water and sediment to adjacent vertex
	      double dw = dperc * cdw->adiff;
	      cdw->nptr->water += dw;
	      cdw->nptr->sed += kc * dw;
	    }
	    curr_aws->alt += kd * (curr_s - cs);
	    curr_aws->sed = (1.0 - kd) * (curr_s - cs);
	    //assert(curr_aws->sed >= 0.0);
	    //assert(curr_aws->alt <= (omax - omin));
	  } else {
	    double erode = min(curr_a, ks * (cs - curr_s));
	    double sperc = (curr_s + erode) / dwp.adsum;

	    // the water can carry more sediment than is available
	    // so we erode a little from the current vertex and
	    // add it to the amount transfered to the adjacent vertex
	    for(int k = dwp.n; k > 0; k--, cdw++) {
	      // dump fraction of water and sediment to adjacent vertex
	      double dw = dperc * cdw->adiff;
	      double sdump = sperc * cdw->adiff;
	      cdw->nptr->water += dw;
	      cdw->nptr->sed += sdump;
	    }
	    curr_aws->alt -= erode;
	    curr_aws->sed = 0.0;
	    //assert(curr_aws->alt >= 0.0);
	    //assert(curr_aws->alt <= (omax - omin));
	  }
	}
      }
    }

    /* add more rain every freq iterations */
    // should rain be proportional to alt + sed (i.e. for really big sed)
    if(freq > 0 && (i + 1) % freq == 0)
      for(x = 0; x < w*h; x++)
	curr_aws_array[x].Rain(rain);

    if(sea_level > 0.0) {
      for(x = 0; x < w*h; x++) {
	if(curr_aws_array[x].alt <= sea) {
	  curr_aws_array[x].water = 0.0;
	  curr_aws_array[x].sed = 0.0;
	}
      }
    }
  }

  // readjust heights
  for(i = 0; i < Size(); i++) {
#ifdef DEBUG
    asum += curr_aws_array[i].alt;
    ssum += curr_aws_array[i].sed;
#endif
#if 1
    // problem if there is too much sediment contained at this vertex!
    Height(i) = curr_aws_array[i].alt + curr_aws_array[i].sed + omin;
#else
    Height(i) = curr_aws_array[i].alt + omin;
#endif
  }

#ifdef DEBUG
  cout << "altitudes, before/after: " << bsum << "/";
  cout << "(" << asum << "/" << asum + ssum << ")" << endl;
#endif

  delete[] curr_aws_array;
}



class Wdiff {
 public:
  Wdiff() : adiff(0.0), wptr(0) { }
  Wdiff(int ok, Real aws, Real* ap, Real* wp)
    {
      wptr = 0;
      if(ok && (adiff = aws - (*ap + *wp)) > 0.0) {
	wptr = wp;
        total += adiff;
        maxad = max(maxad, adiff);
      }
    }
  Real adiff;
  Real* wptr;
  static Real total;
  static Real maxad;
};

Real Wdiff::total = 0.0;
Real Wdiff::maxad = 0.0;

// thermal erosion.  the talus angle is the angle needed between adjacent
// elements for erosion to take place (e.g. slope of sides).  perc is the
// amount of material that may be eroded from any vertex.  iter is the
// number of iterations to perform
MyField* MyField::Water(Real sea, Real stop, Real evap, Real rain,
			int freq, Real thresh, Real* opac, int max)
{
  Real* alts = Heights();             // get heights
  Real* water = new Real[Size()];
  Real wsum, wdrip, omin, omax;
  Wdiff diffs[8];
  int w = Width();
  int h = Height();
  int wadj = w - 2;
  int iter = 0;
  int i;

  // initial rain
  MinMaxHeights(omin, omax);
  for(i = 0; i < Size(); i++) {
    alts[i] -= omin;
    water[i] = rain * alts[i];
  }
  sea *= (omax - omin);

  do {
    Real* altp = alts;
    Real* watp = water;
    wsum = wdrip = 0.0;

    for(int y = 0; y < h; y++) {
      Boolean up = MakeBoolean(y > 0);
      Boolean down = MakeBoolean(y < h - 1);

      for(int x = 0; x < w; x++, altp++, watp++) {
	Real alt = *altp;
	Real wat = *watp;

	if(wat > 0.0) {
	  Boolean left = MakeBoolean(x > 0);
	  Boolean right = MakeBoolean(x < w - 1);
	  Real* aptr = altp - w - 1;
	  Real* wptr = watp - w - 1;
	  Real aws = alt + wat;
	  Wdiff* diffp = diffs;

          Wdiff::total = 0.0;
          Wdiff::maxad = 0.0;

	  *diffp++ = Wdiff(up&&left, aws, aptr, wptr); aptr++; wptr++;
	  *diffp++ = Wdiff(up, aws, aptr, wptr); aptr++; wptr++;
	  *diffp++ = Wdiff(up&&right, aws, aptr, wptr); aptr+=wadj; wptr+=wadj;

	  *diffp++ = Wdiff(left, aws, aptr, wptr); aptr+=2; wptr+=2;
	  *diffp++ = Wdiff(right, aws, aptr, wptr); aptr+=wadj; wptr+=wadj;

	  *diffp++ = Wdiff(down&&left, aws, aptr, wptr); aptr++; wptr++;
	  *diffp++ = Wdiff(down, aws, aptr, wptr); aptr++; wptr++;
	  *diffp = Wdiff(down&&right, aws, aptr, wptr);

	  if(Wdiff::total > 0.0) {
	    double dw = min(Wdiff::maxad, wat);
	    double dperc = dw / Wdiff::total;
	    for(; diffp >= diffs; diffp--) {
	      if(diffp->wptr) {
		Real runoff = dperc * diffp->adiff;
		*diffp->wptr += runoff;
		*watp -= runoff;
		wdrip += runoff;
	      }
	    }
	  }
	}
      }
    }

    // evaporate some water at each iteratation (and possibly rain some more)
    Boolean dorain = MakeBoolean(freq > 0 && (iter + 1) % freq == 0);
    watp = water;
    altp = alts;
    for(y = 0; y < h; y++) {
      for(int x = 0; x < w; x++, watp++, altp++) {
	if(dorain)
	  *watp = *watp * evap + *altp * rain;
	else
	  *watp *= evap;
	if(*watp + *altp < sea) *watp = 0.0;
	wsum += *watp;
      }
    }

#ifdef DEBUG
    if(iter % 10 == 0)
      cout << "iter: " << iter << ", " << wdrip << "/" << wsum << endl;
#endif
    iter++;
  } while(wdrip/wsum > stop && (iter < max || !max));

  // may need to nuke edge water here...
  // might be able to use slope around vertices to determine when to nuke water
  MyField* f = new MyField(*this, True, True);
  for(i = 0; i < f->Size(); i++) {
#if 1
    if(water[i] > (omax - omin) * thresh)
      f->Height(i) += water[i];
    else
      f->Height(i) = omin;
#else
    f->Height(i) = water[i];
#endif
  }

  delete[] alts;
  delete[] water;

  if(opac) {
    for(int i = 0; i < Size(); i++)
      opac[i] = (f->Height(i) == omin) ? 0.0 : 1.0;
  }

  return f;
}
