#include "real.h"
#include "fft.h"
#include "hmap.h"
#include "texture.h"

#define A Height
#define adj (delta * grand())

// Generate a landscape using two dimensional fractal brownian motion
// the algorithm uses the diamond square subdivision algorithm
// (i.e. succesize iterations switch from squares to diamonds)
// at each iteration the corners of the "pieces" are raised or
// lowered by a random amount.
// for this code to work right the field size must be a power of 2
// adapted from code by John Alferness
void MyField::TwoDfbm(int level, Real sigma, Real H, Boolean flg)
{
  Real pow_inc = pow(0.5, 0.5 * H);
  Real delta = sigma;
  int w = Width();
  int h = Height();
  int cw = w;
  int ch = h;
  int chw = cw/2;
  int chh = ch/2;

  // initialize corners
  A(0, 0) = adj;
  A(0, ch-1) = adj;
  A(cw-1, 0) = adj;
  A(cw-1, ch-1) = adj;

  for(int i = 0; i < level; i++) {
    int x, y;

    delta *= pow_inc;

    // sum corners
    for(x = chw; x < w - chw; x += cw) {
      for(y = chh; y < h - chh; y += ch) {
	A(x, y) = adj + (A(x+chw, y+chh) + A(x+chw, y-chh) +
			 A(x-chw, y+chh) + A(x-chw, y-chh))/4.0;
      }
    }

    // indicies ok
    if(flg) {
      for(x = 0; x < w; x += cw)
	for(y = 0; y < h; y += ch)
	  A(x, y) += adj;
    }

    delta *= pow_inc;

    for(x = chw; x < w - chw; x += cw) {
      A(x, 0) = adj + (A(x+chw, 0) + A(x-chw, 0) + A(x, chh)) / 3.0;
      A(x, h-1) = adj + (A(x+chw, h-1) + A(x-chw, h-1) + A(x, h-chh)) / 3.0;
    }
    for(y = chh; y < h - chh; y += ch) {
      A(0, y) = adj + (A(0, y+chh) + A(0, y-chh) + A(chw, y)) / 3.0;
      A(w-1, y) = adj + (A(w-1, y+chh) + A(w-1, y-chh) + A(w-chw, y)) / 3.0;
    }

    for(x = chw; x < w - chw; x += cw)
      for(y = ch; y < h - chh; y += ch)
	A(x, y) = adj + (A(x,y+chh)+A(x,y-chh)+A(x+chw,y)+A(x-chw,y)) / 4.0;

    for(x = cw; x < w - chw; x += cw)
      for(y = chh; y < h - chh; y += ch)
	A(x, y) = adj + (A(x,y+chh)+A(x,y-chh)+A(x+chw,y)+A(x-chw,y)) / 4.0;

    // indicies ok
    if(flg) {
      for(x = 0; x < w; x += cw)
	for(y = 0; y < h; y += ch)
	  A(x, y) += adj;

      for(x = chw; x < w - chw; x += cw)
	for(y = chh; y < h - chh; y += ch)
	  A(x, y) += adj;
    }

    cw >>= 1;
    ch >>= 1;
    chw >>= 1;
    chh >>= 1;
  }
}

// use spectral synthesis of 2d noise to create a landscape.
// first a field of noise is generated using H to control it's "steepness"
// then an inverse FFT is performed on the noise to generate a landscape.
// Its magic!
// assumes h/w are powers of 2.
void MyField::SpectralSynthesis2d(Real H)
{
  MyComplex* cdata = new MyComplex[Size()];
  Real expo = -(H + 1.0) / 2.0;
  int hh = Height()/2;
  int hw = Width()/2;
  int i, j;

  // generate noise for first half of field
  for(i = 0; i <= hw; i++) {
    int i0 = i ? Width() - i : 0;
    for(j = 0; j <= hh; j++) {
      int j0 = j ? Height() - j : 0;
      Real phase = 2.0 * M_PI * rrand(1.0);
      Real rad = (i || j) ? pow(i*i + j*j, expo) * grand() : 0.0;
      cdata[i + j * Width()].r = rad * fcos(phase);
      cdata[i + j * Width()].i = rad * fsin(phase);
      cdata[i0 + j0 * Width()].r = rad * fcos(phase);
      cdata[i0 + j0 * Width()].i = rad * -fsin(phase);
    }
  }
    
  cdata[i-1].i = 0.0;                  // (hw, 0)
  cdata[Width()*(j-1)].i = 0.0;        // (0, hh)
  cdata[i-1 + Width()*(j-1)].i = 0.0;  // (hw, hh)
	  
  // generate noise for second half of field
  for(i = 1; i < hw; i++) {
    int i0 = Width() - i;
    for(j = 1; j < hh; j++) {
      int j0 = Height() - j;
      Real phase = 2.0 * M_PI * rrand(1.0);
      Real rad = (i || j) ? pow(i*i + j*j, expo) * grand() : 0.0;
      cdata[i + j0 * Width()].r = rad * fcos(phase);
      cdata[i + j0 * Width()].i = rad * fsin(phase);
      cdata[i0 + j * Width()].r = rad * fcos(phase);
      cdata[i0 + j * Width()].i = -rad * fsin(phase);
    }
  }


  // perform the inverse FFT
  InvFFT2d(cdata, Width(), Height());

  // suck out the real components of the result as heights
  for(i = 0; i < Width(); i++) {
    for(j = 0; j < Height(); j++) {
      (*this)(i, j).z = cdata[i + j * Width()].r;
    }
  }

  delete[] cdata;
}


#if 0
// 1d spectral synthesis
// for some bizarre/stupid reason it always crashes here if I try to
// allocate memory instead of using the static array....... ARAGARH!
void MyField::SpectralSynthesis1d(Real H)
{
//  MyComplex* cdata = new MyComplex[Width()];
  static MyComplex cdata[10000];
  Real expo = -(H + 1.0) / 2.0;

  // generate noise for 1d field
  for(int i = 0; i < Width()/2; i++) {
    int i0 = Width() - i - 1;
    Real phase = 2.0 * M_PI * rrand(1.0);
    Real rad = i ? pow(i*i, expo) * grand() : 0.0;
    cdata[i].r = rad * fcos(phase);
    cdata[i].i = rad * fsin(phase);
    cdata[i0].r = rad * fcos(phase);
    cdata[i0].i = -rad * fsin(phase);
  }
  cdata[i].i = 0.0;

  // perform the inverse FFT
  InvFFT1d(cdata, Width());

  // suck out the real components of the result as heights
  for(i = 0; i < Width(); i++)
    Height(i) = cdata[i].r;

//  delete[] cdata;
}
#endif
