////////////////////////////////////////////////////////////////////////////
//
// hmap.h - height map class which is the basic class for almost all
//          operations.  The height map class is actually a 3d field of
//          points and colors (normally it is only 2d though)
//
////////////////////////////////////////////////////////////////////////////
#ifndef _HMAP_H_
#define _HMAP_H_

#include "Object.h"
#include "real.h"
#include "color.h"
#include "point.h"

class PNM;
class PBM;
class Palette;
class Texture;
class Filter;

// function that can be used to merge two fields in an arbitrary way
typedef Point3d (*FieldMergeFn)(const Point3d&, const Point3d&, void*);

// basic height map class
class MyField : public Object {
  DeclareStandardMethods(MyField, Object);
 public:
  // copy position data from given field
  virtual void CopyData(const MyField&);

  // copy color data from given field
  virtual void CopyColors(const MyField&);

  // perform thermal erosion (see erode.c for details)
  virtual void T_Erode(Real talus, Real perc, int iter, Real=0.0, Real=0.0);

  // perform hydraulic erosion (see erode.c for details)
  virtual void H_Erode(Real kc, Real kd, Real ks, Real rain, int itr, int frq, Real = 0.0);

  // perform thermal & hydraulic erosion (see erode.c for details)
  virtual void Erode(int iter, Real talus, Real perc, Real wind, Real dir,
		     Real kc, Real kd, Real ks, Real rain, int frq, Real);

  // water stuff
  virtual MyField* Water(Real sea = 0.0, Real = 0.0001, Real = 1.0,
			 Real = 0.05, int = -1, Real = 0.05, Real* = 0,
			 int max = 0);

  // perform craterization (see crater.c for details)
  virtual void Craterize(int itr, Real dp = 1.5, Real db = 0.0, Real sp = 0.9);

  // perform spectral synthesis (see gen.c for details)
  virtual void SpectralSynthesis2d(Real H);

  // perform FBM (see gen.c for details)
  virtual void TwoDfbm(int level, Real sigma, Real H, Boolean = True);

  // push a pyramid/cones/spheres into height map (see hmap.c)
  virtual void Praise(int xc, int yc, int hw, int hh, Real amt, Real zoff = 0.0, Real truncp = 1.0);
  virtual void Craise(int xc, int yc, int rad, Real amt, Real zoff = 0.0, Real truncp = 1.0);
  virtual void Sraise(int xc, int yc, int rad, Real amt, Real zoff = 0.0, Real truncp = 1.0);
  virtual void Raise(int iter, char*);

  // initialize field as a regular grid given origins and deltas
  virtual void InitRegularGrid(Real ox, Real oy, Real oz, Real dx, Real dy, Real dz = 0.0);

  // has this field been colored?
  virtual Boolean IsColored() const { return MakeBoolean(_color); }

  // rotate this field in 3 dimensions by the given angles
  virtual void Rotate(Real, Real, Real);

  // apply a texture to this field
  virtual void Texturize(const Texture& t);

  // colorize this field with the given palette and min/max values
  virtual void Colorize(const Palette&, Real minp = 0.0, Real maxp = 1.0);

  // gray scale this field with the given min/max intensities
  virtual void IntensityMap(GrayScale = 0.0, GrayScale = 1.0);

  // remove the color from this field
  virtual void DeColorize();

  // initialize the color component of this field to RGB values
  virtual void InitRGB() { set_colors(new RGB[Size()]); }

  // initialize the color component of this field to gray scale values
  virtual void InitGrayScale() { set_colors(new GrayScale[Size()]); }

  // save this field to a PNM
  virtual int Save(char* file, int k = 0) const;
  virtual int SaveGif(char* file, Boolean quant = False) const;

  // load this field from a PNM
  // (with optional palette to perform mapping from color to height)
  // (if no palette is given then color intensity is used to get height)
  virtual int Load(char* file, const Palette* = 0);

  // add/subtract/multiply/divide two fields
  virtual const MyField& operator+=(const MyField&);
  virtual const MyField& operator-=(const MyField&);
  virtual const MyField& operator*=(const MyField&);
  virtual const MyField& operator/=(const MyField&);

  // add/subtract/multiply/divide a point to this field
  virtual const MyField& operator+=(const Point3d&);
  virtual const MyField& operator-=(const Point3d&);
  virtual const MyField& operator*=(const Point3d&);
  virtual const MyField& operator/=(const Point3d&);

  // add/subtract/multiply/divide a constant to this field
  virtual const MyField& operator+=(Real);
  virtual const MyField& operator-=(Real);
  virtual const MyField& operator*=(Real);
  virtual const MyField& operator/=(Real);

  // merge this field with another one in an arbitrary way
  virtual void Merge(int xa, int ya, int w, int h,
		     const MyField&, int xb, int yb,
		     int xo, int yo, FieldMergeFn,
		     void* env);

  // average the heights of two fields
  virtual void AverageHeights(const MyField&);

  // add the height components of two fields
  virtual void AddHeights(const MyField& f);

  // subtract the height components of two fields
  virtual void SubtractHeights(const MyField& f);

  int Size() const { return _width*_height*_depth; } // total size of this field
  int Width() const { return _width; }    // width of this field
  int Height() const { return _height; }  // height of this field
  int Depth() const { return _depth; }    // depth of this field

  // return the size in the given dimension 0=x, 1=y, 2=z
  int Dim(int d) const { return !d ? _width : (d == 1 ? _height : _depth); }

  Real Granularity() const { return _gran; }  // granularity of this field

  // does this field own it's data component?
  Boolean Own() const { return _own; }

  // kind of color this field is colored with (RGB or GrayScale)
  ClassId ColorType() const { return color() ? color(0).ObjectId() : 0; }

  // array containing just the heights of this field (must be freed by user)
  Real* Heights() const;

  // set the heights of this field to the heights contained inthe given array
  void SetHeights(Real*);

  // set all the heights to the given value
  void SetHeight(Real);

  // set a rectangular range of heights
  void SetHeight(int x1, int y1, int x2, int y2, Real);

  // find the minimum and maximum heights of a field
  void MinMaxHeights(Real&, Real&) const;

  // scale the heights of a field
  void ScaleHeights(Real = 1.0);

  // find the average height of a field
  Real AverageHeight() const;

  // find the average height over a given region
  Real AverageHeight(int x1, int y1, int x2, int y2) const;

  // find the minimum and maximum extents in one dimension of a field
  void MinMaxDim(int, Real&, Real&) const;
  void MinMaxDim(int, Point3d&, Point3d&) const;

  // scale the positions of a field
  void ScalePositions(Real = 1.0, Real = 1.0, Real = 1.0);

  // perform gaussian smoothing on a field
  void Smooth(Real sig);

  // perform edge detection on a field
  void EdgeDetect(Real);

  // perform arbitrary convolution on a field
  void Convolute(const Filter&);

  // access field coordinates by index
  Point3d& operator()(int i) const
    { return ((MyField*)this)->_data[i]; }
  Point3d& operator()(int i, int j) const
    { return ((MyField*)this)->_data[i+j*_width]; }
  Point3d& operator()(int i, int j, int k) const
    { return ((MyField*)this)->_data[i+j*_width+k*_height*_width]; }

  // access field heights by index
  Real& Height(int i) const
    { return ((MyField*)this)->_data[i].z; }
  Real& Height(int i, int j) const
    { return ((MyField*)this)->_data[i+j*_width].z; }
  Real& Height(int i, int j, int k) const
    { return ((MyField*)this)->_data[i+j*_width+k*_width*_height].z; }

  // access field colors by index
  Color& color(int i) const
    { return ((MyField*)this)->_color[i]; }
  Color& color(int i, int j) const
    { return ((MyField*)this)->_color[i+j*_width]; }
  Color& color(int i, int j, int k) const
    { return ((MyField*)this)->_color[i+j*_width+k*_height*_width]; }

  // convert field to a PNM
  virtual PNM* ToPNM(int=0) const;

  // create an empty field
  MyField();

  // join two fields together
  MyField(const MyField&, const MyField&);

  // load field from given file using palette for height mapping
  // (if no palette is given then color intensity is used to get height)
  MyField(char* file, const Palette* = 0);

  // create field from PNM using palette for height mapping
  // (if no palette is given then color intensity is used to get height)
  MyField(const PNM&, const Palette* = 0);

  // copy field.  the first boolean controls whether the data and color
  // components are copied from the field or if the pointers are just copied.
  // the second boolean controls whether this field owns it's data and color
  // components (i.e. whether it deletes them when it is destroyed)
  MyField(const MyField&, Boolean = False, Boolean = False);

  // create a new field of the given size
  MyField(int w, int h, int d = 1, Real g = 1.0);

  // destroy a field
  ~MyField();

 protected:
  // return references to private member data
  Real& gran() const { return ((MyField*)this)->_gran; }
  Point3d* data() const { return ((MyField*)this)->_data; }
  Color* color() const { return ((MyField*)this)->_color; }
  Boolean& own() const { return ((MyField*)this)->_own; }

  // set field sizes (if = 0 then skip)
  virtual void set_size(int = 0, int = 0, int = 0, Real = 0.0);

  // set field data and color components
  // (deletes old components if they exist)
  virtual void set_data(Point3d* p);
  virtual void set_colors(Color* c);

  // compute dx and dy values for this field
  void compute_deltas();

 private:
  int _width;       // field width
  int _height;      // field height
  int _depth;       // field depth
  Real _gran;       // field granularity (not really used)
  Point3d* _data;   // field data
  Color* _color;    // field colors
  Boolean _own;     // own flag for data and colors
};

#endif
