////////////////////////////////////////////////////////////////////////////
//
// main.c - command line interface to all functions.  generates PNM or GIF
//          files as output (run fractal -help for more info)
//
////////////////////////////////////////////////////////////////////////////
#include <ctype.h>
#include <errno.h>
#include <string.h>
#include "misc.h"
#include "real.h"
#include "planet.h"
#include "hmap.h"
#include "texture.h"
#include "pnm++.h"
#include "terrain.h"
#include "ocean.h"
#include "trig.h"
#include "noise.h"

enum FieldOption {
  FBM                = 0x00000001,  // use FBM to generate landscape
  SPECTRAL_SYNTHESIS = 0x00000002,  // use spectral synthesis to make landscape
  REGULAR_GRID       = 0x00000004,  // make a 2d grid of constant height
  IRREGULAR_GRID     = 0x00000008,  // make a 2d grid with sloped heights

  SRAISE             = 0x00000010,  // make landscape using sphere raising
  PRAISE             = 0x00000020,  // make landscape using pyramid raising
  CRAISE             = 0x00000040,  // make landscape using cone raising

  EROSION            = 0x00000100,  // do thermal & hydraulic erosion
  THERMAL_EROSION    = 0x00000200,  // do thermal erosion
  HYDRAULIC_EROSION  = 0x00000400,  // do hydraulic erosion

  CRATERIZE          = 0x00000800,  // craterize a landscape
  ROTATE             = 0x00001000,  // rotate landscape
  SMOOTH             = 0x00002000,  // smooth a landscape
  EDGE_DETECT        = 0x00004000,  // do edge detection on landscape
  COLORIZE           = 0x00008000,  // colorize a landscape
  GRAY_SCALE         = 0x00010000,  // gray scale a landscape
  SCALE              = 0x00020000,  // scale landscape heights
  PLANET             = 0x00040000,  // create a fractal planet
  HMAP               = 0x00080000,  // create an empty height map
  OCEAN              = 0x00100000,  // create an ocean for a landscape
  TEXTURE            = 0x00200000,  // create a texture field
  TURBULENCE         = 0x00400000,  // create a turbulence field
  WATER              = 0x00800000,  // rain, rain, rain!
  INFO               = 0x20000000,  // print info on landscapes
  TEST_TRIG_TAB      = 0x40000000,  // test trig tables
  HELP               = 0x80000000   // print help
};

// main option table
struct {
  char* name;   // option name
  char* junk;   // unused
  int opt;      // option flag
  char* help;   // help string
} Options[] = {
  { "-fbm", 0, FBM, "generate landscape using fractal brownian motion" },
  { "-ssyn", 0, SPECTRAL_SYNTHESIS, "generate landscape using spectral synthesis" },
  { "-grid", 0, REGULAR_GRID, "generate a regular, constant height grid" },
  { "-igrid", 0, IRREGULAR_GRID, "generate an irregular grid" },
  { "-sraise", 0, SRAISE, "generate landscape using random sphere raising" },
  { "-praise", 0, PRAISE, "generate landscape using random pyramid raising" },
  { "-craise", 0, CRAISE, "generate landscape using random cone raising" },
  { "-terode", 0, THERMAL_EROSION, "perform thermal erosion on a landscape" },
  { "-herode", 0, HYDRAULIC_EROSION, "perform hydraulic erosion on a landscape" },
  { "-erode", 0, EROSION, "perform hydraulic and thermal erosion" },
  { "-crater", 0, CRATERIZE, "craterize a landscape" },
  { "-rotate", 0, ROTATE, "rotate a landscape (busted?)" },
  { "-smooth", 0, SMOOTH, "smooth a landscape using gaussian filter" },
  { "-edge", 0, EDGE_DETECT, "perform edge detection on a landscape" },
  { "-color", 0, COLORIZE, "colorize a landscape" },
  { "-gray", 0, GRAY_SCALE, "create an intensity picture of a landscape" },
  { "-planet", 0, PLANET, "generate a fractal planet" },
  { "-hmap", 0, HMAP, "generate a hieght map" },
  { "-ocean", 0, HMAP | OCEAN, "create an ocean for a landscape" },
  { "-text", 0, TEXTURE, "create a texture" },
  { "-turb", 0, TURBULENCE, "create a turbulence field" },
  { "-water", 0, WATER, "rain, rain and more rain!" },
  { "-info", 0, INFO, "display information about a landscape" },
  { "-ttt", 0, TEST_TRIG_TAB, "test trig tables" },
  { "-help", 0, HELP, "get some help (too many damn options, huh?)" },
  { 0, 0, 0 }
};

// texture types
enum TextureType {
  PLANAR,       // planar cloud texture
  ELLIPSOID,    // ellipsoidal cloud texture
  MARBLE        // marble texture
};

// general options
static int flags = 0;           // main operation
static int width = 256;         // landscape width
static int height = 256;        // landscape height
static int depth = 1;           // landscape depth
static Real granularity = 1.0;  // landscape granularity
static int iter = 500;          // iterations for iterative operations
static Real scale = 0.0;        // scale factor for heights
static Boolean gif = True;      // save as gif file
static Boolean quant = False;   // quantize colors of ppm for gif save

// seed for random numbers
static int seed = 0;

// gray scale options
static Real imin = 0.0;         // minimum gray scale intensity
static Real imax = 1.0;         // maximum gray scale intensity

// texture options
static TextureType text_type = PLANAR;  // default texture type

// turbulence options
static Real tscale = 10.0;      // scale of texture to textured field
static Boolean tlerp = False;   // unused
static int terpw = 0;           // unused

// color/planet/ocean opt
static Real sea_level = 0.5;    // sea level for landscapes

// palette opt
static Real dark = 0.0;
static Real bright = 0.0;
static Real cscale = 0.0;

// angles for rotation, etc.
static Real theta = 45.0;
static Real phi = 45.0;
static Real alpha = 0.0;

// grid options
static Boolean valley = False;
static Boolean sloped = False;
static Boolean dsloped = False;
static Boolean bumpy1 = False;
static Boolean bumpy2 = False;
static Boolean bumpy3 = False;
static Boolean drop = False;

// fbm & ssyn options
static Real H = 0.8;            // related to fractal dimension
static Real sigma = 1.0;        // sigma for gaussian fns and noise fn

// crater options
static int citer = 30000;       // number of crater interations
static Real dbias = sqrt(0.5);  // factor for crater size
static Real dpower = 1.5;       // factor for crater size

// grid options
static Real high = 0.0;         // height of grid

// hydraulic erosion opts
static int hiter = 500;         // number of iterations
static Real kc = 3;             // sediment capacity (proportional to height)
static Real kd = 0.1;           // deposition constant
static Real ks = 0.3;           // soil softness
static Real rain = 0.001;       // rain per rainfall
static Real freq = 0.93;        // frequency of rainfalls

// thermal erosion opts
static int titer = 200;         // number of iterations
static Real tperc = 0.05;       // percent of rubble that can be eroded
static Real talus = 20.0;       // talus angle
static Real twind = 0.0;        // wind

// water opts
static Real evap = 1.0;         // evaporation percentage

// ocean opts
static Real otime = 10.0;       // new ocean time
static Real wave = 0.10;        // wave shape
static Real wind = 2.5;         // wind speed
static Real k0 = 0.1;           // shape constants
static Real kx = 0.1;
static Real kz = 0.2;
static Real grav = 9.81;        // gravity
static Boolean merge = False;   // merge with shore

// minor options
struct {
  char* name;      // name
  char* format;    // format string for option values
  void* ptr[10];   // pointers to option variables
  char* help;      // help string
} ValueOpts[] = {
  { "-w", "%d", { &width }, "width of landscape" },
  { "-h", "%d", { &height }, "height of landscape" },
  { "-d", "%d", { &depth }, "depth of landscape" },
  { "-high", "%f", { &high }, "initial height for grids" },
  { "-g", "%f", { &granularity }, "granularity of landscape" },
  { "-scale", "%f", { &scale }, "scale factor for scale functions" },
  { "-seed", "%d", { &seed }, "seed for random number generator" },
  { "-theta", "%f", { &theta }, "theta angle" },
  { "-phi", "%f", { &phi }, "phi angle" },
  { "-alpha", "%f", { &alpha }, "alpha angle" },
  { "-valley", "%b", { &valley }, "valley grid" },
  { "-sloped", "%b", { &sloped }, "sloped grid" },
  { "-dsloped", "%b", { &dsloped }, "diagonally sloped grid" },
  { "-bumpy1", "%b", { &bumpy1 }, "bumpy grid 1" },
  { "-bumpy2", "%b", { &bumpy2 }, "bumpy grid 2" },
  { "-bumpy3", "%b", { &bumpy3 }, "bumpy grid 3" },
  { "-drop", "%b", { &drop }, "drop grid" },
  { "-cscale", "%f", { &cscale }, "color scaling for terrain palette" },
  { "-dark", "%f", { &dark }, "darkening factor for terrain palette" },
  { "-bright", "%f", { &bright }, "brightening factor for terrain palette" },
  { "-gif", "%b", { &gif }, "save as a gif file" },
  { "-quant", "%b", { &quant }, "quantize ppm to 256 colors" },
  { "-sea", "%f", { &sea_level }, "sea level (as percent of height)" },
  { "-imin", "%f", { &imin }, "minimum intensity for intensity maps" },
  { "-imax", "%f", { &imax }, "maximum intensity for intensity maps" },
  { "-iter", "%d", { &iter }, "number of iterations for generation algs." },
  { "-citer", "%d", { &citer }, "iterations for craterization" },
  { "-dbias", "%f", { &dbias }, "depth bias for craters" },
  { "-dpower", "%f", { &dpower }, "depth power for craters" },
  { "-hiter", "%d", { &hiter }, "iterations for hydraulic erosion" },
  { "-kc", "%f", { &kc }, "sediment capacity for hydraulic erosion" },
  { "-kd", "%f", { &kd }, "sediment deposition factor for hydraulic erosion" },
  { "-ks", "%f", { &ks }, "soil softness for hydraulic erosion" },
  { "-rain", "%f", { &rain }, "amount of rain per rainfall" },
  { "-freq", "%f", { &freq }, "frequency of rainfalls" },
  { "-evap", "%f", { &evap }, "evaporation percentage" },
  { "-titer", "%d", { &titer }, "iterations for thermal erosion" },
  { "-tperc", "%f", { &tperc }, "percent of rubble for thermal erosion" },
  { "-talus", "%f", { &talus }, "talus angle for thermal erosion" },
  { "-twind", "%f", { &twind }, "wind velocity for thermal erosion" },
  { "-H", "%f", { &H }, "fractal dimension (for fbm and ssyn)" },
  { "-sigma", "%f", { &sigma }, "sigma for gaussian functions" },
  { "-time", "%f", { &otime }, "elapsed time for ocean generation" },
  { "-wave", "%f", { &wave }, "wave shape for ocean generation [0,1/pi]" },
  { "-wind", "%f", { &wind }, "wind speed for ocean generation" },
  { "-k0", "%f", { &k0 }, "wave scaling constant for ocean generation" },
  { "-kx", "%f", { &kx }, "wave scaling constant for ocean generation" },
  { "-kz", "%f", { &kz }, "wave scaling constant for ocean generation" },
  { "-grav", "%f", { &grav }, "gravity for ocean generation" },
  { "-merge", "%b", { &merge }, "merge ocean with landscape" },
  { "-tscale", "%f", { &tscale }, "texture scale" },
  { "-tlerp", "%b", { &tlerp }, "use linear interpolation for noise field" },
  { "-terpw", "%d", { &terpw }, "size of interpolation grid for noise field" },
  { "-ttype", "%d", { &text_type }, "texture type to generate" },
  { 0, 0, { 0 }, 0 }
};


extern "C" char* sys_errlist[];

// return system error message for given number
inline static const char* err_msg(int num)
{
  return sys_errlist[num];
}

// check to make sure given argument is a number
inline static int is_num_arg(char* arg)
{
  return (isdigit(*arg) || (*arg == '-' && isdigit(*(arg + 1))));
}

// see descriptions of options
int main(int argc, char* argv[])
{
  Real p, lastp = 0.0;    // for progress counters
  char* outfile = 0;      // output file
  MyField* h = 0;         // field being manipulated

  // parse options
  for(int i = 1; i < argc; i++) {

    // first check major options
    for(int j = 0; Options[j].name; j++) {
      if(!strcmp(argv[i], Options[j].name)) {
	flags |= Options[j].opt;
	break;
      }
    }

    // if not found check minor options
    if(!Options[j].name) {
      Boolean found = False;
      for(int j = 0; ValueOpts[j].name; j++) {
	if(!strcmp(argv[i], ValueOpts[j].name)) {
	  // parse option values
	  char* fmt = ValueOpts[j].format;
	  int pind = 0;
	  found = True;
	  while(*fmt) {
	    // integer option value
	    if(fmt[1] == 'd') {
	      int* valptr = (int*)ValueOpts[j].ptr[pind];
	      if(i < argc - 1 && is_num_arg(argv[i+1])) {
		*valptr = atoi(argv[++i]);
	      } else {
		cerr << "no value given for " << ValueOpts[j].name << endl;
	      }
	    // boolean option value
	    } else if(fmt[1] == 'b') {
	      int* valptr = (int*)ValueOpts[j].ptr[pind];
	      if(i < argc - 1 && is_num_arg(argv[i+1])) {
		*valptr = atoi(argv[++i]);
	      } else {
		*valptr = !*valptr;
	      }
	    // floating point option value
	    } else if(fmt[1] == 'f') {
	      Real* valptr = (Real*)ValueOpts[j].ptr[pind];
	      if(i < argc - 1 && is_num_arg(argv[i+1])) {
		sscanf(argv[++i], "%f", valptr);
	      } else {
		cerr << "no value given for " << ValueOpts[j].name << endl;
	      }
	    }
	    fmt += 2;
	    pind++;
	  }
	  break;
	}
      }
      // option is either unknown or output file name
      if(!found && *argv[i] == '-') {
	cout << "unknown option: " << argv[i] << " ignored" << endl;
      } else if(!ValueOpts[j].name && !outfile && *argv[i] != '-') {
	outfile = argv[i];
      }
    }
  }

#ifdef atarist
  _binmode(1);
#endif

  // print help strings
  if(flags & HELP) {
    for(int j = 0; Options[j].name; j++)
      cout << Options[j].name << "\t\t" << Options[j].help << endl;
    for(j = 0; ValueOpts[j].name; j++) {
      char* fmt = ValueOpts[j].format;
      int pind = 0;
      cout << ValueOpts[j].name << "\t\t" << ValueOpts[j].help << " (";
      while(*fmt) {
	if(fmt[1] == 'd') {
	  cout << *(int*)ValueOpts[j].ptr[pind];
	} else if(fmt[1] == 'b') {
	  cout << (*(Boolean*)ValueOpts[j].ptr[pind] ? "True" : "False");
	} else if(fmt[1] == 'f') {
	  cout << *(Real*)ValueOpts[j].ptr[pind];
	}
	fmt += 2;
	pind++;
	if(*fmt) cout << ", ";
      }
      cout << ")" << endl;
    }
    exit(0);
  }

  // set randomness, 0 = use system clock
  InitRand(seed);

#if 0
  double min, max;
  for(int l = 0; l < 300000; l++) {
    double tmp = ngrand(2.0);
    if(!l) min = max = tmp;
    if(min > tmp) min = tmp;
    if(max < tmp) max = tmp;
  }
  cout << "ngrand min/max: " << min << "/" << max << endl;
#endif

  // test trig tab
  if(flags & TEST_TRIG_TAB) {
    for(Real a = theta; a <= phi; a += 1.0) {
      cout << "sin(" << a << ") = " << fsin(deg2rad(a)) << endl;
      cout << "cos(" << a << ") = " << fcos(deg2rad(a)) << endl;
    }
  }

  // turn degrees into radians
  theta = deg2rad(theta);
  phi = deg2rad(phi);
  alpha = deg2rad(alpha);

  // adjust sizes by granularity
  width = (int)(width / granularity);
  height = (int)(height / granularity);
  depth = (int)(depth / granularity);

  // default to making a planet
  if(!(flags & (HMAP|PLANET|TEXTURE|TURBULENCE)))
    flags |= PLANET | COLORIZE;

  if(flags & HMAP) {
    cout << "Making " << width << " x " << height << " hmap" << endl;
    h = new MyField(width, height, 1, 1.0);
  } else if(flags & PLANET) {
    cout << "Making " << width << " x " << height << " planet, " << iter << " iterations" << endl;
    h = new Planet(width, height, 100.0, iter, 1.0, sea_level, 0);
  } else if(flags & TEXTURE) {
    Texture* t = 0;
    h = new MyField(width, height, depth, 1.0);

    if(text_type == MARBLE) {
      cout << "Making " << width << " x " << height << " marble texture" << endl;
      t = new MarbleTexture(width, height, depth);
    } else if(text_type == ELLIPSOID) {
      cout << "Making " << width << " x " << height << " x " << depth << " ellipsoidal texture" << endl;
      t = new EllipsoidalTexture;
      h->SetHeight(high);
    } else if(text_type == PLANAR) {
      cout << "Making " << width << " x " << height << " planar texture" << endl;
      t = new PlanarTexture;
    } else {
      cout << "no texture specified" << endl;
      delete h;
      return -1;
    }
    h->Texturize(*t);
    delete t;
  } else if(flags & TURBULENCE) {
    Real min, max, ave;
    cout << "Making " << width << " x " << height << " x ";
    cout << depth << " turbulence field" << endl;
    h = new MyField(width, height, 1, 1.0);

    lastp = 0.0;
    cout << "turbulence -" << flush;

    //Real dx = 5.0/width, dy = 5.0/height;
    Real dx = 0.10, dy = 0.10;

    for(int j = 0; j < h->Height(); j++) {
      for(int i = 0; i < h->Width(); i++) {
	h->Height(i, j) = Noise(i * dx, j * dy, 1.0);
	//h->Height(i, j) = Turbulence(h->Width(), i, j, 0.5);
	assert(!isnan(h->Height(i, j)));
      }
      p = (100.0 * j) / (Real)h->Height();
      if(p - lastp >= 10.0) {
	cout << " " << p << "%" << flush;
	lastp = p;
      }
    }
    cout << " 100%" << endl;
  } else {
    cout << "no field type specified!" << endl;
    return -1;
  }
  if(scale) h->ScaleHeights(scale);

  if(flags & REGULAR_GRID) {
    cout << "making regular grid" << endl;
    h->SetHeight(high);
    if(scale) h->ScaleHeights(scale);
  }

  if(flags & IRREGULAR_GRID) {
    if(valley) {
      for(int j = 0; j < h->Height(); j++) {
	for(int i = 0; i < h->Width(); i++) {
	  Point3d& p = (*h)(i, j);
	  p.z = (p.x + fabs(p.y - h->Width()/2))/15.0;
	}
      }
    } else if(sloped) {
      for(int j = 0; j < h->Height(); j++) {
	for(int i = 0; i < h->Width(); i++) {
	  Point3d& p = (*h)(i, j);
	  p.z = p.x/10.0;
	}
      }
    } else if(dsloped) {
      for(int j = 0; j < h->Height(); j++) {
	for(int i = 0; i < h->Width(); i++) {
	  Point3d& p = (*h)(i, j);
	  p.z = (p.x + p.y - h->Width()/2)/15.0;
	}
      }
    } else if(bumpy1) {
      for(int j = 0; j < h->Height(); j++) {
	for(int i = 0; i < h->Width(); i++) {
	  Point3d& p = (*h)(i, j);
	  if(p.x < h->Width()/2)
	    p.z = 2 * fsin(deg2rad(10*p.x));
	  else
	    p.z = -2;
	}
      }
    } else if(bumpy2) {
      for(int j = 0; j < h->Height(); j++) {
	for(int i = 0; i < h->Width(); i++) {
	  Point3d& p = (*h)(i, j);
	  if(p.x < 2.0*h->Width()/3.0)
	    p.z = 2*fsin(2*deg2rad(p.y+0.8*p.x))*fsin(deg2rad(10*p.x));
	  else
	    p.z = -2;
	}
      }
    } else if(bumpy3) {
      for(int j = 0; j < h->Height(); j++) {
	Real phi = M_PI * j / 20.0;
	for(int i = 0; i < h->Width(); i++) {
	  Point3d& p = (*h)(i, j);
	  Real rx = M_PI * (rrand(2.0) - 1.0) / 4.0;
	  Real ry = M_PI * (rrand(2.0) - 1.0) / 4.0;
	  Real theta = (M_PI * i) / 15.0;
	  p.z = 10.0 * fsin(theta+phi+rx) * fcos(phi+ry);
	}
      }
    } else if(drop) {
      for(int j = 0; j < h->Height(); j++) {
	for(int i = 0; i < h->Width(); i++) {
	  Point3d& p = (*h)(i, j);
	  if(p.x > (h->Width()/2 + p.y)/5.0)
	    p.z = 0.0;
	  else
	    p.z = 15.0;
	}
      }
    } else {
      cout << "no irregular grid type selected!" << endl;
      return -1;
    }
  }

  if(flags & FBM) {
    Real min, max, ave;
    if(sigma == 0.0) sigma = sqrt(h->Width());
    iter = (int)round(log(max(h->Width(), h->Height()))/log(2));
    cout << "performing fbm " << iter << " iterations" << endl;
    h->TwoDfbm(iter, sigma, H, True);
    h->MinMaxHeights(min, max);
    ave = h->AverageHeight();
    cout << "min/max/ave heights: " << min << ", " << max << ", " << ave << endl;
    if(scale) h->ScaleHeights(scale);
  }

  if(flags & SPECTRAL_SYNTHESIS) {
    cout << "performing spectral synthesis" << endl;
    h->SpectralSynthesis2d(H);
    if(scale) h->ScaleHeights(scale);
  }

  if(flags & SRAISE) {
    h->Raise(iter, "sphere");
    if(scale) h->ScaleHeights(scale);
  }

  if(flags & PRAISE) {
    h->Raise(iter, "pyramid");
    if(scale) h->ScaleHeights(scale);
  }

  if(flags & CRAISE) {
    h->Raise(iter, "cone");
    if(scale) h->ScaleHeights(scale);
  }

  if(flags & CRATERIZE) {
    h->Craterize(citer, dpower, dbias);
    if(scale) h->ScaleHeights(scale);
  }

  if(flags & EROSION) {
    MyField hnew(*h, True, True);
    hnew.Erode(titer, talus, tperc, twind, alpha, kc, kd, ks, rain, (int)((1.0 - freq)*titer), 0.0);
    h->Print(cout, 1); cout << endl;
    hnew.Print(cout, 1); cout << endl;
    MyField* tmp = new MyField(*h, hnew);
    delete h;
    h = tmp;
    if(scale) h->ScaleHeights(scale);
  }

  if(flags & THERMAL_EROSION) {
    MyField hnew(*h, True, True);
    hnew.T_Erode(talus, tperc, titer, twind, alpha);
    h->Print(cout, 1); cout << endl;
    hnew.Print(cout, 1); cout << endl;
    MyField* tmp = new MyField(*h, hnew);
    delete h;
    h = tmp;
    if(scale) h->ScaleHeights(scale);
  }

  if(flags & HYDRAULIC_EROSION) {
    MyField hnew(*h, True, True);
    hnew.H_Erode(kc, kd, ks, rain, hiter, (int)((1.0 - freq)*hiter));
    h->Print(cout, 1); cout << endl;
    hnew.Print(cout, 1); cout << endl;
    MyField* tmp = new MyField(*h, hnew);
    delete h;
    h = tmp;
    if(scale) h->ScaleHeights(scale);
  }

  if(flags & WATER) {
    cout << "raining on " << width << " x " << height << " field" << endl;
    MyField* water = h->Water(0.0, 0.008, evap, rain, -1, 0.05);

    water->IntensityMap(0.3);
//    water->SaveGif("watermap.gif");

    MyField* both = new MyField(*h, *water);

    TerrainPalette tp(0.0);
    if(dark != 0.0) tp.Darken(dark);
    if(bright != 0.0) tp.Brighten(bright);
    if(cscale != 0.0) tp.Scale(cscale);
    both->Colorize(tp);
    for(i = h->Width(); i < both->Width(); i++) {
      for(int j = 0; j < both->Height(); j++) {
	Real I = water->color(i-h->Width(), j).Intensity();
	both->color(i, j) = RGB(I, I, I);
      }
    }

    delete water;
    delete h;
    h = both;
    quant = True;
  }

  if(flags & OCEAN) {
    cout << "Making " << width << " x " << height << " ocean at time " << otime << endl;
    Ocean o(*h, wind, wave, sea_level, k0, kx, kz, grav);
    cout << o << endl;
    Real crap;
    MyField* tmp = o.Compute(otime, alpha, crap, merge);
    delete h;
    h = tmp;
    cout << "newh " << h->AverageHeight() << endl;
    cout << "new sea level = " << sea_level << endl;
    if(scale) h->ScaleHeights(scale);
  }

  if(flags & ROTATE) {
    cout << "rotating by " << rad2deg(theta) << " x " << rad2deg(phi);
    cout << " x " << rad2deg(alpha) << endl;
    h->Rotate(theta, phi, alpha);
  }

  if(flags & SMOOTH) {
    cout << "smoothing hmap" << endl;
    h->Smooth(2.0);
    cout << "done smoothing hmap" << endl;
    if(scale) h->ScaleHeights(scale);
  }

  if(flags & EDGE_DETECT) {
    h->EdgeDetect(-1.0);
  }

  if(flags & COLORIZE) {
    TerrainPalette tp(sea_level);
    if(dark != 0.0) tp.Darken(dark);
    if(bright != 0.0) tp.Brighten(bright);
    if(cscale != 0.0) tp.Scale(cscale);
    h->Colorize(tp);
  }

  if(flags & GRAY_SCALE) {
    h->IntensityMap(imin, imax);
  }

  if(flags & INFO) {
    Real min, max;
    h->MinMaxHeights(min, max);
    cout << *h << endl;
    cout << "min/max heights: " << min << "/" << max << endl;
    cout << "average height: " << h->AverageHeight() << endl;
  }

  if(outfile) {
    cout << "saving hmap to " << outfile << endl;
    if(!h->IsColored()) {
      TerrainPalette tp(sea_level);
      if(dark != 0.0) tp.Darken(dark);
      if(bright != 0.0) tp.Brighten(bright);
      if(cscale != 0.0) tp.Scale(cscale);
      h->Colorize(tp);
    }
    if(gif) {
      if(!h->SaveGif(outfile, quant))
	cerr << "error saving to " << outfile << ": " << err_msg(errno) << endl;
    } else {
      if(!h->Save(outfile))
	cerr << "error saving to " << outfile << ": " << err_msg(errno) << endl;
    }
  }

  delete h;

  return 0;
}
