#include "noise.h"

static const double Epsilon = 0.0001;

inline static double hpoly1(double t) { return ((2.0*t-3.0)*t*t+1.0); }
inline static double hpoly2(double t) { return (-2.0*t+3.0)*t*t; }
inline static double hpoly3(double t) { return ((t-2.0)*t+1.0)*t; }
inline static double hpoly4(double t) { return (t-1.0)*t*t; }

inline static double frand(long s)
{
  s = s << 13 ^ s;
  return 1.0-((s*(s*s*15731+789221)+1376312589)&0x7fffffff)/1073741824.0;
}

inline static double frand3(long x, long y, long z)
{
  return frand(17*x + 23*y + 29*z);
}

inline static double rand3a(long x, long y, long z)
{
  return frand(67 * x + 59 * y + 71 * z);
}

inline static double rand3b(long x, long y, long z)
{
  return frand(73 * x + 79 * y + 83 * z);
}

inline static double rand3c(long x, long y, long z)
{
  return frand(89 * x + 97 * y + 101 * z);
}

inline static double rand3d(long x, long y, long z)
{
  return frand(103 * x + 107 * y + 109 * z);
}

static double delta[3];
static long xlim[3][2];

// recursive interpolation function for perlin noise function
static void Interpolate(double* f, int i, int n)
{
  double f0[4];
  double f1[4];

  // if we are at an integer latice point generate a random value
  if(n == 0) {
    int x = i&1, y = i>>1&1, z = i>>2;
    f[0] = rand3a(xlim[0][x], xlim[1][y], xlim[2][z]);
    f[1] = rand3b(xlim[0][x], xlim[1][y], xlim[2][z]);
    f[2] = rand3c(xlim[0][x], xlim[1][y], xlim[2][z]);
    f[3] = rand3d(xlim[0][x], xlim[1][y], xlim[2][z]);
  } else {  // otherwise interpolate between adjacent noise values
    n--;
    Interpolate(f0, i, n);
    Interpolate(f1, i | 1<<n, n);

    // use linear interpolation for a, b, c
    double dn = delta[n];
    double hp1 = hpoly1(dn);
    double hp2 = hpoly2(dn);
	    
    f[0] = f0[0]*hp1 + f1[0]*hp2;
    f[1] = f0[1]*hp1 + f1[1]*hp2;
    f[2] = f0[2]*hp1 + f1[2]*hp2;
    f[3] = f0[3]*hp1 + f1[3]*hp2 + f0[n]*hpoly3(dn) + f1[n]*hpoly4(dn);
  }
}

// compute noise value at x, y, z (from graphics gems II)
Real Noise(Real x, Real y, Real z)
{
  double f[4];

  xlim[0][0] = (long)floor(x); // next and previous integer x coords
  xlim[0][1] = xlim[0][0] + 1;

  xlim[1][0] = (long)floor(y); // next and previous integer x coords
  xlim[1][1] = xlim[1][0] + 1;

  xlim[2][0] = (long)floor(z); // next and previous integer x coords
  xlim[2][1] = xlim[2][0] + 1;

  delta[0] = x - xlim[0][0];   // dx
  delta[1] = y - xlim[1][0];   // dy
  delta[2] = z - xlim[2][0];   // dz

  Interpolate(f, 0, 3);

//  return f.Dp(x, y, z);
  //return f[0]*x + f[1]*y + f[2]*z + f[3];
  return f[3];
}

/* compute fractal noise function */
Real FractalNoise(Real x, Real y, Real z)
{
  long  t[3], v[3], beg[3];
  double  fval[8], fc;
  int  branch;
  register long  s;
  register int  i, j;
  
  /* get starting cube */
  s = (long)(1.0/Epsilon);
  
  t[0] = (long)(s * x);
  t[1] = (long)(s * y);
  t[2] = (long)(s * z);
  beg[0] = (long)(s * floor(x));
  beg[1] = (long)(s * floor(y));
  beg[2] = (long)(s * floor(z));
  
  for (j = 0; j < 8; j++) {
    for (i = 0; i < 3; i++) {
      v[i] = beg[i];
      if (j & 1<<i)
	v[i] += s;
    }
    fval[j] = frand3(v[0],v[1],v[2]);
  }
  /* compute fractal */
  for ( ; ; ) {
    fc = 0.0;
    for (j = 0; j < 8; j++)
      fc += fval[j];
    fc *= 0.125;
    if ((s >>= 1) == 0)
      return(fc);		/* close enough */
    branch = 0;
    for (i = 0; i < 3; i++) {	/* do center */
      v[i] = beg[i] + s;
      if (t[i] > v[i]) {
	branch |= 1<<i;
      }
    }
    fc += s*Epsilon*frand3(v[0],v[1],v[2]);
    fval[~branch & 7] = fc;
    for (i = 0; i < 3; i++) {	/* do faces */
      if (branch & 1<<i)
	v[i] += s;
      else
	v[i] -= s;
      fc = 0.0;
      for (j = 0; j < 8; j++)
	if (~(j^branch) & 1<<i)
	  fc += fval[j];
      fc = 0.25*fc + s*Epsilon*frand3(v[0],v[1],v[2]);
      fval[~(branch^1<<i) & 7] = fc;
      v[i] = beg[i] + s;
    }
    for (i = 0; i < 3; i++) {	/* do edges */
      j = (i+1)%3;
      if (branch & 1<<j)
	v[j] += s;
      else
	v[j] -= s;
      j = (i+2)%3;
      if (branch & 1<<j)
	v[j] += s;
      else
	v[j] -= s;
      fc = fval[branch & ~(1<<i)];
      fc += fval[branch | 1<<i];
      fc = 0.5*fc + s*Epsilon*frand3(v[0],v[1],v[2]);
      fval[branch^1<<i] = fc;
      j = (i+1)%3;
      v[j] = beg[j] + s;
      j = (i+2)%3;
      v[j] = beg[j] + s;
    }
    for (i = 0; i < 3; i++)		/* new cube */
      if (branch & 1<<i)
	beg[i] += s;
  }
}


// turbulence function (from Foley and Van Dam, Chapter 20)
Real Turbulence(Real x, Real y, Real z, int n)
{
  Real turb = 0.0;
  Real s = 1.0;
  while(n > 0) {
    turb += fabs(s * Noise(x/s, y/s, z/s));
    s /= 2;
    n >>= 1;
  }
  return turb;
}
