#include "ocean.h"
#include "trig.h"

// init - randomly perturb heights according to paper
// allow change in x direction (rotation)
// add effects of cumulative wave trains (just average??)

// for each y strip precompute Sx, Sz, etc...?

Point3d Ocean::Rotate(Real x, Real y, Real z) const
{
  x -= _xc;
  y -= _yc;
//  return Point3d(x*_ncos - y*_nsin + _xc, x*_nsin + y*_ncos + _yc, z);
  return Point3d(x*_cos - y*_sin + _xc, x*_sin + y*_cos + _yc, z);
}

Point3d& Ocean::Rpoint(int i, int j, Real ang) const
{
  if(ang == 0.0) {
    //return (*this)(fixrng(i, 0, Width()), fixrng(j, 0, Height()));
    return (*this)(i, j);
  } else {
    Real x = i - _xc;
    Real y = j - _yc;
    int ni = (int)round(x*_cos - y*_sin + _xc);
    int nj = (int)round(x*_sin + y*_cos + _yc);
    return (*this)(fixrng(ni, 0, Width()), fixrng(nj, 0, Height()));
  }
}

Point3d& Ocean::Shore(int i, int j, Real ang) const
{
  if(ang == 0.0) {
    //return _shore(fixrng(i, 0, Width()), fixrng(j, 0, Height()));
    return _shore(i, j);
  } else {
    Real x = i - _xc;
    Real y = j - _yc;
    int ni = (int)round(x*_cos - y*_sin + _xc);
    int nj = (int)round(x*_sin + y*_cos + _yc);
    return _shore(fixrng(ni, 0, Width()), fixrng(nj, 0, Height()));
  }
}


// compute Phi angle for wave equation
Real Ocean::Phi(int i, int j, Real t, Real ang) const
{
  Point3d lastp = Rpoint(0, j, ang);
  Real phi = -_w * t;
  for(int x = 0; x <= i; x++) {
    Point3d op = Rpoint(x, j, ang);
    Point3d sp = Shore(x, j, ang);
    Real depth = op.z - sp.z;
    if(depth != 0.0) {
      Real kinf = _K*tanh(_K * depth);
      phi += (kinf * op.x - lastp.x) / sqrt(tanh(kinf * depth));
    }
    lastp = op;
  }
  return phi;
}

// compute ocean new coordinates at time t
Point3d Ocean::NewCoords(int i, int j, Real t, Real ang) const
{
  Point3d p = (*this)(i, j);
  Real depth = p.z - _shore(i, j).z;
  Point3d np, sp;

  if(ang == 0.0 && i == Width() - 1) {
    np = p;
    sp = p;
  } else {
    np = Rpoint(i+1, j, ang);
    sp = Shore(i+1, j, ang);
  }

  Real ndepth = np.z - sp.z;
  Real alpha = (np.x != p.x) ?  Alpha(depth, ndepth, np.x - p.x) : 0.0;
  Real phi = Phi(i, j, t, ang);
  Real sx = Sx(depth);
  Real sz = Sz(sx, depth);

  Real cosp = fcos(phi);
  Real cosa = fcos(alpha);
  Real sinp = fsin(phi);
  Real sina = fsin(alpha);

  Real x = p.x + _r * sx * cosa * sinp + sz * sina * cosp;
  Real y = p.y;
  Real z = p.z - _r * sz * cosa * cosp + sx * sina * sinp;

  return Rotate(x, y, z);
}


// compute new ocean at time t
MyField* Ocean::Compute(Real t, Real ang, Real& sea, Boolean merge)
{
  MyField* ocean = new MyField(Width(), Height());
  PercCounter pc("ocean", Size());
  Real zmax = AverageHeight();
  Real minh, maxh;
  int k = 0;

  _cos = fcos(ang);
  _sin = fsin(ang);
  _ncos = fcos(-ang);
  _nsin = fsin(-ang);

  ocean->CopyData(*this);
  for(int j = 0; j < Height(); j++) {
    for(int i = 0; i < Width(); i++, k++, pc++) {
      if(Height(i, j) - _shore.Height(i, j) > 0.0) {
	(*ocean)(k) = NewCoords(i, j, t, ang);
	if((*ocean)(k).z > zmax) zmax = (*ocean)(k).z;
      }
    }
  }
  ocean->MinMaxHeights(minh, maxh);
  sea = zmax / (maxh - minh);

  return ocean;
}

inline static Real amplitude(double x, Boolean rnd = False)
{
  double ang = x * 180.0 - 90.0;
  if(rnd) ang += irand(100) - 40;
  Real tmp = fcos(deg2rad(ang));
  if(tmp < 0.0) tmp = 0;
  return sqrt(sqrt(tmp)) - 0.5;
}

inline static int fact(int n)
{
  if(n == 0)
    return 1;
  else {
    int f = n;
    while(--n) f *= n;
    return f;
  }
}


// mean > 0
inline static void MakeDip(MyField& f, Real mean,
			   Real scale, int i, int j,
			   int rad)
{
  double mexp = scale*exp(-mean)/2.0;
  Real maxz = 0.0;
  for(int x = 0; x < rad; x++) {
    for(int y = 0; y < rad; y++) {
      int px = (int)round(sqrt(x*x + y*y));
      double dz = (px < 13) ? mexp*pow(mean, px)/fact(px) : 0;
      if(i - x >= 0) {
	if(j - y >= 0) f.Height(i - x, j - y) -= dz;
	if(j + y < f.Height()) f.Height(i - x, j + y) -= dz;
      } else if(i + x < f.Width()) {
	if(j - y >= 0) f.Height(i + x, j - y) -= dz;
	if(j + y < f.Height()) f.Height(i + x, j + y) -= dz;
      }
      if(dz > maxz) maxz = dz;
    }
  }
#ifdef DEBUG
//  cout << "Hole: (" << i << ", " << j << "), rad = " << rad;
//  cout << ", zmax = " << maxz << endl << flush;
#endif
}


// initialize constants and ocean field
void Ocean::Init()
{
  Real minh, maxh;
  _w = w();
  _K = K();
  _r = R();
  _xc = Width()/2.0;
  _yc = Height()/2.0;

  // init sea height data
  _shore.MinMaxHeights(minh, maxh);
  _sea_height = (_sea_level * (maxh - minh)) + minh;

  CopyData(_shore);  // to get same x & y coords!

#if 0
  SetHeight(_sea_height);
#else
  double dx = 1.0 / Width();
  double dy = 1.0 / Height();
  for(int j = 0; j < Height(); j++) {
    Real ya = amplitude(j * dy);
    for(int i = 0; i < Width(); i++) {
      Real depth = _sea_height - _shore(i, j).z;
      Real sz = depth > 0.0 ? Sz(depth) : 1.0;
      Real xa = amplitude(i * dx, True);
      //Height(i, j) = _sea_height + (2.0 * _r * ya * xa * sz) - _r;
      Height(i, j) = _sea_height + (2.0 * _r * ya * xa) - _r;
    }
  }
#endif

  // put some "holes" into the ocean to break up crests
  // diameter is on the order of the wavelength
  // and depth has a poisson distribution
  // P(x) = exp(-mean) * (mean^x) / fact(x)
  int _num_holes = (int)(sqrt(Width()*Height()));
  for(int n = 0; n < _num_holes; n++) {
    int rad = irand((int)(3*L())) + 2;
    MakeDip(*this, 0.8, 2*_r, irand(Width()), irand(Height()), rad);
  }
#ifdef DEBUG
  cout << "shore min/max " << minh << ", " << maxh << endl;
  cout << "sea level: " << _sea_height << endl;
#endif
}


// create new ocean
Ocean::Ocean(const MyField& shore, Real V, Real delta,
	     Real sea_level, Real K0, Real Kx, Real Kz,
	     Real g)
: MyField(shore.Width(), shore.Height()), _shore(shore),
  _delta(delta), _V(V), _sea_level(sea_level), _K0(K0),
  _Kx(Kx), _Kz(Kz), _g(g)
{
  Init();
}

ostream& Ocean::Print(ostream& os, int) const
{
  MyField::Print(os, 1);
  os << endl;
  os << "wave shape = " << _delta << endl;
  os << "wind = " << _V << " m/s" << endl;
  os << "wave number (K) = " << _K << " rad/m" << endl;
  os << "K0/Kx/Kz (rad/m) = " << _K0 << "/" << _Kx << "/" << _Kz << endl;
  os << "gravity = " << _g << " m/s^2" << endl;
  os << "angular speed (w) = " << _w << " rad/sec" << endl;
  os << "wave radius (r) = " << _r << " m" << endl;
  os << "wavelength (L) = " << L() << " m" << endl;
  os << "wave period (T) = " << T() << " s" << endl;
  return os;
}
