////////////////////////////////////////////////////////////////////////////
//
// ocean.h - simulated ocean class (see A Simple Model of Ocean Waves
//           in references for more info)
//
////////////////////////////////////////////////////////////////////////////
#ifndef _OCEAN_H_
#define _OCEAN_H_

#include "hmap.h"

// Ocean class
class Ocean : public MyField {
  DeclareStandardMethods(Ocean, MyField);
 public:
  Point3d& Rpoint(int i, int j, Real ang) const;
  Point3d& Shore(int i, int j, Real ang) const;
  Real Depth(int i, int j, Real ang) const
    { return Rpoint(i,j,ang).z - Shore(i,j,ang).z; }

  // wave shape, ranges between 0 and 1/pi, default 0.1
  // the higher the value the higher the wave peaks
  Real& Shape() const { return ((Ocean*)this)->_delta; }

  // wind speed in m/s
  Real& WindSpeed() const { return ((Ocean*)this)->_V; }

  // ocean depth field
  const MyField& Shore() const { return ((Ocean*)this)->_shore; }

  // sea level of ocean (with respect to shore)
  Real& SeaLevel() const { return ((Ocean*)this)->_sea_level; }

  // wave shape constants
  Real& K0() const { return ((Ocean*)this)->_K0; }
  Real& Kx() const { return ((Ocean*)this)->_Kx; }
  Real& Kz() const { return ((Ocean*)this)->_Kz; }

  // gravitational constant (usually 9.8 m/s^2)
  Real& Gravity() const { return ((Ocean*)this)->_g; }

  // compute new ocean field at given time, return new field
  // which contains the new ocean (optionally merged with the
  // shore field).  also set sea to be the new maximum sea height
  MyField* Compute(Real t, Real ang, Real& sea, Boolean = True);

  // create a new ocean with the given shore and parameters
  // V = wind velocity in m/s, delta = wave shape, sea = height of sea
  // as % of shore height, K0, Kx, Kz = wave shape constants, g = gravity
  Ocean(const MyField& shore, Real V, Real delta = 0.1, Real sea = 0.5,
	Real K0 = 1.0, Real Kx = 1.0, Real Kz = 1.0, Real g = 9.8);
 protected:
  // compute angular wave speed
  Real w() const { return (sqrt(2.0/3.0) * _g) / _V; }

  // wave period
  Real T() const { return (2.0 * M_PI) / w(); }

  // wave wavelength
  Real L() const { return (_g * sqr(T())) / (2.0 * M_PI); }

  // wave number
  Real K() const { return (2.0 * M_PI) / L(); }

  // radius of wave orbit
  Real R() const { return (_delta * L()) / 2.0; }

  // x scale factor
  Real Sx(Real depth) const { return 1.0 / (1.0 - exp(-_Kx * depth)); }

  // z scale factor
  Real Sz(Real depth) const { return Sx(depth) * (1.0 - exp(-_Kz * depth)); }
  Real Sz(Real sx, Real depth) const { return sx * (1.0 - exp(-_Kz * depth)); }

  // alpha angle for wave equation
  Real Alpha(Real d1, Real d2, Real dx) const
    { return ((d2 - d1) * exp(-_K0 * d1)) / dx; }

  // phi angle for wave equation
  Real Phi(int i, int j, Real t, Real ang) const;

  // compute new x & z coords for given point at time t
  Point3d NewCoords(int i, int j, Real t, Real ang) const;

  // initialize some constants and heights
  void Init();
 protected:
  Point3d Rotate(Real x, Real y, Real z) const;
 private:
  // parameters
  const MyField& _shore; // depth of shore meters
  Real _delta;           // wave shape range = [0 - 1/pi], ave = 0.1
  Real _V;               // wind velocity in m/s
  Real _sea_level;       // as percent of height of _shore
  Real _sea_height;      // average sea height
  Real _K0;              // scaling constant
  Real _Kx;              // scaling constant
  Real _Kz;              // scaling constant
  Real _g;               // gravity in m/s^2, normal = 9.8 m/s^2
  // computed constants
  Real _w;               // angular speed     radians/sec
  Real _K;               // wave number       radians/meter
  Real _r;               // radius of orbit   meters
  Real _xc;              // x center of ocean field
  Real _yc;              // y center of ocean field
  Real _cos;             // cos of wavefront angle
  Real _sin;             // sin of wavefront angle
  Real _ncos;            // cos of negative wavefront angle
  Real _nsin;            // sin of negative wavefront angle
};

#endif
