#include "planet.h"
#include "trig.h"

// returns the sphere defined by planet
// (normally coordinates are stored as a 2d height map)
Point3d* Planet::Sphere() const
{
  Point3d* sp = new Point3d[Size()];
  for(int i = 0; i < Size(); i++)
    sp[i] = SpCoord(i);
  return sp;
}

// return the ith coordinate of the planet as a spherical coordinate
SphericalPoint Planet::SpCoord(int i) const
{
  Point3d p = (*this)(i);
  return SphericalPoint((deg2rad(360.0) * p.x)/Width(),
			(deg2rad(180.0) * p.y)/Height(),
			p.z);
}

// return the ith by jth coordinate of the planet as a spherical coordinate
SphericalPoint Planet::SpCoord(int i, int j) const
{
  Point3d p = (*this)(i, j);
  return SphericalPoint((deg2rad(360.0) * p.x)/Width(),
			(deg2rad(180.0) * p.y)/Height(),
			p.z);
}


// evolve the planet a certain number of iterations.
// works by generating "faults" and raising or lowering land on either
// side of the fault
// Note: setup time is slow, so it pays to do alot of iterations at once
void Planet::Evolve(int iter)
{
  PercCounter pc("evolve", iter);
  Vector3d norm, v1, v2;
  Real* heights = Heights();    // get the heights since this is
                                // what we will be modifying
  Point3d* sphere = Sphere();   // get 3d coordinates of sphere
  register Point3d* pptr;
  register Real* pd;
  register Real ni, nj, nk;

  for(int n = 0; n < iter; n++, pc++) {
    // pick two points to define "fault"
#if 1
    int x1 = irand(Width());
    int y1 = irand(Height());
    int x2 = irand(Width());
    int y2 = irand(Height());
#else
    int x1 = (int)ngrand(Width());
    int y1 = (int)ngrand(Height());
    int x2 = (int)ngrand(Width());
    int y2 = (int)ngrand(Height());
#endif

    // compute plane that defines fault
    v1 = sphere[x1 + y1 * Width()];
    v2 = sphere[x2 + y2 * Width()];
    norm = v1.Cross(v2).Normalize();
    ni = norm.i; nj = norm.j; nk = norm.k;
    pd = heights;
    pptr = sphere;

    // raise land on one side of plane and lower land
    // on the other side
    // (multiplying nk by pptr->z probably unecessary)
    for(int i = Size() - 1; i >= 0; i--, pptr++) {
      if((pptr->x * ni + pptr->y * nj) > pptr->z * nk)
	*pd++ += _delta;
      else
	*pd++ -= _delta;
    }
  }

  // adjust age and set new heights
  _age += iter;
  SetHeights(heights);

  // delete tmp data
  delete[] heights;
  delete[] sphere;
}

void Planet::Cap(Real cap)
{
  if(cap != 0.0) {
    Real H = 0.25, D = 2.5;

//    char* Hs = getenv("H");
//    char* Ds = getenv("D");
//    if(Hs) sscanf(Hs, "%f", &H);
//    if(Ds) sscanf(Ds, "%f", &D);

#ifdef DEBUG
    cout << "capping: " << H << ", " << D << endl;
#endif

    Real maxh = cap * Height();
    Real minh = maxh / D;
    Real min, max;

    MinMaxHeights(min, max);
#if 0
    for(int i = 0; i < Width(); i++) {
      int cl = (int)round(maxh + rrand(minh/2.0));
      for(int j = 0; j < cl; j++)
	Height(i, j) = max;
      cl = (int)round(maxh + rrand(minh/2.0));
      for(j = Height() - cl; j < Height(); j++)
	Height(i, j) = max;
    }
#else
    // can use multiple lines so thing is as long as needed
    MyField poles(128, 128, 1);

    poles.SpectralSynthesis2d(H);

    int nstart = poles.Size()/3;
    int sstart = 2*poles.Size()/3;
    int nend = nstart + Width();
    int send = sstart + Width();

    for(int i = 0; i < poles.Size(); i++) {
      if(i < nstart || i > send || (i > nend && i < sstart))
	poles.Height(i) = 0.0;
    }
    poles.ScaleHeights(maxh - minh);

    for(i = 0; i < Width(); i++) {
      int nh = (int)(poles.Height(nstart + i) + minh);
      int sh = (int)(poles.Height(sstart + i) + minh);
      for(int j = 0; j < nh; j++)
	Height(i, j) = max;
      for(j = 0; j < sh; j++)
	Height(i, Height() - j - 1) = max;
    }
#endif
  }
}

// seed planet with pseudo continents (just a bunch of rectangles)
void Planet::Seed(int seed)
{
  for(int i = 0; i < seed; i++) {
    int tstart = irand(Width());
    int pstart = irand(Height());
    int tlength = irand(Width()/8) + 4;
    int plength = irand(Height()/8) + 4;
    for(int k = 0; k < tlength; k++) {
      int x = tstart + k;
      if(x >= Width()) x -= Width();
      for(int j = 0; j < plength; j++) {
	int y = pstart + j;
	if(y >= Height()) y -= Height();
	(*this)(x, y).z += _delta;
      }
    }
  }
}

// initialize heights to radius
void Planet::SetRadius(Real radius)
{
  for(int i = 0; i < Size(); i++)
    Height(i) = radius;
}

// load planet from a PNM
int Planet::Load(char* file, const Palette* p)
{
  int res = Load(file, p);
  _age = 0;
  _radius = AverageHeight();
  _sea_level = 0.50;
  return res;
}

// create empty planet
Planet::Planet()
: MyField(), _age(0), _radius(0), _sea_level(0), _delta(0)
{
}

// copy a planet
Planet::Planet(const Planet& p, Boolean copy, Boolean own)
: MyField(p, copy, own), _age(p._age), _radius(p._radius),
  _sea_level(p._sea_level), _delta(p._delta)
{
}

// create planet using f as initial data
Planet::Planet(const MyField& f, Real rad,
	       int iter, Real delta, Real sea,
	       int seed, Real cap, Boolean copy,
	       Boolean own)
: MyField(f, copy, own), _age(0), _radius(f.AverageHeight()+rad),
  _sea_level(sea), _delta(delta)
{
  // delta is proportional to # of iterations and height of
  // perturbation as a percent of the radius
  // (i.e. we want the heights of the planet to be only about 5%
  //  above and below the mean level)
  if(_delta == 0.0) _delta = 2.0 * (_radius * 0.05) / iter;

  // modify initial field by the radius
  for(int i = 0; i < Size(); i++)
    Height(i) += rad;

  // seed and evolve
  Seed(seed);
  Evolve(iter);
  Cap(cap);
}

// create a new planet of the given size
Planet::Planet(int w, int h, Real rad, int iter,
	       Real delta, Real sea, int sd,
	       Real cap)
: MyField(w, h), _age(0), _radius(rad), _sea_level(sea), _delta(delta)
{
  // see comments from previous ctor
  if(_delta == 0.0) _delta = 2.0 * (_radius * 0.05) / iter;
  SetRadius(rad);
  Seed(sd);
  Evolve(iter);
  Cap(cap);
}

// create a new planet of the given granularity
// (see previous ctors and planet.h file)
Planet::Planet(Real gran, Real rad, int iter, Real delta,
	       Real sea, int seed, Real cap)
: MyField((int)(360.0/gran), (int)(180.0/gran), 1, gran),
  _age(0), _radius(rad), _sea_level(sea), _delta(delta)
{
  if(_delta == 0.0) _delta = 2.0 * (_radius * 0.05) / iter;
  SetRadius(rad);
  Seed(seed);
  Evolve(iter);
  Cap(cap);
}
