////////////////////////////////////////////////////////////////////////////
//
// planet.h - fractal planet class.  Planets are created by randomly
//            generating faults (great circles on a sphere).  At each
//            fault the land to one side is raised a certain amount
//            and the land on the other side is lowered by the same
//            amount.  After a few hundred iterations it starts to
//            look like a planet!
//
////////////////////////////////////////////////////////////////////////////
#ifndef _PLANET_H_
#define _PLANET_H_

#include "hmap.h"

// Planet class
class Planet : public MyField {
  DeclareStandardMethodsNoPrint(Planet, MyField);
 public:
  // return the coordinates of the sphere that the planet defines
  Point3d* Sphere() const;

  // return the ith spherical coordinate of the planet
  SphericalPoint SpCoord(int i) const;

  // return the ith by jth spherical coordinate of the planet
  SphericalPoint SpCoord(int i, int j) const;

  // return the age of the planet (i.e. # of iterations)
  int Age() const { return _age; }

  // planet radius
  Real Radius() const { return _radius; }

  // planet's sea level (unused at the moment)
  Real SeaLevel() const { return _sea_level; }

  // evolve the planet by given number of iterations
  void Evolve(int iter = 1);
  void Cap(Real cap = 0.0);

  // load a planet from a PNM
  int Load(char*, const Palette* = 0);

  // create an empty planet
  Planet();

  // copy a planet
  Planet(const Planet&, Boolean = False, Boolean = False);

  // create a new planet by copying original data from given field
  // the boolean arguments control whether the original data is copied
  // or referenced from the input field.
  // The seed argument controls the number of pseudo continents that
  // the planet's surface is initialized with
  Planet(const MyField&, Real rad, int iter, Real delta = 0.0,
	 Real sea = 0.5, int seed = 0, Real cap = 0.0,
	 Boolean = False, Boolean = False);

  // create a new planet of the given size
  // The seed argument controls the number of pseudo continents that
  // the planet's surface is initialized with
  Planet(int w, int h, Real rad, int iter,
	 Real delta=0.0, Real sea=0.5,
	 int seed=0, Real cap = 0.0);

  // create a new planet of the given granularity.
  // The relation between granularity and size is:
  //   width = 360/gran, height = width/2;
  // The seed argument controls the number of pseudo continents that
  // the planet's surface is initialized with
  Planet(Real gran, Real rad, int iter, Real delta=0.0,
	 Real sea=0.5, int seed=0, Real cap = 0.0);
 protected:
  void Seed(int);        // seed a planet with given number of "continents"
  void SetRadius(Real);  // set the radius of a planet
 private:
  int _age;         // age of planet (# of iterations evolved)
  Real _radius;     // radius of planet
  Real _sea_level;  // sea level of planet (not used)
  Real _delta;      // raise/lower amount at faults
};

#endif
