#include "texture.h"
#include "trig.h"
#include "noise.h"

// evaulate functional texture at given coordinates
// using pseudo FFT procedure
Real FunctionalTexture::operator()(Real x, Real y, Real z) const
{
  Real sum1 = 0.0;
  Real sum2 = 0.0;
  for(int i = 1; i < _size; i++) {
    sum1 += _Ci[i] * fsin(_fx[i] * x + _px(*this, i, x, y, z)) + _t0;
    sum2 += _Ci[i] * fsin(_fy[i] * y + _py(*this, i, x, y, z)) + _t0;
  }
  return _k * sum1 * sum2;
}

// create functional texture
FunctionalTexture::FunctionalTexture(TextureFunc tx, TextureFunc ty, int n,
				     Real C0, Real fx0, Real fy0, Real t0)
: _px(tx), _py(ty), _size(n+1), _Ci(new Real[n+1]),
  _fx(new Real[n+1]), _fy(new Real[n+1]), _t0(t0),
  _k(1.0)
{
  assert(tx && ty);

  // initialize constant arrays
  _Ci[0] = C0;
  _fx[0] = fx0;
  _fy[0] = fy0;
  for(int i = 1; i < _size; i++) {
    _Ci[i] = _Ci[i-1] * sqrt(2.0) / 2.0;
    _fx[i] = 2.0 * _fx[i-1];
    _fy[i] = 2.0 * _fy[i-1];
  }

#if 0
  Real min = RealMax;
  Real max = -RealMax;
  Real dx = 1.0/100.0;
  Real dy = 1.0/100.0;
  Real dz = 1.0/200.0;
  for(Real x = 0.0; x <= 1.0; x += dx) {
    for(Real y = 0.0; y <= 1.0; y += dy) {
      for(Real z = 0.0; z <= 1.0; z += dz) {
	Real sum1 = 0.0;
	Real sum2 = 0.0;
	for(i = 1; i < _size; i++) {
	  sum1 += _Ci[i] * fsin(_fx[i] * x + _px(*this, i, x, y, z)) + _t0;
	  sum2 += _Ci[i] * fsin(_fy[i] * y + _py(*this, i, x, y, z)) + _t0;
	}
	Real tot = sum1 * sum2;
	if(min > tot) min = tot;
	if(max < tot) max = tot;
      }
    }
    cout << "*" << flush;
  }
  cout.precision(10);
  cout << "min/max = " << min << "/" << max << endl;
#endif
}


// planar texure functions
inline static Real planar_x(const FunctionalTexture& t, int i, Real, Real y, Real)
{
  return 0.5 * M_PI * fsin(0.5 * t.fy(i) * y);
}

inline static Real planar_y(const FunctionalTexture& t, int i, Real x, Real, Real)
{
  return 0.5 * M_PI * fsin(0.5 * t.fx(i) * x);
}

// make planar texture
PlanarTexture::PlanarTexture(int n, Real C0, Real fx0, Real fy0, Real t0)
: FunctionalTexture(planar_x, planar_y, n, C0, fx0, fy0, t0)
{
}


// ellipsoidal texture functions
static Real ellipsoidal_x(const FunctionalTexture& t, int i, Real x, Real y, Real z)
{
  return planar_x(t, i, x, y, z) + M_PI * fsin(0.5 * t.fx(i) * z);
}

static Real ellipsoidal_y(const FunctionalTexture& t, int i, Real x, Real y, Real z)
{
  return planar_y(t, i, x, y, z) + M_PI * fsin(0.5 * t.fx(i) * z);
}

// make ellipsoidal texture
EllipsoidalTexture::EllipsoidalTexture(int n, Real C0, Real fx0, Real fy0, Real t0)
: FunctionalTexture(ellipsoidal_x, ellipsoidal_y, n, C0, fx0, fy0, t0)
{
}



// marble textures

Real MarbleTexture::operator()(Real x, Real y, Real z) const
{
  return (fsin(x + Turbulence(x, y, z)) + 1.0) / 2.0;
}
