////////////////////////////////////////////////////////////////////////////
//
// texture.h - definition of texture class and functional textures
//
////////////////////////////////////////////////////////////////////////////
#ifndef _TEXTURE_H_
#define _TEXTURE_H_

#include "real.h"

class Texture;
class FunctionalTexture;

// abstract texture class
class Texture {
 public:
  // return texture value at given coordinates
  virtual Real operator()(Real x, Real y, Real z) const = 0;
  // destroy a texture
  virtual ~Texture() { }
 protected:
  // create an empty texture
  Texture() { }
};

// function used to generate functional texture
typedef Real (*TextureFunc)(const FunctionalTexture& t, int, Real x, Real y, Real z);

// FunctionalTexture class - texture values are generated by a user
// defined function and pseudo FFT procedure.
// Note: these textures are NOT normalized!
// (this class was derived from Gardner's paper (see references))
class FunctionalTexture : public Texture {
 public:
  TextureFunc PX() const { return _px; }  // texture function for x direction
  TextureFunc PY() const { return _py; }  // texture function for y direction
  int Iter() const { return _size; }      // iterations for pseudo FFT

  // scale factor for texture value
  Real& k() const { return ((FunctionalTexture*)this)->_k; }

  // various constants for texture generation (see C++ code)
  Real& T0() const { return ((FunctionalTexture*)this)->_t0; }
  Real& fx(int n) const { return ((FunctionalTexture*)this)->_fx[n]; }
  Real& fy(int n) const { return ((FunctionalTexture*)this)->_fy[n]; }
  Real& C(int n) const { return ((FunctionalTexture*)this)->_Ci[n]; }

  // compute texture value at given location
  Real operator()(Real x, Real y, Real z) const;

  // create new functional texture
  FunctionalTexture(TextureFunc tx, TextureFunc ty, int n = 7,
		    Real C0 = 0.5, Real fx0 = 0.5, Real fy0 = 0.5,
		    Real t0 = 1.0);

  // delete functional texture
  ~FunctionalTexture() { delete[] _Ci; delete[] _fx; delete[] _fy; }
 private:
  TextureFunc _px;   // x texture func
  TextureFunc _py;   // y texture func
  int _size;         // iterations for pseudo FFT
  Real* _Ci;          // scaling constant
  Real* _fx;         // x scale constants
  Real* _fy;         // y scale constants
  Real _t0;          // additive constant
  Real _k;           // scale constant
  Real _min;         // min texture value
  Real _max;         // max texture value
};

// planar texture - used for generating planar cloud fields
class PlanarTexture : public FunctionalTexture {
 public:
  PlanarTexture(int n = 7, Real C0 = 0.5, Real fx0 = 0.5,
		Real fy0 = 0.5, Real t0 = 1.0);
};

// ellipsoidal texture - used for generating 3d cloud fields
class EllipsoidalTexture : public FunctionalTexture {
 public:
  EllipsoidalTexture(int n = 7, Real C0 = 0.5, Real fx0 = 0.5,
		     Real fy0 = 0.5, Real t0 = 1.0);
};

// marble texture
class MarbleTexture : public Texture {
 public:
  Real operator()(Real x, Real y, Real z) const;
  MarbleTexture() { }
  ~MarbleTexture() { }
};

#endif
