#include <stream.h>
#include "conv.h"

// print a filter
ostream& operator<<(ostream& os, const Filter& f)
{
  return f.Print(os);
}

// print a filter
ostream& Filter::Print(ostream& os, int) const
{
  os << "Filter(";
  for(int i = 0; i < _width; i++) {
    for(int j = 0; j < _height; j++)
      os << (*this)(i, j) << " ";
    os << endl;
  }
  os << ")";
  return os;
}

// create an averaging filter
AverageFilter::AverageFilter(int width, int height)
: Filter(width, height)
{
  Real value = 1.0/(width*height);
  for(int i = 0; i < width; i++)
    for(int j = 0; j < height; j++)
      Set(i, j, value);
}

// create a gaussian filter
GaussianFilter::GaussianFilter(Real sigma, int xy)
: Filter(xy ? (8*(int)sigma)+1 : 1, xy ? 1 : (8*(int)sigma)+1)
{
  int size = 4*(int)sigma;
  int index = 0;

  if(xy) {
    for(int x = -size; x <= size; x++)
      Set(index++, 0, Gaussian(sigma, (Real)x));
  } else {
    for(int y = -size; y <= size; y++)
      Set(0, index++, Gaussian(sigma, (Real)y));
  }
}

// create a delta filter
GradientFilter::GradientFilter(int xy)
: Filter(2, 2)
{
  if(xy) {
    Set(0, 0, -1.0);
    Set(0, 1, -1.0);
    Set(1, 0, 1.0);
    Set(1, 1, 1.0);
  } else {
    Set(0, 0, -1.0);
    Set(0, 1, 1.0);
    Set(1, 0, -1.0);
    Set(1, 1, 1.0);
  }
}

// perform convolution on input data with the given filter
// and store results in out (in and out must be different!)
void Convolute(Real* in, int w, int h, Filter& f, Real* out)
{
  int hw = f.Width()/2;
  int hh = f.Height()/2;
  Real* optr = out;

  for(int y = 0; y < h; y++) {
    for(int x = 0; x < w; x++) {
      const Real* fptr = f.Ptr(0, 0);
      Real sum = 0.0;
      for(int j = 0; j < f.Height(); j++) {
	int yindex = fixrng(y + j - hh, 0, h);
	for(int i = 0; i < f.Width(); i++) {
	  int xindex = fixrng(x + i - hw, 0, w);
	  sum += in[xindex + yindex * w] * *fptr++;
	}
      }
      *optr++ = sum;
    }
  }
}


// find the gradients and gradient magintudes of the given data
// if dxi is 0 don't compute x gradients
// if dyi is 0 don't compute y gradients
// if magi is 0 don't compute magnitudes of gradients
// if ave is 0 don't compute average
void Gradients(Real* data, int w, int h, Real* dxi,
	       Real* dyi, Real* magi, Real* ave)
{
  GradientFilter fx(1), fy(0);
  Real* xptr = dxi;
  Real* yptr = dyi;
  Real* mptr = magi;

  // find x & y gradients
  if(dxi) Convolute(data, w, h, fx, dxi);
  if(dyi) Convolute(data, w, h, fy, dyi);

  // find average gradient magnitudes and average magnitude
  if(ave) *ave = 0.0;
  if(magi && dxi && dyi) {
    for(int i = w * h - 1; i >= 0; i--) {
      Real dx = *xptr++;
      Real dy = *yptr++;
      if(ave)
	*ave += *mptr++ = sqrt((dx * dx) + (dy * dy));
      else
	*mptr++ = sqrt((dx * dx) + (dy * dy));
    }
  }
  if(ave) *ave /= w*h;
}

// perform smoothing on given data using a gaussian filter
void Smooth(Real* data, int w, int h, Real sigma)
{
  GaussianFilter fx(sigma, 0), fy(sigma, 1);
  Real* tmp = new Real[w*h];
  Convolute(data, w, h, fy, tmp);
  Convolute(tmp, w, h, fx, data);
  delete[] tmp;
}

// perform edge detection on the given data using
// changes in gradient intensity to detect edges
// out must point to an array of sufficient size to
// hold the results
void EdgeDetect(Real* data, int w, int h, Real th, unsigned char* out)
{
  Real* dxi = new Real[w*h];
  Real* dyi = new Real[w*h];
  Real* magi = new Real[w*h];
  Real* xptr = dxi;
  Real* yptr = dyi;
  Real* mptr = magi;
  unsigned char* optr = out;
  Real mag, pp, pm, ave = 0.0;
  int x, y, dx, dy, xpm, xpp, ypm, ypp;
  int total = 0;

  assert(out);
  MemClear(out, w*h*sizeof(*out));

  // compute gradients
  Gradients(data, w, h, dxi, dyi, magi, &ave);
  if(th < 0.0) th = ave / 1.5;

  for(y = 0; y < h; y++) {
    for(x = 0; x < w; x++, xptr++, yptr++, optr++) {
      mag = *mptr++;
      if(mag != 0.0) {
	dx = (int)round(*xptr / mag);
	dy = (int)round(*yptr / mag);
#if 0
	assert(dx == -1 || dx == 0 || dx == 1);
	assert(dy == -1 || dy == 0 || dy == 1);
#endif
	xpp = fixrng(x + dx, 0, w);
	xpm = fixrng(x - dx, 0, w);
	ypp = fixrng(y + dy, 0, h);
	ypm = fixrng(y - dy, 0, h);
	pp = magi[xpp + ypp * w];
	pm = magi[xpm + ypm * w];
	/* could do mag > th here! */
	if(mag > th && ((mag > pp && mag >= pm) || (mag > pm && mag >= pp))) {
	  *optr = 255;
#if 0
	  total++;
#endif
	}
      }
    }
  }
#if 0
  printf("total edge pixels detected = %d\n", total);
  printf("average magnitude = %2.2f\n", ave);
#endif

  // delete tmp data
  delete[] dxi;
  delete[] dyi;
  delete[] magi;
}

