////////////////////////////////////////////////////////////////////////////
//
// conv.h - classes for performing convolutions and convolution filters
//
////////////////////////////////////////////////////////////////////////////
#ifndef _CONV_H_
#define _CONV_H_

#include "real.h"

// generic convolution filter
class Filter {
 public:
  int Width() const { return _width; }     // width of filter
  int Height() const { return _height; }   // height of filter

  // access filter elements
  Real operator()(int i, int j) const { return _data[i+j*_width]; }

  // return pointer to filter elements (for faster access)
  const Real* Ptr(int i, int j) { return &_data[i+j*_width]; }

  // set a filter element
  void Set(int i, int j, Real r) { _data[i+j*_width] = r; }

  // create an empty filter
  Filter() : _width(0), _height(0), _data(0) { }

  // create empty filter of the given size
  Filter(int w, int h) : _width(w), _height(h), _data(new Real[w*h])
    { MemClear(_data, sizeof(*_data)*w*h); }

  // destroy a filter
  ~Filter() { delete[] _data; }

  // print filter
  friend ostream& operator<<(ostream&, const Filter&);
  ostream& Print(ostream&, int=0) const;
 protected:
  // set size of a filter
  void SetSize(int w, int h)
    {
      delete[] _data;
      _width = w;
      _height = h;
      _data = new Real[w*h];
      MemClear(_data, sizeof(*_data)*w*h);
    }
 private:
  int _width;    // filter width
  int _height;   // filter height
  Real* _data;   // filter data
};

// averaging filter
class AverageFilter : public Filter {
 public:
  AverageFilter(int, int);
};

// gaussian filter
class GaussianFilter : public Filter {
 public:
  GaussianFilter(Real sigma, int xy);
};

// gradient filter
class GradientFilter : public Filter {
 public:
  GradientFilter(int xy);
};

// perform convolution on input data with given filter
// out must point to an array big enough to hold results
// (it must also be different from in)
extern void Convolute(Real* in, int w, int h, Filter& f, Real* out);

// compute gradients, gradient magnitudes and average magnitude for the
// given input.  dxi, dxi, mag and ave can be 0, in which case that
// particular component will not be computed (note: dxi and dyi are
// needed to compute mag and mag is needed to compute ave)
extern void Gradients(Real* d, int w, int h, Real* dxi, Real* dyi, Real* magi, Real* ave);

// perform smoothing using gaussian filter
extern void Smooth(Real* data, int w, int h, Real sigma);

// perform edge detection on input data (out must point to pre-allocated data)
extern void EdgeDetect(Real* data, int w, int h, Real t, unsigned char* out);

#endif
