#include "fft.h"

/*	FOURN  --  Multi-dimensional fast Fourier transform

	Called with arguments:

	   data       A  one-dimensional  array  of  floats  (NOTE!!!	NOT
		      DOUBLES!!), indexed from one (NOTE!!!   NOT  ZERO!!),
		      containing  pairs of numbers representing the complex
		      valued samples.  The Fourier transformed results	are
		      returned in the same array.

	   nn	      An  array specifying the edge size in each dimension.
		      THIS ARRAY IS INDEXED FROM  ONE,	AND  ALL  THE  EDGE
		      SIZES MUST BE POWERS OF TWO!!!

	   ndim       Number of dimensions of FFT to perform.  Set to 2 for
		      two dimensional FFT.

	   isign      If 1, a Fourier transform is done; if -1 the  inverse
		      transformation is performed.

        This  function  is essentially as given in Press et al., "Numerical
        Recipes In C", Section 12.11, pp.  467-470.
*/

void FFT(MyComplex* data, int nn[], int dim, Boolean inv)
{
  int isign = inv ? -1 : 1;
  int ntot = 1;
  int nprev = 1;

  for(int idim = 1; idim <= dim; idim++)
    ntot *= nn[idim];

  for(idim = dim; idim >= 1; idim--) {
    int n = nn[idim];
    int nrem = ntot / (n * nprev);
    int ip1 = nprev;
    int ip2 = ip1 * n;
    int ip3 = ip2 * nrem;
    int i2rev = 1;

    for(int i2 = 1; i2 <= ip2; i2 += ip1) {
      if(i2 < i2rev) {
	for(int i1 = i2; i1 <= i2 + ip1 - 1; i1++)
	  for(int i3 = i1; i3 <= ip3; i3 += ip2)
	    tswap(MyComplex, data[i3], data[i2rev + i3 - i2]);
      }
      int ibit = ip2;
      while(ibit >= (2*ip1) && (2*i2rev) > ibit) {
	i2rev -= ibit/2;
	ibit >>= 1;
      }
      i2rev += ibit/2;
    }
    int ifp1 = ip1, ifp2;
    while(ifp1 < ip2) {
      ifp2 = ifp1 << 1;
      Real theta = isign * (2.0 * M_PI) / (ifp2 / ip1);
      Real wtemp = fsin(0.5 * theta);
      Real wpr = -2.0 * wtemp * wtemp;
      Real wpi = fsin(theta);
      MyComplex w(1.0, 0.0);

      for(int i3 = 1; i3 <= ifp1; i3 += ip1) {
	for(int i1 = i3; i1 <= i3 + ip1 - 1; i1++) {
	  for(i2 = i1; i2 <= ip3; i2 += ifp2) {
	    int k2 = i2 + ifp1;
	    MyComplex tmp = w * data[k2];
	    data[k2] = data[i2] - tmp;
	    data[i2] += tmp;
	  }
	}
	wtemp = w.r;
	w.r = wtemp * wpr - w.i * wpi + w.r;
	w.i = w.i * wpr + wtemp * wpi + w.i;
      }
      ifp1 = ifp2;
    }
    nprev *= n;
  }
}

// perform 1d discrete inverse FFT
void InvFFT1d(MyComplex* data, int w)
{
  int nn[1];
  nn[0] = w;
  FFT(data-1, nn-1, 1, True);
}

// perform 1d discrete FFT
void FFT1d(MyComplex* data, int w)
{
  int nn[1];
  nn[0] = w;
  FFT(data-1, nn-1, 1, False);
}

// perform 2d discrete inverse FFT
void InvFFT2d(MyComplex* data, int w, int h)
{
  int nn[2];
  nn[0] = w; nn[1] = h;
  FFT(data-1, nn-1, 2, True);
}

// perform 2d discrete FFT
void FFT2d(MyComplex* data, int w, int h)
{
  int nn[2];
  nn[0] = w; nn[1] = h;
  FFT(data-1, nn-1, 2, False);
}
