////////////////////////////////////////////////////////////////////////////
//
// ftab.h - defines FuncTab class which is used to store and lookup
//          precomputed function values (Note: this implementation
//          does not provide any type of interpolation for intermediate
//          values)
//
////////////////////////////////////////////////////////////////////////////
#ifndef _FUNCTAB_H_
#define _FUNCTAB_H_

#include "misc.h"

// different types of double functions
typedef double (*DoubleFn1)(double);
typedef double (*DoubleFn2)(double, double);
typedef double (*DoubleFn3)(double, double, double);

// FuncTab class - used to store and lookup precomputed function values
// Note: add option to interpolate intermediate values
class FuncTab {
 public:
  // return size of func tab (e.g. number of entries)
  int Size() const { return _size; }

  // return granularity of functab (e.g. sampling size for input values)
  double Granularity() const { return _gran; }

  // lookup function value at v
  double operator()(double v) const { return _data[(int)(v * _index)]; }

  // return nth function value
  double operator[](int n) const;

  // copy a func tab's data
  void Copy(const FuncTab&);

  // create empty functab
  FuncTab();

  // copy a functab
  FuncTab(const FuncTab&);

  // create functab from given function, max value and granularity
  FuncTab(DoubleFn1 fn, double max, double gran);

  // destroy a func tab
  ~FuncTab() { delete[] _data; }

  // print FuncTab
  friend inline ostream& operator<<(ostream&, const FuncTab&);
  ostream& Print(ostream& o, int=0) const;
 protected:
  // return references to functab values so they can be modified
  double& Value(double v) const { return _data[(int)(v * _index)]; }
  double& Value(int n) const { return _data[n]; }
  double Index() const { return _index; }
 private:
  int _size;      // number of precomputed values
  double _index;  // scale to convert lookup value to array index
  double _gran;   // granularity (sampling rate) of func tab
  double* _data;  // actual precomputed data
  double _max;    // maximum value that func tab is computed for
                  // (i.e. data is computed for inputs from [0, max]
                  //  with _gran as the sampling rate)
  DoubleFn1 _fn;  // actual function used to compute the data
};


// FuncTab with linera interpolation
class InterpFuncTab : public FuncTab {
 public:
  // lookup function value at v
  double operator()(double v) const;

  // create empty functab
  InterpFuncTab() : FuncTab() { }

  // copy a functab
  InterpFuncTab(const InterpFuncTab& ift) : FuncTab(ift) { }

  // create functab from given function, max value and granularity
  InterpFuncTab(DoubleFn1 fn, double max, double gran)
    : FuncTab(fn, max, gran) { }
};


#endif
