////////////////////////////////////////////////////////////////////////////
//
// point.h - simple 3d point and spherical point classes
//
////////////////////////////////////////////////////////////////////////////
#ifndef _POINT3D_H_
#define _POINT3D_H_

#include "real.h"
#include "trig.h"

class Vector3d;
class SphericalPoint;

// 3d point
class Point3d {
 public:
  Real x, y, z;   // x, y, z components of point

  // allow access to point as if it were an array of Real[3]
  Real& operator[](int i) { return !i ? x : (i == 1 ? y : z); }

  Point3d operator*(const Point3d& p) const   // multiply two points
    { return Point3d(x*p.x, y*p.y, z*p.z); }  // (element by element)
  Point3d operator/(const Point3d& p) const   // divide two points
    { return Point3d(x/p.x, y/p.y, z/p.z); }  // (element by element)
  Point3d operator+(const Point3d& p) const   // add two points
    { return Point3d(x+p.x, y+p.y, z+p.z); }  // (element by element)
  Point3d operator-(const Point3d& p) const   // subtract two points
    { return Point3d(x-p.x, y-p.y, z-p.z); }  // (element by element)

  // multiple/divide/add/subtract constant to point
  Point3d operator*(Real r) const { return Point3d(x*r, y*r, z*r); }
  Point3d operator/(Real r) const { return Point3d(x/r, y/r, z/r); }
  Point3d operator+(Real r) const { return Point3d(x+r, y+r, z+r); }
  Point3d operator-(Real r) const { return Point3d(x-r, y-r, z-r); }

  // assignment operator versions of above functions
  const Point3d& operator*=(const Point3d& p) { return (*this = *this * p); }
  const Point3d& operator/=(const Point3d& p) { return (*this = *this / p); }
  const Point3d& operator+=(const Point3d& p) { return (*this = *this + p); }
  const Point3d& operator-=(const Point3d& p) { return (*this = *this - p); }

  const Point3d& operator*=(Real r) { return (*this = *this * r); }
  const Point3d& operator/=(Real r) { return (*this = *this / r); }
  const Point3d& operator+=(Real r) { return (*this = *this + r); }
  const Point3d& operator-=(Real r) { return (*this = *this - r); }

  // create point at origin
  Point3d() : x(0.0), y(0.0), z(0.0) { }

  // create a point from a vector
  Point3d(const Vector3d& v);

  // convert spherical point to rectangular point
  Point3d(const SphericalPoint& p);

  // copy another point
  Point3d(const Point3d& p) : x(p.x), y(p.y), z(p.z) { }

  // create point from x, y and z coords
  Point3d(Real xx, Real yy, Real zz) : x(xx), y(yy), z(zz) { }

  // destroy a point
  ~Point3d() { }

  // print point
  friend ostream& operator<<(ostream&, const Point3d&);
  ostream& Print(ostream& o, int=0) const;
};

// spherical point, defined by two angles and a radius
class SphericalPoint {
 public:
  Real theta, phi, radius;  // spherical point components

  // return x, y, and z coords of this spherical point
  Real x() const { return radius*fsin(phi)*fcos(theta); }
  Real y() const { return radius*fsin(phi)*fsin(theta); }
  Real z() const { return radius*fcos(phi); }

  // create spherical point at origin
  SphericalPoint() : theta(0.0), phi(0.0), radius(0.0) { }

  // convert regular point to spherical point
  SphericalPoint(const Point3d& p);

  // copy spherical point
  SphericalPoint(const SphericalPoint& p) : theta(p.theta), phi(p.phi), radius(p.radius) { }

  // create spherical point from angles and radius
  SphericalPoint(Real t, Real p, Real r) : theta(t), phi(p), radius(r) { }

  // destroy spherical point
  ~SphericalPoint() { }

  // print spherical point
  friend ostream& operator<<(ostream&, const SphericalPoint&);
  ostream& Print(ostream& o, int=0) const;
};

// actual constructors down here to prevent circular dependencies
#include "vector.h"

inline Point3d::Point3d(const Vector3d& v) : x(v.i), y(v.j), z(v.k) { }
inline Point3d::Point3d(const SphericalPoint& p) : x(p.x()), y(p.y()), z(p.z()) { }

#endif
