#include <stream.h>
#include "quadric.h"

ostream& operator<<(ostream& os, const Ellipsoid& e)
{
  return e.Print(os);
}

ostream& Ellipsoid::Print(ostream& os, int) const
{
  int oldp = os.precision();
  os.precision(4);
  os << "Q(x, y, z) = " << _q[0] << "x^2";
  os << " + " << _q[1] << "y^2";
  os << " + " << _q[2] << "z^2";
  os << " + " << _q[3] << "xy";
  os << " + " << _q[4] << "xz";
  os << " + " << _q[5] << "yz";
  os << " + " << _q[6] << "x";
  os << " + " << _q[7] << "y";
  os << " + " << _q[8] << "z";
  os << " + " << _q[9];
  os << ", LimbMaxXY = " << _limb_max_xy;
  return os;
}

Real Ellipsoid::LimbXZ(Real x, Real z, Real f) const
{
  Real A = _q[0]*x*x+_q[1]*f*f + _q[2]*z*z + _q[3]*x*f + _q[4]*x*z + _q[5]*f*z;
  Real B = _q[6]*x + _q[7]*f + _q[8]*z;
  Real C = _q[9];
  return sqr(B) - 4 * A * C;
}

// maximum is at geometric center of limb
Real Ellipsoid::MaxLimbXZ(Real, Real, Real f) const
{
  return sqr(f) * _limb_max_xz;
}


Real Ellipsoid::LimbXY(Real x, Real y, Real f) const
{
  Real A = _q[0]*x*x+_q[1]*y*y + _q[2]*f*f + _q[3]*x*y + _q[4]*x*f + _q[5]*y*f;
  Real B = _q[6]*x + _q[7]*y + _q[8]*f;
  Real C = _q[9];
  return sqr(B) - 4 * A * C;
}

// maximum is at geometric center of limb
Real Ellipsoid::MaxLimbXY(Real, Real, Real f) const
{
  return sqr(f) * _limb_max_xy;
}



// F(x, y, z) = 0
// normal = (dF/dx, dF/dy, dF/dz)

// evaluate dF/dx at given point for this ellipse
Real Ellipsoid::dx(Real x, Real y, Real z) const
{
  return 2.0*_q[0]*x + _q[1]*y*y + _q[2]*z*z + _q[3]*y +
         _q[4]*z + _q[5]*y*z + _q[6] + _q[7]*y + _q[8]*z;
}

// evaluate dF/dy at given point for this ellipse
Real Ellipsoid::dy(Real x, Real y, Real z) const
{
  return _q[0]*x*x + 2.0*_q[1]*y + _q[2]*z*z + _q[3]*x +
         _q[4]*x*z + _q[5]*z + _q[6]*x + _q[7] + _q[8]*z;
}

// evaluate dF/dz at given point for this ellipse
Real Ellipsoid::dz(Real x, Real y, Real z) const
{
  return _q[0]*x*x + _q[1]*y*y + 2.0*_q[2]*z + _q[3]*x*y +
         _q[4]*x + _q[5]*y + _q[6]*x + _q[7]*y + _q[8];
}

// compute the normal vector at given point on this ellipse
Vector3d Ellipsoid::Normal(const Point3d& p) const
{
  Vector3d n(dx(p.x, p.y, p.z), dy(p.x, p.y, p.z), dz(p.x, p.y, p.z));
  return n.Normalize();
}

void Ellipsoid::FindLimbMax()
{
  Real dmax1 = _q[6]*_q[8]*_q[4] + 4*_q[0]*_q[9]*_q[2] - _q[0]*_q[8]*_q[8]
             - _q[9]*_q[4]*_q[4] - _q[2]*_q[6]*_q[6];
  Real dmax2 = -_q[6]*_q[6] + 4 * _q[9] * _q[0];

  Real nmax1 = -_q[9]*(-4*_q[0]*_q[9]*sqr(_q[5])+16*_q[0]*_q[2]*_q[9]*_q[1]
	               -4*_q[0]*_q[1]*sqr(_q[8])-4*_q[0]*_q[2]*sqr(_q[7])
		       +4*_q[0]*_q[7]*_q[8]*_q[5]-4*_q[2]*sqr(_q[3])*_q[9]
		       +4*_q[9]*_q[5]*_q[3]*_q[4]-4*_q[9]*_q[1]*sqr(_q[4])
		       +4*_q[1]*_q[6]*_q[8]*_q[4]-4*_q[2]*_q[1]*sqr(_q[6])
		       +4*_q[2]*_q[3]*_q[6]*_q[7]+sqr(_q[7])*sqr(_q[4])
		       +sqr(_q[3])*sqr(_q[8])+sqr(_q[5])*sqr(_q[6])
		       -2*_q[3]*_q[6]*_q[8]*_q[5]-2*_q[7]*_q[8]*_q[3]*_q[4]
		       -2*_q[5]*_q[6]*_q[7]*_q[4]);
  Real nmax2 = -4*_q[9]*(4*_q[0]*_q[9]*_q[1]-_q[9]*sqr(_q[3])-_q[0]*sqr(_q[7])
			 -_q[1]*sqr(_q[6])+_q[3]*_q[6]*_q[7]);

  assert(dmax1 != 0.0 && dmax2 != 0.0);

#ifdef DEBUG
  cout << "xz limbs: " << nmax1/dmax1 << ", " << nmax2/dmax2 << endl;
#endif

  if(dmax1 == 0.0) {
    _limb_max_xz = nmax2/dmax2;
  } else if(dmax2 == 0.0) {
    _limb_max_xz = nmax1/dmax1;
  } else {
    _limb_max_xz = max(nmax1/dmax1, nmax2/dmax2);
  }

  nmax2 = -4*_q[9]*(-_q[0]*_q[8]*_q[8] - _q[9]*_q[4]*_q[4] - _q[2]*_q[6]*_q[6]
		    + 4*_q[9]*_q[0]*_q[2] + _q[4]*_q[6]*_q[8]);

  dmax1 = -4*_q[9]*_q[0]*_q[1] + _q[9]*_q[3]*_q[3] + _q[0]*_q[7]*_q[7]
          + _q[1]*_q[6]*_q[6] - _q[3]*_q[6]*_q[7];

  nmax1 = -_q[9]*(4*_q[9]*_q[0]*sqr(_q[5])-16*_q[9]*_q[0]*_q[2]*_q[1]
		  +4*_q[9]*_q[2]*sqr(_q[3])-4*_q[9]*_q[5]*_q[4]*_q[3]
		  +4*_q[9]*_q[1]*sqr(_q[4])+4*_q[0]*sqr(_q[8])*_q[1]
		  +4*_q[0]*_q[2]*sqr(_q[7])-4*_q[0]*_q[5]*_q[7]*_q[8]
		  -sqr(_q[3])*sqr(_q[8])+2*_q[3]*_q[5]*_q[6]*_q[8]
		  +2*_q[3]*_q[4]*_q[7]*_q[8]-4*_q[3]*_q[2]*_q[6]*_q[7]
		  +4*_q[2]*_q[1]*sqr(_q[6])-sqr(_q[4])*sqr(_q[7])
		  -4*_q[1]*_q[4]*_q[6]*_q[8]+2*_q[5]*_q[4]*_q[6]*_q[7]
		  -sqr(_q[5])*sqr(_q[6]));

#ifdef DEBUG
  cout << "xy limbs: " << nmax1/dmax1 << ", " << nmax2/dmax2 << endl;
#endif

  if(dmax1 == 0.0) {
    _limb_max_xy = nmax2/dmax2;
  } else if(dmax2 == 0.0) {
    _limb_max_xy = nmax1/dmax1;
  } else {
    _limb_max_xy = max(nmax1/dmax1, nmax2/dmax2);
  }

  _limb_max_yz = 0.0;
}



// create a new ellipsoid with the given table of coefficients
Ellipsoid::Ellipsoid(Real* q)
{
  MemCopy(q, _q, 10*sizeof(*q));
  FindLimbMax();
}

// create a new ellipsoid with the given list of coefficients
Ellipsoid::Ellipsoid(Real q0, Real q1, Real q2, Real q3, Real q4,
		     Real q5, Real q6, Real q7, Real q8, Real q9)
{
  _q[0] = q0;
  _q[1] = q1;
  _q[2] = q2;
  _q[3] = q3;
  _q[4] = q4;
  _q[5] = q5;
  _q[6] = q6;
  _q[7] = q7;
  _q[8] = q8;
  _q[9] = q9;
  FindLimbMax();
}


Ellipsoid::Ellipsoid(const Point3d& P, const Point3d& Q,
		     const Point3d& R, const Point3d& T)
{
  _q[0] = - sqr(R.y) * sqr(Q.z) - sqr(Q.y) * sqr(R.z) - sqr(R.z) * sqr(P.y)
       - sqr(P.z) * sqr(R.y) + 2 * P.z * R.y * R.z * P.y + 2 * Q.y * R.z
       * R.y * Q.z - sqr(Q.y) * sqr(P.z) - sqr(P.y) * sqr(Q.z) + 2 * Q.y
       * P.z * P.y * Q.z;

  _q[1] = 2 * R.x * P.z * R.z * P.x + 2 * P.x * Q.z * Q.x * P.z - sqr(P.x)
     * sqr(Q.z) - sqr(R.x) * sqr(P.z) - sqr(R.z) * sqr(Q.x) - sqr(R.x)
     * sqr(Q.z) + 2 * R.x * Q.z * R.z * Q.x - sqr(Q.x) * sqr(P.z) - sqr(R.z)
     * sqr(P.x);

  _q[2] = - sqr(R.y) * sqr(P.x) - sqr(Q.x) * sqr(P.y) + 2 * Q.y * R.x * R.y
     * Q.x + 2 * P.x * Q.y * Q.x * P.y - sqr(P.x) * sqr(Q.y) + 2 * R.x * P.y
     * R.y * P.x - sqr(R.y) * sqr(Q.x) - sqr(R.x) * sqr(P.y) - sqr(Q.y)
     * sqr(R.x);

  _q[3] = ((2*sqr(Q.z)+2*sqr(R.z))*P.y+(-2*Q.z*Q.y-2*R.z*R.y)*P.z)*P.x+(-2*R.x*R.z-2*Q.x*Q.z)*P.z*P.y+(2*R.x*R.y+2*Q.x*Q.y)*sqr(P.z)+(2*Q.y*sqr(R.z)-2*Q.z*R.y*R.z)*Q.x+2*R.y*sqr(Q.z)*R.x-2*Q.y*R.z*R.x*Q.z;

  _q[4] = ((-2*Q.z*Q.y-2*R.z*R.y)*P.y+(2*sqr(R.y)+2*sqr(Q.y))*P.z)*P.x+(2*Q.x*Q.z+2*R.x*R.z)*sqr(P.y)+(-2*R.x*R.y-2*Q.x*Q.y)*P.z*P.y+(-2*Q.y*R.z*R.y+2*sqr(R.y)*Q.z)*Q.x-2*Q.y*R.x*R.y*Q.z+2*sqr(Q.y)*R.z*R.x;

  _q[5] = (2*R.z*R.y+2*Q.z*Q.y)*sqr(P.x)+((-2*R.x*R.z-2*Q.x*Q.z)*P.y+(-2*R.x*R.y-2*Q.x*Q.y)*P.z)*P.x+(2*sqr(R.x)+2*sqr(Q.x))*P.z*P.y+2*R.y*sqr(Q.x)*R.z+(-2*R.x*R.y*Q.z-2*Q.y*R.x*R.z)*Q.x+2*Q.y*sqr(R.x)*Q.z;

  _q[6] = ((2*Q.y*T.z*Q.z-2*sqr(R.z)*T.y-2*T.y*sqr(Q.z)+2*R.z*T.z*R.y)*P.y+(2*T.y*Q.z*Q.y-2*sqr(R.y)*T.z-2*sqr(Q.y)*T.z+2*R.z*T.y*R.y)*P.z)*P.x+(2*sqr(R.z)*T.x-2*Q.x*T.z*Q.z-2*R.x*T.z*R.z+2*T.x*sqr(Q.z))*sqr(P.y)+((2*T.y*Q.z+2*T.z*Q.y)*Q.x-4*T.x*Q.y*Q.z+(2*T.z*R.y+2*T.y*R.z)*R.x-4*R.y*T.x*R.z)*P.z*P.y+(2*T.x*sqr(Q.y)-2*Q.x*T.y*Q.y-2*R.x*T.y*R.y+2*sqr(R.y)*T.x)*sqr(P.z)+((-2*sqr(R.z)*T.y+2*R.z*T.z*R.y)*Q.y+(-2*sqr(R.y)*T.z+2*R.z*T.y*R.y)*Q.z)*Q.x+(2*sqr(R.z)*T.x-2*R.x*T.z*R.z)*sqr(Q.y)+((2*T.z*R.y+2*T.y*R.z)*R.x-4*R.y*T.x*R.z)*Q.z*Q.y+(-2*R.x*T.y*R.y+2*sqr(R.y)*T.x)*sqr(Q.z);

  _q[7] = (2*T.y*sqr(Q.z)-2*Q.y*T.z*Q.z+2*sqr(R.z)*T.y-2*R.z*T.z*R.y)*sqr(P.x)+((-2*T.x*sqr(Q.z)+2*R.x*T.z*R.z+2*Q.x*T.z*Q.z-2*sqr(R.z)*T.x)*P.y+((2*T.z*Q.y-4*T.y*Q.z)*Q.x+2*T.x*Q.y*Q.z+(-4*T.y*R.z+2*T.z*R.y)*R.x+2*R.y*T.x*R.z)*P.z)*P.x+(-2*sqr(Q.x)*T.z+2*R.x*R.z*T.x-2*sqr(R.x)*T.z+2*Q.x*T.x*Q.z)*P.z*P.y+(-2*R.x*R.y*T.x-2*Q.x*T.x*Q.y+2*sqr(R.x)*T.y+2*sqr(Q.x)*T.y)*sqr(P.z)+(2*sqr(R.z)*T.y-2*R.z*T.z*R.y)*sqr(Q.x)+((2*R.x*T.z*R.z-2*sqr(R.z)*T.x)*Q.y+((-4*T.y*R.z+2*T.z*R.y)*R.x+2*R.y*T.x*R.z)*Q.z)*Q.x+(2*R.x*R.z*T.x-2*sqr(R.x)*T.z)*Q.z*Q.y+(-2*R.x*R.y*T.x+2*sqr(R.x)*T.y)*sqr(Q.z);

  _q[8] = (-2*R.z*T.y*R.y+2*sqr(Q.y)*T.z+2*sqr(R.y)*T.z-2*T.y*Q.z*Q.y)*sqr(P.x)+(((-4*T.z*Q.y+2*T.y*Q.z)*Q.x+2*T.x*Q.y*Q.z+(2*T.y*R.z-4*T.z*R.y)*R.x+2*R.y*T.x*R.z)*P.y+(-2*T.x*sqr(Q.y)+2*Q.x*T.y*Q.y+2*R.x*T.y*R.y-2*sqr(R.y)*T.x)*P.z)*P.x+(-2*Q.x*T.x*Q.z-2*R.x*R.z*T.x+2*sqr(Q.x)*T.z+2*sqr(R.x)*T.z)*sqr(P.y)+(2*R.x*R.y*T.x-2*sqr(Q.x)*T.y+2*Q.x*T.x*Q.y-2*sqr(R.x)*T.y)*P.z*P.y+(-2*R.z*T.y*R.y+2*sqr(R.y)*T.z)*sqr(Q.x)+(((2*T.y*R.z-4*T.z*R.y)*R.x+2*R.y*T.x*R.z)*Q.y+(2*R.x*T.y*R.y-2*sqr(R.y)*T.x)*Q.z)*Q.x+(-2*R.x*R.z*T.x+2*sqr(R.x)*T.z)*sqr(Q.y)+(2*R.x*R.y*T.x-2*sqr(R.x)*T.y)*Q.z*Q.y;

  _q[9] = ((sqr(R.z)-sqr(T.z))*sqr(Q.y)+(2*T.z*T.y-2*R.z*R.y)*Q.z*Q.y+(sqr(R.y)-sqr(T.y))*sqr(Q.z)-sqr(T.z)*sqr(R.y)+2*R.z*T.y*T.z*R.y-sqr(R.z)*sqr(T.y))*sqr(P.x)+((((2*sqr(T.z)-2*sqr(R.z))*Q.y+(-2*T.z*T.y+2*R.z*R.y)*Q.z)*Q.x+(2*R.x*R.z-2*T.z*T.x)*Q.z*Q.y+(-2*R.x*R.y+2*T.y*T.x)*sqr(Q.z)+(2*sqr(T.z)*R.y-2*R.z*T.y*T.z)*R.x-2*R.z*T.x*T.z*R.y+2*sqr(R.z)*T.y*T.x)*P.y+(((-2*T.z*T.y+2*R.z*R.y)*Q.y+(2*sqr(T.y)-2*sqr(R.y))*Q.z)*Q.x+(2*T.z*T.x-2*R.x*R.z)*sqr(Q.y)+(-2*T.y*T.x+2*R.x*R.y)*Q.z*Q.y+(-2*T.y*T.z*R.y+2*sqr(T.y)*R.z)*R.x-2*R.y*T.x*R.z*T.y+2*sqr(R.y)*T.x*T.z)*P.z)*P.x+((sqr(R.z)-sqr(T.z))*sqr(Q.x)+(2*T.z*T.x-2*R.x*R.z)*Q.z*Q.x+(-sqr(T.x)+sqr(R.x))*sqr(Q.z)-sqr(R.z)*sqr(T.x)-sqr(R.x)*sqr(T.z)+2*R.x*T.z*R.z*T.x)*sqr(P.y)+((2*T.z*T.y-2*R.z*R.y)*sqr(Q.x)+((2*R.x*R.z-2*T.z*T.x)*Q.y+(-2*T.y*T.x+2*R.x*R.y)*Q.z)*Q.x+(2*sqr(T.x)-2*sqr(R.x))*Q.z*Q.y+2*sqr(R.x)*T.z*T.y+(-2*T.x*R.z*T.y-2*T.z*R.y*T.x)*R.x+2*R.y*sqr(T.x)*R.z)*P.z*P.y+((sqr(R.y)-sqr(T.y))*sqr(Q.x)+(-2*R.x*R.y+2*T.y*T.x)*Q.y*Q.x+(-sqr(T.x)+sqr(R.x))*sqr(Q.y)+2*R.x*T.y*R.y*T.x-sqr(R.x)*sqr(T.y)-sqr(R.y)*sqr(T.x))*sqr(P.z)+(-sqr(T.z)*sqr(R.y)+2*R.z*T.y*T.z*R.y-sqr(R.z)*sqr(T.y))*sqr(Q.x)+(((2*sqr(T.z)*R.y-2*R.z*T.y*T.z)*R.x-2*R.z*T.x*T.z*R.y+2*sqr(R.z)*T.y*T.x)*Q.y+((-2*T.y*T.z*R.y+2*sqr(T.y)*R.z)*R.x-2*R.y*T.x*R.z*T.y+2*sqr(R.y)*T.x*T.z)*Q.z)*Q.x+(-sqr(R.z)*sqr(T.x)-sqr(R.x)*sqr(T.z)+2*R.x*T.z*R.z*T.x)*sqr(Q.y)+(2*sqr(R.x)*T.z*T.y+(-2*T.x*R.z*T.y-2*T.z*R.y*T.x)*R.x+2*R.y*sqr(T.x)*R.z)*Q.z*Q.y+(2*R.x*T.y*R.y*T.x-sqr(R.x)*sqr(T.y)-sqr(R.y)*sqr(T.x))*sqr(Q.z);

  FindLimbMax();
}
