////////////////////////////////////////////////////////////////////////////
//
// quadric.h - defines different types of quadric surfaces
//             (for now only ellipsoids are defined)
//
////////////////////////////////////////////////////////////////////////////
#ifndef _QUADRIC_H_
#define _QUADRIC_H_

#include "point.h"
#include "vector.h"

// Ellipsoidal quadric surface, defined by equation:
// F(x, y, z) = q0 x^2 + q1 y^2 + q2 z^2 + q3 xy + q4 xz +
//              q5 yz + q6 x + q7 y + q8 z + q9 
class Ellipsoid {
 public:
  ostream& Print(ostream&, int=0) const;

  // return the nth coefficient of the ellipsoid equation
  Real& Q(int i) const { return ((Ellipsoid*)this)->_q[i]; }

  // compute normal at the given point
  Vector3d Normal(const Point3d&) const;

  // compute value of the limb function (ellipsoid projected onto xz plane)
  Real LimbXZ(Real x, Real z, Real f = 1.0) const;

  // compute the maximum of the xz limb function
  Real MaxLimbXZ(Real x, Real z, Real f = 1.0) const;

  // compute value of the limb function (ellipsoid projected onto xy plane)
  Real LimbXY(Real x, Real y, Real f = 1.0) const;

  // compute the maximum of the xy limb function
  Real MaxLimbXY(Real x, Real y, Real f = 1.0) const;

  // compute F(x, y, z)
  Real operator()(Real x, Real y, Real z) const
  {
    return _q[0]*x*x + _q[1]*y*y + _q[2]*z*z + _q[3]*x*y + _q[4]*x*z +
           _q[5]*y*z + _q[6]*x + _q[7]*y + _q[8]*z + _q[9];
  }

  Real dx(Real x, Real y, Real z) const;  // compute dF/dx at given point
  Real dy(Real x, Real y, Real z) const;  // compute dF/dy at given point
  Real dz(Real x, Real y, Real z) const;  // compute dF/dz at given point

  // create ellipsoid from array of coefficients
  Ellipsoid(Real*);

  // create ellipsoid from list of coefficients
  Ellipsoid(Real, Real, Real, Real, Real,
	    Real, Real, Real, Real, Real);

  // create ellipsoid from control points and origin
  Ellipsoid(const Point3d&, const Point3d&, const Point3d&, const Point3d&);

  // destroy ellipsoid
  ~Ellipsoid() { }

  friend ostream& operator<<(ostream&, const Ellipsoid&);
 protected:
  void FindLimbMax();
 private:
  Real _q[10];  // array of coefficients
  Real _limb_max_xz;
  Real _limb_max_xy;
  Real _limb_max_yz;
};

#endif
