#include <sys/time.h>
#include <stream.h>
#include "real.h"
#include "trig.h"

extern "C" void srand48(long);

/* number of samples to take for gaussian random number generator */
static int NumSamples = 4;
static double GaussAdd = sqrt(3.0 * NumSamples);
static double GaussFac = (2.0 * GaussAdd) / (double)NumSamples;

/* minimum and maximum possible values for gaussian random number generator */
double grand_min = -GaussAdd;
double grand_max = GaussFac * NumSamples - GaussAdd;

// initialize random number generator
static int did_init = InitRand();

// if seed != 0 then use seed as the seed value, otherwise
// use the microseconds component of the current time as the
// seed value
int InitRand(int seed)
{
  /* seed random number generator with current time */
  if(!seed) {
    struct timeval tv;
    gettimeofday(&tv, 0);
    srand48(tv.tv_usec);
  } else {
    srand48(seed);
  }
  NumSamples = 4;
  GaussAdd = sqrt(3.0 * NumSamples);
  GaussFac = (2.0 * GaussAdd) / (double)NumSamples;
  grand_min = -GaussAdd;
  grand_max = GaussFac * NumSamples - GaussAdd;
  return 1;
}

// generate random numbers that have a gaussian distribution
double grand()
{
  double sum = 0.0;
  for(int i = 0; i < NumSamples; i++)
    sum += rrand(1.0);
  return GaussFac*sum - GaussAdd;
}

// normalized gaussian random number generator
// (i.e. return values in [0, x])
double ngrand(double x)
{
  return x * ((grand() - grand_min) / (grand_max - grand_min));
}

// computes gaussian function at given time t and sigma values
Real Gaussian(Real sigma, Real t)
{
  Real sigsq = sigma*sigma;
  Real coeff = 1.0 / sqrt(2.0 * M_PI);
  return (coeff/sigsq) * exp(-0.5 * ((t * t)/sigsq));
}

// use precomputed sin/cos tables or actual sin/cos functions?
#ifdef TRIG_TABLE

// generate precomputed sin/cos tables for all angles that
// are multiples of 0.1 degrees

#define SinCosGran deg2rad(0.05)
#define TanhGran deg2rad(0.0005)
#define SqrtGran 0.0001

SinTab fsin(SinCosGran);
CosTab fcos(SinCosGran);
TanhTab ftanh(TanhGran);
InterpFuncTab fsqrt1(sqrt, 1.0, SqrtGran);

// initialize sin/cos tables (for shared library)
int TrigInit()
{
  SinTab stmp(SinCosGran);
  CosTab ctmp(SinCosGran);
  TanhTab ttmp(TanhGran);
  FuncTab sqtmp(sqrt, 1.0, SqrtGran);
  fsin.Copy(stmp);
  fcos.Copy(ctmp);
  ftanh.Copy(ttmp);
  fsqrt1.Copy(sqtmp);
  return 1;
}

#else

// dummy trig init when not using sin/cos tables
int TrigInit() { return 1; }

#endif

// create a new sin table with given granularity
SinTab::SinTab(double g)
: FuncTab(sin, deg2rad(360.0), g)
{
  // make sure important values are correct
  Value(deg2rad(180.0)) = 0.0;
}

// create a new cos table with given granularity
CosTab::CosTab(double g)
: FuncTab(cos, deg2rad(360.0), g)
{
  // make sure important values are correct
  Value(deg2rad(90.0)) = 0.0;
  Value(deg2rad(270.0)) = 0.0;
}

// create a new tanh table with given granularity
TanhTab::TanhTab(double g)
: FuncTab(tanh, deg2rad(720.0), g)
{
}
