////////////////////////////////////////////////////////////////////////////
//
// trig.h - miscellaneous trig functions and definition of precomputed
//          sin/cos function table classes
//           
//
////////////////////////////////////////////////////////////////////////////
#ifndef _TRIG_H_
#define _TRIG_H_

#include "real.h"
#include "ftab.h"

// convert degrees to radians and back
inline Real deg2rad(Real x) { return (x*M_PI)/180.0; }
inline Real rad2deg(Real x) { return (x*180.0)/M_PI; }

// class for precomputed sin values
class SinTab : public FuncTab {
 public:
  // lookup precomputed sin value. since table only holds values
  // from [0-360] we must make sure that v falls within that range
  double operator()(double v) const
    {
      if(v < 0.0)
	v = M_PI - v;
      if(v >= 2.0 * M_PI)
	v = fmod(v, 2.0 * M_PI);
      return FuncTab::operator()(v);
    }
  SinTab(double g);                   // create sin tab with given granularity
  SinTab() : FuncTab() { }            // create an "empty" sin tab
  SinTab(const SinTab& t) : FuncTab(t) { } // copy a sin tab
};

// class for precomputed cos values
class CosTab : public FuncTab {
 public:
  // lookup precomputed cos value. since table only holds values
  // from [0-360] we must make sure that v falls within that range
  double operator()(double v) const
    {
      if(v < 0.0)
	v = -v;
      if(v >= 2.0 * M_PI)
	v = fmod(v, 2.0 * M_PI);
      return FuncTab::operator()(v);
    }
  CosTab(double g);                   // create cos tab with given granularity
  CosTab() : FuncTab() { }            // create an "empty" cos tab
  CosTab(const CosTab& t) : FuncTab(t) { } // copy a cos tab
};

// class for precomputed tanh values
class TanhTab : public FuncTab {
 public:
  // lookup precomputed tanh value. since table only holds values
  // from [0-720] we must make sure that v falls within that range
  double operator()(double v) const
    {
      if(v < 0.0) {
	return (v <= -4.0 * M_PI) ? -1.0 : -FuncTab::operator()(-v);
      } else {
	return (v >= 4.0 * M_PI) ? 1.0 : FuncTab::operator()(v);
      }
    }
  TanhTab(double g);                  // create tanh tab with given granularity
  TanhTab() : FuncTab() { }           // create an "empty" tanh tab
  TanhTab(const TanhTab& t) : FuncTab(t) { } // copy a tanh tab
};

// use precomputed sin/cos values or actual sin/cos functions
#ifdef TRIG_TABLE
extern SinTab fsin;
extern CosTab fcos;
extern TanhTab ftanh;
extern InterpFuncTab fsqrt1;
#else
#define fsin sin
#define fcos cos
#define ftanh tanh
#define fsqrt1 sqrt;
#endif

// compute tangent (using either tables or real functions
// depending on TRIG_TABLE
inline double ftan(double a) { return fsin(a)/fcos(a); }

// initialize trig tables (in case global ctors cannot be called)
extern int TrigInit();

#endif
