////////////////////////////////////////////////////////////////////////////
//
// vector.h - simple 3d vector class
//
////////////////////////////////////////////////////////////////////////////
#ifndef _VECTOR3D_H_
#define _VECTOR3D_H_

#include "real.h"
#include "trig.h"

class Point3d;

// 3d vector class
class Vector3d {
 public:
  Real i, j, k;   // i, j, k, components of vector

  // allow access to vector as if it were an array of Real[3]
  Real& operator[](int n) { return !n ? i : (n == 1 ? j : k); }

  Vector3d operator*(const Vector3d& v) const    // multiply vectors
    { return Vector3d(i*v.i, j*v.j, k*v.k); }    // (element by element)
  Vector3d operator+(const Vector3d& v) const    // add vectors
    { return Vector3d(i+v.i, j+v.j, k+v.k); }    // (element by element)
  Vector3d operator-(const Vector3d& v) const    // subtract vectors
    { return Vector3d(i-v.i, j-v.j, k-v.k); }    // (element by element)
  Vector3d operator/(const Vector3d& v) const    // divide vectors
    { return Vector3d(i/v.i, j/v.j, k/v.k); }    // (element by element)

  // multiply vector by constant
  Vector3d operator*(Real r) const { return Vector3d(i*r, j*r, k*r); }

  // divide vector by constant
  Vector3d operator/(Real r) const { return Vector3d(i/r, j/r, k/r); }

  // increment vector by constant
  Vector3d operator+(Real r) const { return Vector3d(i+r, j+r, k+r); }

  // decrement vector by constant
  Vector3d operator-(Real r) const { return Vector3d(i-r, j-r, k-r); }

  // compute dot product
  Real Dot(const Vector3d& v) const { return i*v.i + j*v.j + k*v.k; }

  // compute cross product
  Vector3d Cross(const Vector3d& v) const
    { return Vector3d((j*v.k)-(v.j*k), (v.i*k)-(i*v.k), (i*v.j)-(v.i*j)); }

  // compute vector magnitude
  Real Magnitude() const { return sqrt(i*i + j*j + k*k); }

  // normalize vector
  const Vector3d& Normalize()
    { Real mag = Magnitude(); i /= mag; j /= mag; k /= mag; return *this; }

  // create default vector (at origin)
  Vector3d() : i(0.0), j(0.0), k(0.0) { }

  // create vector from a 3d point
  Vector3d(const Point3d& p);

  // create vector from i, j, k components
  Vector3d(Real ii, Real jj, Real kk) : i(ii), j(jj), k(kk) { }

  // destroy a vector
  ~Vector3d() { }

  // print vectors
  friend ostream& operator<<(ostream&, const Vector3d&);
  ostream& Print(ostream& o, int=0) const;
};

// need this include here to prevent circular dependency in includes
#include "point.h"

// actual constructor from point
inline Vector3d::Vector3d(const Point3d& p) : i(p.x), j(p.y), k(p.z) { }

#endif
