#include <errno.h>
#include "pnm++.h"

extern "C" char** pm_allocarray ARGS(( int cols, int rows, int size ));
extern "C" void pm_freearray ARGS(( char** its, int rows ));

// initialize pnm library
static int argc = 1;
static char* argv[] = { "dummy" };
static int init = (pnm_init(&argc, argv), 0);

// print brief information about PNM
ostream& PNM::Print(ostream& os, int) const
{
  Object::Print(os);
  os << "(" << _width << " x " << _height << ", " << _maxval << ")";
  return os;
}

// save PNM to given file
int PNM::Save(char* file) const
{
  FILE* f = fopen(file, "w");
  if(f) {
    pnm_writepnm(f, _data, _width, _height, _maxval, _format, 0);
    fclose(f);
  }
  return errno;
}

// load PNM from given file and return new PNM object
PNM* PNM::Load(char* file)
{
  FILE* f = fopen(file, "r");
  xel** pnm;
  xelval maxval;
  int w, h, format;

  if(f) {
    pnm = pnm_readpnm(f, &w, &h, &maxval, &format);
    fclose(f);

    if(PBM_FORMAT_TYPE(format))
      return new PBM(w, h, pnm, format, maxval);
    else if(PGM_FORMAT_TYPE(format))
      return new PGM(w, h, pnm, format, maxval);
    else if(PPM_FORMAT_TYPE(format))
      return new PPM(w, h, pnm, format, maxval);
    else
      return 0;
  } else {
    return 0;
  }
}

// initialize an empty PNM
PNM::PNM()
: _width(0), _height(0), _maxval(0), _data(0), _format(0)
{
}

// copy a PNM
PNM::PNM(const PNM& pnm)
: _width(pnm._width), _height(pnm._height), _maxval(pnm._maxval),
  _data(pnm_allocarray(pnm._width, pnm._height)), _format(pnm._format)
{
  for(int i = 0; i < _width; i++)
    for(int j = 0; j < _height; j++)
      *Elt(i, j) = *((PNM&)pnm).Elt(i, j);
}

// create a new PNM from data
PNM::PNM(int format, int w, int h, xel** xels, xelval maxv)
: _width(w), _height(h), _maxval(maxv),
  _data(xels ? xels : pnm_allocarray(w, h)),
  _format(format)
{
}

// destroy a PNM
PNM::~PNM()
{
  pnm_freearray(_data, _height);
}

// create a PBM
PBM::PBM(int w, int h, unsigned char* xels, int f, xelval maxv)
: PNM(f, w, h, 0, maxv)
{
  for(int i = 0; i < w; i++)
    for(int j = 0; j < h; j++)
      (*this)(i, j) = *xels++;
}

// create a PGM
PGM::PGM(int w, int h, unsigned char* xels, int f, xelval maxv)
: PNM(f, w, h, 0, maxv)
{
  for(int i = 0; i < w; i++)
    for(int j = 0; j < h; j++)
      (*this)(i, j) = *xels++;
}

// create a PPM
PPM::PPM(int w, int h, RGB* xels, int f, xelval maxv)
: PNM(f, w, h, 0, maxv)
{
  for(int i = 0; i < w; i++)
    for(int j = 0; j < h; j++)
      (*this)(i, j) = *xels++;
}


// convert an Xel to a Color
Xel::operator const Color&() const
{
  static RGB rgb;
  static GrayScale gray;
  static BitColor bit;

  if(_type == PPM_TYPE)
    return (rgb = (RGB)(*this));
  else if(_type == PBM_TYPE)
    return (bit = (BitColor)(*this));
  else if(_type == PGM_TYPE)
    return (gray = (GrayScale)(*this));
  else {
    assert(0);
    return (bit = 0);
  }
}

// assign an xel an Xel
const Xel& Xel::operator=(const xel& x)
{
  if(_type == PPM_TYPE)
    PPM_ASSIGN(*_val, PPM_GETR(x), PPM_GETG(x), PPM_GETB(x));
  else
    PNM_ASSIGN1(*_val, PNM_GET1(x));
  return *this;
}

// assign an xelval to an Xel (only for PGMs and PBMs)
const Xel& Xel::operator=(const xelval& x)
{
  PNM_ASSIGN1(*_val, x);
  return *this;
}

// assign one Xel to another
const Xel& Xel::operator=(const Xel& x)
{
  if(_type == PPM_TYPE) {
    if(x._type == PPM_TYPE)
      *_val = *x._val;
    else
      PPM_ASSIGN(*_val, (xelval)x, (xelval)x, (xelval)x);
  } else if(_type == PGM_TYPE) {
    if(x._type == PPM_TYPE)
      PNM_ASSIGN1(*_val, (int)(PGM_MAXMAXVAL * (PPM_LUMIN((xel)x) / PPM_MAXMAXVAL)));
    else
      PNM_ASSIGN1(*_val, (xelval)x);
  } else {
    if(x._type == PPM_TYPE)
      PNM_ASSIGN1(*_val, x.Red() + x.Green() + x.Blue());
    else
      PNM_ASSIGN1(*_val, (xelval)x);
  }
  return *this;
}

// assign a boolean to an Xel (only for PBMs)
const Xel& Xel::operator=(Boolean b)
{
  assert(_type == PBM_TYPE);
  PNM_ASSIGN1(*_val, b);
  return *this;
}

// assign an integer intensity to an Xel (only for PGMs)
const Xel& Xel::operator=(unsigned int intens)
{
  assert(_type == PGM_TYPE);
  PNM_ASSIGN1(*_val, intens);
  return *this;
}

// assign an arbitrary color an Xel (for any kind of PNM)
const Xel& Xel::operator=(const Color& c)
{
  if(c.IsA(RGB::Id())) {
    assert(_type == PPM_TYPE);
    const RGB& rgb = (const RGB&)c;
    PPM_ASSIGN(*_val,
	       (int)(rgb.r() * PPM_MAXMAXVAL),
	       (int)(rgb.g() * PPM_MAXMAXVAL),
	       (int)(rgb.b() * PPM_MAXMAXVAL));
  } else if(c.IsA(BitColor::Id())) {
    return operator=(((const BitColor&)c).Bit());
  } else if(c.IsA(GrayScale::Id())) {
    Real i = c.Intensity();
    return operator=((unsigned int)(i * PGM_MAXMAXVAL));
  } else {
    assert(0);
  }
  return *this;
}

// print info about an Xel
ostream& operator<<(ostream& os, const Xel& x)
{
  switch(x.Type()) {
  case PBM_TYPE:
    os << "#<BitXel " << (int)x.BitValue() << ">";
    return os;
  case PGM_TYPE:
    os << "#<GrayXel " << x.Intensity() << ">";
    return os;
  case PPM_TYPE:
    os << "#<ColorXel " << x.Red() << ", " << x.Green() << ", " << x.Blue() << ">";
    return os;
  }
  return os;
}
