////////////////////////////////////////////////////////////////////////////
//
// pnm++.h - this file implements a C++ wrapper around pnm objects from
//           the pnm library. (do man libpnm for more info on pnms)
//
////////////////////////////////////////////////////////////////////////////
#ifndef _PNM_PP_H_
#define _PNM_PP_H_

#include <stdio.h>
#include "Object.h"
#include "color.h"

extern "C" {
// this weirdness is because pbmplus.h is a mess
#ifndef __GNUC__
#define _PBMPLUS_H_
#define ARGS(a) a
#else
#define strlen stupid_strlen
#endif
#include <pnm.h>
#undef strlen
#undef const
}

#include "xel++.h"

// original PGM maxval is too big, reset it to a "sane" value
#undef PGM_MAXMAXVAL
#define PGM_MAXMAXVAL 255

// generic abstract PNM class
// this class defines the basic data and operations on all PNM
// objects.  It also has methods for loading unknown types of
// PNMs and creating the correct object for them (i.e. PGM, PPM
// or PBM)
class PNM : public Object {
  DeclareStandardMethods(PNM, Object);
 public:
  // load an arbitrary pnm file and return a newly
  // created PNM object of the correct type for it.
  static PNM* Load(char*);

  // save this PNM
  int Save(char*) const;

  // return the number of elements in this PNM
  int Size() const { return _width*_height; }

  // return the width of this PNM
  int Width() const { return _width; }
  int Height() const { return _height; }

  // return the maximum value of this PNM. this value
  // defines the ranges that Xel values can have.
  // (e.g. for PGMs, gray scale values will be between 0
  // and MaxVal().  for PPMs and the RGB values will be
  // between 0 and MaxVal())
  xelval MaxVal() const { return _maxval; }

  // return the format type of this PNM (see libpnm for format types)
  int Format() const { return _format; }

  // return a pointer to the given xel in this PNM
  xel* Elt(int i, int j) { return &((PNM*)this)->_data[j][i]; }

  // return Xel at given position in this PNM
  virtual Xel operator()(int i, int j) = 0;

  // delete a PNM
  ~PNM();

 protected:
  PNM();              // create an empty PNM
  PNM(const PNM&);    // copy a PNM
  // create a PNM with given format size and data
  PNM(int format, int w, int h, xel** = 0, xelval = PNM_MAXMAXVAL);
 private:
  int _width;      // width of PNM in pixels
  int _height;     // height of PNM in pixels
  xelval _maxval;  // defines maximum range of Xel values
  xel** _data;     // pointer to actual image data
  int _format;     // format of this PNM
};

// PBM class (binary image)
class PBM : public PNM {
  DeclareStandardMethodsNoPrint(PBM, PNM);
 public:
  // return PBM Xel at given location
  Xel operator()(int i, int j) { return Xel(PBM_TYPE, Elt(i, j)); }

  // create empty PBM
  PBM() { }

  // copy an existing PBM
  PBM(const PBM& pbm) : PNM(pbm) { }

  // create PBM of given size and data
  PBM(int w, int h, xel** xels = 0, int f = PBM_FORMAT,
      xelval maxv = pnm_pbmmaxval) : PNM(f, w, h, xels, maxv) { }

  // create PBM of given size and an array of char data
  PBM(int w, int h, unsigned char* xels, int f = PBM_FORMAT,
      xelval maxv = pnm_pbmmaxval);
};


// PGM class (gray scale image)
class PGM : public PNM {
  DeclareStandardMethodsNoPrint(PGM, PNM);
 public:
  // return PGM Xel at given location
  Xel operator()(int i, int j) { return Xel(PGM_TYPE, Elt(i, j)); }

  // create empty PGM
  PGM() { }

  // copy an existing PGM
  PGM(const PGM& pgm) : PNM(pgm) { }

  // create PGM of given size and data
  PGM(int w, int h, xel** xels = 0, int f = PGM_FORMAT,
      xelval maxv = PGM_MAXMAXVAL) : PNM(f, w, h, xels, maxv) { }

  // create PGM of given size and an array of char data
  PGM(int w, int h, unsigned char* xels, int f = PGM_FORMAT,
      xelval maxv = PGM_MAXMAXVAL);
};


// PPM class (RGB image)
class PPM : public PNM {
  DeclareStandardMethodsNoPrint(PPM, PNM);
 public:
  // return PPM Xel at given location
  Xel operator()(int i, int j) { return Xel(PPM_TYPE, Elt(i, j)); }

  // create empty PPM
  PPM() { }

  // copy an existing PPM
  PPM(const PPM& ppm) : PNM(ppm) { }

  // create PPM of given size and data
  PPM(int w, int h, xel** xels = 0, int f = PPM_FORMAT,
      xelval maxv = PPM_MAXMAXVAL) : PNM(f, w, h, xels, maxv) { }

  // create PPM of given size and an array of RGB data
  PPM(int w, int h, RGB* xels, int f = PPM_FORMAT,
      xelval maxv = PPM_MAXMAXVAL);
};

#endif
