////////////////////////////////////////////////////////////////////////////
//
// xel++.h - this file implements a C++ wrapper around the xel object from
//           the pnm library. (do man libpnm for more info on xels & pnms)
//
////////////////////////////////////////////////////////////////////////////
#ifndef _XEL_PP_H_
#define _XEL_PP_H_

// The Xel class is a C++ wrapper around the xel object from the pnm library.
// PNM pictures are composed of Xels (which represent gray scale, RGB or
// binary values).  Each Xel points to a specific element in a PNM, so 
// assignment to Xel's change the actual data inside the PNM.
// Note: should really get maxval from PNM that Xel came from!
class Xel {
 public:
  // return the type of this Xel (PBM_TYPE, PPM_TYPE or PGM_TYPE)
  int Type() const { return _type; }

  // convert to xel (actual libpnm object)
  operator xel() const { return *_val; }

  // convert to xelval (component of xel, only for PGMs and PBMs)
  operator xelval() const { return PNM_GET1(*_val); }

  // convert this Xel to an RGB value (only for PPMs)
  operator RGB() const { return RGB(Red(), Green(), Blue(), PPM_MAXMAXVAL); }

  // convert this Xel to an GrayScale value (only for PGMs)
  operator GrayScale() const { return GrayScale(Intensity(), PGM_MAXMAXVAL); }

  // convert this Xel to a bit value (only for PBMs)
  operator BitColor() const { return BitColor(BitValue()); }

  // convert this Xel into a Color (works for and type)
  operator const Color&() const;

  // return the binary value of this Xel (only for PBMs)
  Boolean BitValue() const { return MakeBoolean((xelval)(*this)); }

  // return the intensity value of this Xel (only for PBMs or PGMs)
  unsigned int Intensity() const { return (xelval)(*this); }

  // return the red/green/blue components of this Xel (only for PPMs)
  int Red() const { return PPM_GETR(*_val); }
  int Green() const { return PPM_GETG(*_val); }
  int Blue() const { return PPM_GETB(*_val); }

  // different assignment operators (see pnm++.c for more info)
  const Xel& operator=(const xel&);
  const Xel& operator=(const xelval& x);
  const Xel& operator=(const Xel& x);
  const Xel& operator=(Boolean);
  const Xel& operator=(unsigned int);
  const Xel& operator=(const Color&);

  // create a new Xel by copying another one
  Xel(const Xel& x) : _val(x._val), _type(x._type) { }

  // delete an Xel
  ~Xel() { }
 protected:
  // create empty Xel
  Xel() : _val(0), _type(0) { }
  // create Xel which actually points to an element of a PNM
  Xel(int type, xel* x) : _val(x), _type(type) { }

  // only PNM objects can create Xels this way
  friend class PNM;
  friend class PGM;
  friend class PPM;
  friend class PBM;

  // print an Xel
  friend ostream& operator<<(ostream&, const Xel&);
 private:
  xel* _val;                // pointer to actual xel in a PNM
  unsigned short _type;     // type of Xel (PPM_TYPE, PGM_TYPE, or PBM_TYPE)
};

#endif
