////////////////////////////////////////////////////////////////////////////
//
// color.h - this file defines color objects (RGB, GrayScale and BitColor)
//           used for representing object color.
//
////////////////////////////////////////////////////////////////////////////
#ifndef _COLOR_H_
#define _COLOR_H_

#include <stream.h>
#include <limits.h>
#include "Object.h"
#include "real.h"

// abstract color object, all color objects are the same size to make
// certain things easier.  Even though this is very wasteful for gray
// scale and binary colors.  (I'm assuming RGB colors will be the most
// common anyway)
class Color : public Object {
  DeclareStandardMethodsNoPrint(Color, Object);
 public:
  // return intensity of this color
  virtual Real Intensity() const = 0;
  virtual void Brighten(Real) = 0;
  virtual void Darken(Real) = 0;
  virtual void Scale(Real) = 0;

  // assign to this color
  Boolean operator==(const Color& c) const;

  // destroy a color object
  ~Color() { }
 protected:
  // create an empty "black" color
  Color() : _r(0.0), _g(0.0), _b(0.0) { }

  // copy another color
  Color(const Color& c) : _r(c._r), _g(c._g), _b(c._b) { }

  // create color using single value (for gray scale and bit colors)
  Color(unsigned int i) : _r((Real)i), _g(0.0), _b(0.0) { }

  // set rgb color components (for color values)
  Color(Real r, Real g, Real b) : _r(r), _g(g), _b(b) { }

  // return integer value of this color (for gray scale and bit colors)
  unsigned int intval() const { return (unsigned int)_r; }

  // return rgb components (for RGB colors)
  Real& rval() { return _r; }
  Real& gval() { return _g; }
  Real& bval() { return _b; }

  // set integer color value (for gray and bit colors)
  void set(unsigned int i) { _r = (Real)i; }

  // set RGB color values (for RGB colors)
  void set(Real r, Real g, Real b) { _r = r; _g = g; _b = b; }
 private:
  Real _r, _g, _b;  // color data
};

// normalized rgb color (i.e. all components range between [0, 1.0])
class RGB : public Color {
  DeclareStandardMethods(RGB, Color);
 public:
  RGB operator*(Real) const;          // multiply color by constant factor
  RGB operator*(const Color&) const;  // multiply two colors
  RGB operator+(Real) const;          // add constant factor to color
  RGB operator+(const Color&) const;  // add two colors

  // return red/green/blue components of this color
  Real r() const { return ((RGB*)this)->rval(); }
  Real g() const { return ((RGB*)this)->gval(); }
  Real b() const { return ((RGB*)this)->bval(); }

  // set red/green/blue components of this color
  void SetRed(Real v) { rval() = v; Check(); }
  void SetGreen(Real v) { gval() = v; Check(); }
  void SetBlue(Real v) { bval() = v; Check(); }

  // set red/green/blue components as an integer
  // max controls the maximum value that any of these
  // integers can be.
  void Set(int r, int g, int b, unsigned int max = 1<<24)
    { Real m = max; Set(r/m, g/m, b/m); }

  // set red/green/blue components at once
  void Set(Real r, Real g, Real b) { set(r, g, b); Check(); }

  // compute intensity of this color
  Real Intensity() const { return 0.299*r() + 0.587*g() + 0.114*b(); }
  void Brighten(Real);
  void Darken(Real);
  void Scale(Real);

  // create "black" color
  RGB() : Color(0.0, 0.0, 0.0) { }

  // create RGB from float values
  RGB(float i) { Set(i, i, i); }

  // copy another color
  RGB(const Color& c) : Color(c) { }

  // create RGB color for integers that range between [0, max]
  RGB(int r, int g, int b, unsigned int max = 1<<24) { Set(r, g, b, max); }

  // create RGB from float values
  RGB(float r, float g, float b) { Set(r, g, b); }

  // create RGB from double values
  RGB(double r, double g, double b) { Set((float)r, (float)g, (float)b); }

  // destroy an RGB
  ~RGB() { }
 protected:
  // verify all color components are in the correct range
  void Check()
    {
      assert(r() >= 0.0 && r() <= 1.0);
      assert(g() >= 0.0 && g() <= 1.0);
      assert(b() >= 0.0 && b() <= 1.0);
    }
};

// normalized gray scale intensity (i.e. [0, 1.0])
class GrayScale : public Color {
  DeclareStandardMethods(GrayScale, Color);
 public:
  // convert to Real value (just return intensity)
  operator Real() const { return Intensity(); }
  GrayScale operator*(Real) const;         // multiply color by constant factor
  GrayScale operator*(const Color&) const; // multiply two colors
  GrayScale operator+(Real) const;         // add constant factor to color
  GrayScale operator+(const Color&) const; // add two colors

  // assign intensities to gray scale
  const GrayScale& operator=(Real i) { Set(i); return *this; }

  // assign integer intensities to gray scale (assumes each
  // intensity ranges between [0, max for integer type])
  const GrayScale& operator=(unsigned char i) { Set(i, UCHAR_MAX); return *this; }
  const GrayScale& operator=(unsigned short i) { Set(i, USHRT_MAX); return *this; }
  const GrayScale& operator=(unsigned int i) { Set(i, UINT_MAX); return *this; }

  // set intensity
  void Set(Real i) { rval() = i; Check(); }
  void Set(unsigned int i, unsigned int max = UCHAR_MAX) { Set(i/(Real)max); Check(); }

  // return intensity of this color
  Real Intensity() const { return ((GrayScale*)this)->rval(); }
  void Brighten(Real f) { if((f *= rval()) > 1.0) f = 1.0; Set(f); }
  void Darken(Real f) { Set(rval()/f); }
  void Scale(Real f) { Brighten(f); }

  // create "black" color
  GrayScale() { }

  // copy another color
  GrayScale(const Color& c) : Color(c) { }

  // create GrayScale from integer intensity (see operator=)
  GrayScale(unsigned char i) { Set(i, UCHAR_MAX); }
  GrayScale(unsigned short i) { Set(i, USHRT_MAX); }
  GrayScale(unsigned int i, unsigned int max = UINT_MAX) { Set(i, max); }
  
  // create GrayScale from float/double intensity
  GrayScale(float i) { Set((Real)i); }
  GrayScale(double i) { Set((Real)i); }

  // destroy a gray scale
  ~GrayScale() { }
 protected:
  // make sure intensity value is in range
  void Check() { assert(rval() >= 0.0 && rval() <= 1.0); }
};

// bit color (0 or 1)
class BitColor : public Color {
  DeclareStandardMethods(BitColor, Color);
 public:
  // convert binary color to integer
  operator unsigned int() const { return Bit(); }
  BitColor operator*(Real) const;         // multiply by constant (i.e. "and")
  BitColor operator*(const Color&) const; // multiply two colors (i.e. "and")
  BitColor operator+(Real) const;         // add constant (i.e. "or")
  BitColor operator+(const Color&) const; // add colors (i.e. "or")

  // set bit color from real value
  // (if value is in between 0 & 1 it is thresholded at 0.5)
  const BitColor& operator=(Real i) { set((i >= 0.5) ? 1 : 0); return *this; }

  // set bit color from integer
  const BitColor& operator=(int i) { set(i ? 1 : 0); return *this; }

  // return intensity of bit color (e.g 0 or 1)
  Real Intensity() const { return (Real)intval(); }
  void Brighten(Real) { }
  void Darken(Real) { }
  void Scale(Real) { }

  // return value of bit color (e.g 0 or 1)
  unsigned int Bit() const { return intval(); }

  // create "off" bit color
  BitColor() : Color(0) { }

  // copy bit color
  BitColor(const Color& c) : Color(c) { }

  // create bit color from integer
  BitColor(unsigned int i) : Color(i ? 1 : 0) { }

  // create bit color from real value (see operator=)
  BitColor(Real i) : Color(i >= 0.5 ? 1 : 0) { }

  // destroy bit color
  ~BitColor() { }
};

#endif
