////////////////////////////////////////////////////////////////////////////
//
// world.h - this file defines a "World" class which holds environment
//           data used for rendering (e.h. light source, camera, etc.)
//
////////////////////////////////////////////////////////////////////////////
#ifndef _WORLD_H_
#define _WORLD_H_

#include "real.h"
#include "light.h"

// World class
class World {
 public:
  // distance from eye to image plane
  Real F() const { return _f; }
  // constant for ambient light in lighting equation
  Real& Ka() const { return ((World*)this)->_Ka; }
  // constant for diffuse light in lighting equation
  Real& Kd() const { return ((World*)this)->_Kd; }
  // constant for specular light in lighting equation
  Real& Ks() const { return ((World*)this)->_Ks; }

  // compute ambient color
  RGB Ia() const
    {
      Real rc = fixrrng(_Ka * _light.color().r() * _Ia.r(), 0.0, 1.0);
      Real gc = fixrrng(_Ka * _light.color().g() * _Ia.g(), 0.0, 1.0);
      Real bc = fixrrng(_Ka * _light.color().b() * _Ia.b(), 0.0, 1.0);
      return RGB(rc, gc, bc);
    }

  // compute diffuse color component with given object normal and color
  RGB Id(const Vector3d& n, const RGB& c) const
    {
      Real tmp = _Kd * n.Dot(_light);
      Real rc = fixrrng(_light.color().r() * c.r(), 0.0, 1.0);
      Real gc = fixrrng(_light.color().g() * c.g(), 0.0, 1.0);
      Real bc = fixrrng(_light.color().b() * c.b(), 0.0, 1.0);
      return RGB(rc, gc, bc);
    }

  // compute specular color value given the object normal, color and shininess
  RGB Is(const Vector3d& n, const RGB& c, Real e) const
    {
      Real fact = 2.0 * n.Dot(_light.Direction());
      Vector3d nn(n.i * fact, n.j * fact, n.k * fact);
      Real base = _eye.Dot(nn - _light.Direction());
      Real tmp = (base > 0.0) ? _Ks * pow(base, e) : 0.0;
      Real rc = fixrrng(c.r() * tmp, 0.0, 1.0);
      Real gc = fixrrng(c.g() * tmp, 0.0, 1.0);
      Real bc = fixrrng(c.b() * tmp, 0.0, 1.0);
      return RGB(rc, gc, bc);
    }

  // compute color value given object normal, color and shininess
  RGB I(const Vector3d& n, const RGB& c, Real e) const
    { return Ia() + Id(n, c) + Is(n, c, e); }

  // return location of eye
  Vector3d& Eye() const { return ((World*)this)->_eye; }

  // return light
  Light& light() const { return ((World*)this)->_light; }

  // create a new world object with light, eye and lighting constants
  World(const Light& l, const Vector3d& e, Real f, Real ka = 0.1,
	Real kd = 0.9, Real ks = 0.2, RGB Ia = 0.2)
    : _light(l), _eye(e), _Ka(ka), _Kd(kd), _Ks(ks), _Ia(Ia), _f(f)
    { _eye.Normalize(); }
 private:
  Light _light;   // world light (inifinitely distant point light source)
  Vector3d _eye;  // eye location
  Real _Ka;       // coefficient for ambient light
  Real _Kd;       // coefficient for diffuse light
  Real _Ks;       // coefficient for specular light
  RGB _Ia;       // default amount of ambient light
  Real _f;        // distance to image plane from eye
};

#endif
